module AVL
  ( Tree, empty, insert, depth, search, toList, isEmpty, fmapM )
 where

data Tree k v = Empty | Node Int (Tree k v) k v (Tree k v) 
              deriving (Eq, Show)

instance Functor (Tree k) where
  fmap _ Empty = Empty
  fmap f (Node d l k v r) = Node d (fmap f l) k (f v) (fmap f r)

fmapM _ Empty = return Empty
fmapM f (Node d l k v r) = do l' <- fmapM f l
                              v' <- f v
                              r' <- fmapM f r
                              return (Node d l' k v' r')

empty :: Tree k v 
empty = Empty

isEmpty :: Tree k v -> Bool
isEmpty Empty = True
isEmpty _ = False

depth :: Tree k v -> Int
depth Empty = 0
depth (Node d _ _ _ _) = d

toList :: Tree k v -> [(k,v)]
toList t = inorder t []

inorder Empty = id
inorder (Node _ l k v r) = inorder l . ((k,v):) . inorder r

search :: Ord k => k -> Tree k v -> Maybe v
search _ Empty = Nothing
search sk (Node _ l k v r) =
  case (compare sk k) of
    EQ -> Just v
    LT -> search sk l
    GT -> search sk r

insert :: Ord k => k -> v -> Tree k v -> Tree k v
insert nk nv Empty = Node 1 Empty nk nv Empty
insert nk nv (Node d l k v r) = 
  case (compare nk k) of
    EQ -> Node d l k nv r
    LT -> join (insert nk nv l) k v r
    GT -> join l k v (insert nk nv r)

{- Non balancing insert; for comparisons
nbinsert :: Ord k => k -> v -> Tree k v -> Tree k v
nbinsert nk nv Empty = Node 1 Empty nk nv Empty
nbinsert nk nv (Node d l k v r) = 
  case (compare nk k) of
    EQ -> Node d l k nv r
    LT -> let nl = nbinsert nk nv l
              nd = 1 + max (depth nl) (depth r)
           in Node nd nl k v r
    GT -> let nr = nbinsert nk nv r
              nd = 1 + max (depth l) (depth nr)
           in Node nd l k v nr
-}
    
-- Auxiliary functions for insert
join l k v r  
  | abs (ld-rd) <= 1 = sJoin l k v r
  | ld == rd+2       = lJoin l k v r
  | ld+2 == rd       = rJoin l k v r
  | ld > rd+2        = join ll lk lv (join lr k v r)
  | ld+2 < rd        = join (join l k v rl) rk rv rr
    where ld = depth l   -- DO NOT CHANGE these two lines, because they are
          rd = depth r   --  there to permit Empty cases in the two lines below.
          (Node _ ll lk lv lr) = l  -- These two lines are not reached when l or
          (Node _ rl rk rv rr) = r  --  r are empty because of lazy evaluation. 

lJoin l k v r
  | lld >= lrd = sJoin ll lk lv (sJoin lr k v r)
  | otherwise  = sJoin (sJoin ll lk lv lrl) 
                          lrk lrv
                          (sJoin lrr k v r)
    where (Node ld ll lk lv lr) = l
          lld = depth ll
          lrd = depth lr
          (Node _ lrl lrk lrv lrr) = lr

rJoin l k v r
  | rrd >= rld = sJoin (sJoin l k v rl) rk rv rr
  | otherwise  = sJoin (sJoin l k v rll) 
                          rlk rlv
                          (sJoin rlr rk rv rr)
    where (Node rd rl rk rv rr) = r
          rrd = depth rr
          rld = depth rl
          (Node _ rll rlk rlv rlr) = rl

sJoin l k v r = Node (1+ max (depth l) (depth r)) l k v r          
