-- Parsing library in Haskell
--
-- This version produces good error messages and fixes a long-time space leak
--
--  By John Hughes (1997)
--  Additions and modifications: Pablo E. Martinez Lopez (1999)

module BasicParser
  ( module Monad
  , module MoreMonad
  , Parser
  , returnP, bindP, failP, orelse, opt, cut
  , mark, withError
  , mapP, binP, apP, doP, pairP, consP
  , satisfy, literal, token, eof
  , position --, offside
  , parse, parseIO --, reparse
  ) 
 where
import Monad
import MoreMonad
import Error
import Position

infixl 8 `apP`, `doP`, `bindP`
infixr 6 `consP`, `pairP`
infix  5 `opt`
infixr 2 `orelse`
infix  1 `withError`

-------------
-- The parsing monad -- keeps track of high water mark
-------------
data Input t = Input Position (PosList t)
                     -- Last token seen in this thread, remaining input
data HighWaterMark = HWM Position (Error ())
                         -- Position and error of deepest attempt

combineHWM hwm1@(HWM pos1 emsg1) hwm2@(HWM pos2 emsg2) =
  case compare pos1 pos2 of
    LT -> hwm2
    EQ -> HWM pos1 (emsg1 `addErrors` emsg2)
    GT -> hwm1

newtype Parser t a = P (Input t -> Result t a)
newtype Result t a = 
  R (HighWaterMark ->
       (HighWaterMark,
        Maybe ( a, Input t, Maybe (Result t a) )
       )
    )
-- A result maps the high-water-mark from deepest parsing attempt 
--  (the position and possible error at that position) to
--   1.- the high-water-mark including the next attempt at parsing
--   2.- Maybe indicating whether or not a parse was found
--        a.- value 
--        b.- remaining input: last token seen and unconsumed input
--        c.- Maybe indicating whether or not more parses *might* be found, 
--            (permitting more space efficient code in the deterministic case).

-- Convenience functions to perform pattern matching
unParser :: Parser t a -> Input t -> Result t a
unParser (P p) = p

unR :: Result t a -> HighWaterMark ->
       ( HighWaterMark, Maybe ( a, Input t, Maybe (Result t a) ) )
unR (R x) = x

-------------
-- Monadic presentation of parsers
returnP :: a -> Parser t a
returnP x = P$ \inp -> returnR (Just (x,inp,Nothing))

returnR :: Maybe (a, Input t, Maybe (Result t a)) -> Result t a
returnR x = R$ \hwm -> (hwm, x)

bindP :: Parser t a -> (a -> Parser t b) -> Parser t b          
P p `bindP` f = P$ \inp -> (p inp) `after` f
  where 
    (R r) `after` f = 
     R$ \hwm ->
     case (r hwm) of
      (hwm', Nothing)             -> (hwm', Nothing)
      (hwm', Just(x,inp',Nothing)) -> unR (unParser (f x) inp') hwm'
      (hwm', Just(x,inp',Just r')) -> unR ((unParser (f x) inp')
                                           `orelseR` 
                                           (r' `after` f)
                                          ) hwm'
         
-------------
-- Primitives that depend on the monad
errorP :: String -> Parser t a
errorP msg = P$ \(Input curpos _) -> errorR (HWM curpos (Error msg))
                  
errorR :: HighWaterMark -> Result t a
errorR cwm = R$ \hwm -> let new_hwm = combineHWM hwm cwm
                         in new_hwm `seq` (new_hwm, Nothing)

failP :: Parser t a
failP = P$  \(Input curpos _) -> 
               failR (HWM curpos (Error "Failed (no error report)"))

failR :: HighWaterMark -> Result t a
failR cwm@(HWM cpos cmsg) =
   R$ \hwm@(HWM hpos emsg) -> 
        let new_hwm = case compare hpos cpos of
                        LT -> cwm
                        EQ -> hwm
                        GT -> hwm
         in new_hwm `seq` (new_hwm, Nothing)

orelse :: Parser t a -> Parser t a -> Parser t a
P f `orelse` P g = P$ \inp -> f inp `orelseR` g inp

orelseR :: Result t a -> Result t a -> Result t a
R f `orelseR` q@(R g) = 
  R$ \hwm ->
  case f hwm of
    (hwm', Nothing) -> case hwm' of
                        HWM _ (FatalError _) -> (hwm', Nothing)
                        _                    -> g hwm'
    (hwm', Just (x, inp, cont)) ->
      (hwm', Just (x, inp, case cont of 
                            Nothing -> Just q
                            Just r  -> Just (r `orelseR` q)
                  ))

-- It the parsing fails, it provides an alternative success element
opt :: Parser t a -> a -> Parser t a
P f `opt` x = 
  P$ \inp@(Input _ ts) -> R$ \hwm ->
   let cpos = pos ts in cpos `seq`
   let (hwm'@(HWM epos emsg), res) = unR (f inp) hwm 
    in case res of
        Nothing -> case emsg of
                    FatalError _ -> (hwm', Nothing)
                    Success _    -> (hwm', Just (x,inp,Nothing))
                    Error _      -> if cpos == epos
                                     then (HWM epos (Success()), Just(x,inp,Nothing))
                                     else (hwm', Just (x,inp,Nothing))
        Just _  -> (hwm', res)

withError :: Parser t a -> String -> Parser t a
P f `withError` msg = 
  P$ \inp@(Input _ ts) -> R$ \hwm ->
   let cpos = pos ts in cpos `seq`
   let (hwm'@(HWM epos emsg), res) = unR (f inp) (HWM cpos (Success())) 
    in case emsg of
        FatalError _ -> (HWM epos emsg, res)
        Success _    -> (combineHWM hwm hwm', res)
        Error _      -> if cpos == epos
                         then (combineHWM hwm (HWM cpos (Error msg)), res)
                         else (combineHWM hwm hwm', res)

cut :: Parser t a -> Parser t a
cut (P f) = 
  P$ \inp -> R$ \hwm ->
  let (hwm', res) = unR (f inp) hwm in
  case res of
    Nothing -> (hwm', Nothing)
    Just (y,inp',_)  -> (hwm', Just (y, inp', Nothing))

mark :: Parser t a -> Parser t a
mark (P f) = 
  P$ \inp -> R$ \hwm ->
  let pres@(HWM epos _, res) = unR (f inp) hwm 
   in case res of
        Nothing -> pres
        Just _  -> (HWM epos (Success ()), res)

-------------
-- Primitives corresponding strictly to parsing
-------------

-------------
-- Low-level parsing primitives
-- It parses a token if it satisfies the predicate p
satisfy :: Show t => (t -> Bool) -> Parser t t
satisfy p  = do tok <- token
                if (p tok)
                 then return tok
                 else errorP ("Unexpected " ++ show tok ++ "!") 
                
-- It parses a token if it is equal to t                
literal :: (Eq t, Show t) => t -> Parser t t
literal t = do tok <- token `withError` (show t ++ " expected!")
               if (tok==t)
                then return tok
                else errorP (show t ++ " expected!") 

-- It returns the next token, if it exists, or it fails, otherwise
token :: Parser t t
token = 
  P$ \(Input cpos ts) -> 
        case ts of
         (npos,t):ts' -> returnR (Just (t,Input npos ts', Nothing))
         []           -> errorR (HWM EndofFile (Error "No tokens!"))

-- It succeed on end of input, fails otherwise
eof :: Parser t ()
eof = P$ \(Input cpos ts) ->
      case ts of
        _:_ -> errorR (HWM cpos (Error "Not eof!"))
        []  -> returnR (Just ((), Input EndofFile [], Nothing))

-------------
-- Primitives to handle the position of high-water-mark
-- It returns the current position in the input token-list
position :: Parser t Position
position = P$ \inp@(Input cpos _) -> returnR (Just (cpos,inp,Nothing))

-------------
-- Invoking a parser
parse :: Parser t a -> PosList t -> Error a
parse (P p) ts = 
  let fpos = pos ts in fpos `seq`
  case unR (p (Input NoPos ts)) (HWM NoPos (Success())) of
    (_, Just (x,_,_)) -> Success x
    (HWM epos emsg, Nothing) ->
      case emsg of
       FatalError msg -> Error ("Fatal error: '"++msg++"' at "++displayPos epos)
       Error msg      -> Error("Syntax error: '"++msg++"' at "++displayPos epos)
       Success ()     -> Error("Syntax error with no error message at "
                               ++displayPos epos)

parseIO :: Show a => Parser t a -> Error (PosList t) -> IO ()
parseIO p lt = case lt of
                 FatalError e -> putStr e
                 Error e      -> putStr e
                 Success ts   -> case (parse p ts) of
                                  FatalError e -> putStr e
                                  Error e      -> putStr e
                                  Success x    -> print x

{-
-- It parses the token-list ss, and adds the current token-list ts to the rest ss'
reparse :: Parser t a -> PosList t -> Parser t a
reparse (P f) ss = P$ \ts -> rep (f ss) ts
  where 
   rep (R r) ts = 
    R$ \h emsg-> 
     let h'= if (h==EndofFile) then pos ts else h in h' `seq` 
      case r h emsg of
       (h',FatalError msg,Nothing)    -> (h', FatalError msg, Nothing)
       (h',Error msg,Nothing)         -> (h', Error msg, Nothing)
       (h',Success _,Nothing)         -> (h', Success (), Nothing)
       (h',emsg',Just(x,ss',Nothing)) -> (h', emsg', Just(x,ss'++ts,Nothing))
       (h',emsg',Just(x,ss',Just r')) -> (h', emsg', Just(x,ss'++ts,Just (rep r' ts)))

-- Combinator to manage the offside rule
offside :: Parser a b -> Parser a b
offside e = do p   <- position
               inp <- many $ do q <- position
                                if (column p <= column q)
                                 then do t <- token
                                         return (q,t)
                                 else fail "Offside wreckage!"
               reparse e inp
-}

-------------
-- This parsers are monads
-------------
instance Monad (Parser t) where
   return = returnP
   (>>=)  = bindP
   fail s = errorP s 
   
instance MonadPlus (Parser t) where
   mzero = failP
   mplus = orelse

instance Functor (Parser t) where
   fmap = mapP

mapP :: (a -> b) -> Parser t a -> Parser t b
mapP f m = m `using` f

binP :: (a -> b -> c) -> Parser t a -> Parser t b -> Parser t c
binP = binM

apP :: Parser t (a->b) -> Parser t a -> Parser t b              
apP = ap

doP :: Parser t a -> Parser t b -> Parser t a
doP = doM

pairP :: Parser t a -> Parser t b -> Parser t (a,b)
pairP = pairM

consP :: Parser t a -> Parser t [a] -> Parser t [a]
consP = consM

instance MonadOption (Parser t) where
  option = opt

  
