module Date where

import Char

-- **********************************************************************
--  Date datatype
-- **********************************************************************

data Date = Date { year :: Int, month :: Int, day :: Int }
          deriving (Eq, Read)

{-
makeDate      :: Int -> Int -> Int -> Date
              -- year   month  day
validDate     :: Date -> Bool
addDays       :: Int -> Date -> Date
addMonths     :: Int -> Date -> Date
addYears      :: Int -> Date -> Date
addFixed      :: Int -> Date -> Date
startMonth    :: Date -> Date
weekDay       :: Date -> String
monthYear     :: Date -> String
monthDay      :: Date -> Int    -> Int ->  (Int, Date)
              --         day 0..6  week 0.. day
calendarDay   :: Date -> (Int,      Int)
              --          day 0..6, week 0..
calendarMonth :: Date -> [(Int,      Int)]
              --           day 0..6, week 0..
-}

-- *********************************************************************
-- Date datatype
-- *********************************************************************

instance Show Date where
  showsPrec _ (Date year month day) = 
      shows day . showString "-" . shows month . showString "-" . shows year

showDate format date = 
   let (s, field, t) = nextField format
    in case field of
         Nothing -> s ++ t
         Just fnw -> s ++ formatField fnw date ++ showDate t date

formatField ('#', n, w) date = let v = extractFieldAsNumber w date
                                in fixedLen '0' (show v) n
formatField ('$', n, w) date = let v = extractFieldAsString w date
                                in fixedLen ' ' v n

extractFieldAsNumber 'd' = day
extractFieldAsNumber 'm' = month
extractFieldAsNumber 'y' = year

extractFieldAsString 'd' = (\d -> days!!(d-1) ++ ", " ++ show d) . day
extractFieldAsString 'm' = (\m -> months!!(m-1)) . month
extractFieldAsString 'y' = show . year

fixedLen c xs 0 = xs
fixedLen c xs n = let m = length xs
                   in case (compare n m) of
                        LT -> take n xs
                        EQ -> xs
                        GT -> (replicate (n-m) c) ++ xs

nextField df = (s1, field, s3)            
      where (s1, s2) = break (\c -> c == '#' || c == '$') df
            (field,s3) = 
               case s2 of
                ""           -> (Nothing,"")
                (f:disc:s2') -> if (isDigit disc) 
                                 then (Just (f,n,head s2'), tail s2') 
                                 else (Just (f,0,disc), s2')
                                 where n = ord disc - ord '0' 

dateFormat = "#2d-#2m-#4y"

{-
instance Read Date where
  readsPrec _ s = Date { year=y, month=m, day=d }
  where
    nosep x = x <> '-' && x <> '/'
    (s1, s2) = span nosep (read s)
    d = (toInt . show) s1
    (s3, s4) = (span nosep . drop 1) s2
    m = (toInt . show) s3
    y = (toInt . show . drop 1) s4
-}
instance Ord Date where
  (Date y1 m1 d1) < (Date y2 m2 d2) = 
     y1 < y2 || (y1 == y2 && (m1 < m2 || (m1 == m2 && d1 < d2)))

makeDate :: Int -> Int -> Int -> Date
makeDate y m d = Date { year=y, month=m, day=d }

validDate :: Date -> Bool
validDate (Date year month day) = 
    year > 0  && 
    month >=1 && month <= 12 && 
    day >= 1  && day <= monthDays year month

addDays :: Int -> Date -> Date
addDays n (Date year month day) = 
   normaliseDate (Date { year=year, month=month, day=day+n })

addMonths :: Int -> Date -> Date
addMonths n (Date year month day) = 
   normaliseDate (Date { year=year, month=month+n, day=day })

addYears :: Int -> Date -> Date
addYears n (Date year month day) = 
   normaliseDate (Date { year=year+n, month=month, day=day })

addFixed :: Int -> Date -> Date
addFixed n date1@(Date _ _ day) = 
 addDays (wn*7 + (dn1 - dn2 + 7) `mod` 7) date2
 where
  date2 = (startMonth . addMonths n) date1
  dn1 = dayNr date1
  dn2 = dayNr date2
  wn = (day-1) `div` 7

startMonth :: Date -> Date
startMonth (Date year month _) = Date { year=year, month=month, day=1 }

weekDay :: Date -> String
weekDay date = days !! (dayNr date)

monthYear :: Date -> String
monthYear (Date year month _) = 
   months !! (month - 1) ++ " " ++ show year

monthDay :: Date -> Int -> Int -> (Int, Date)
monthDay date@(Date year month _) d w = 
  (day'', Date { year=year, month=month, day=day''})
 where
  dnr   = (dayNr . startMonth) date
  day'  = d + w*7 - dnr + 1
  day'' = if (day' >= 1 && day' <= monthDays year month) 
           then day' 
           else 0

calendarDay :: Date -> (Int, Int)
calendarDay date@(Date _ _ day) = 
 ((dnr+day-1) `mod` 7, (dnr+day-1) `div` 7)
 where
  dnr = (dayNr . startMonth) date

calendarMonth :: Date -> [(Int, Int)]
calendarMonth date@(Date year month day) = 
  [((dnr+d) `mod` 7, (dnr+d) `div` 7) | d <- [0..monthDays year (month-1)]]
 where
  dnr = (dayNr . startMonth) date

days = ["sunday",   "monday", "tuesday", "wednesday",
        "thursday", "friday", "saturday"]

months = ["january",   "february", "march",    "april", 
          "may",       "june",     "july",     "august",
          "september", "october",  "november", "december"]

normaliseDate (Date year month day)
  | month < 1
    = normaliseDate (Date { year=year-1, month=month+12, day=day })
  | month > 12
    = normaliseDate (Date { year=year+1, month=month-12, day=day })
  | day < 1
    = normaliseDate (Date { year=year,
                            month=month-1, 
                            day=day+monthDays year (month-1) 
                          })
  | day > daysMonth
    = normaliseDate (Date { year=year, month=month+1, day=day-daysMonth })
  | otherwise
    = Date { year=year, month=month, day=day }
 where
   daysMonth = monthDays year month

leapYear y
  | y `mod` 100 == 0 = y `mod` 400 == 0
  | otherwise        = y `mod` 4 == 0

monthLengths y = [31, if (leapYear y) then 29 else 28, 
                  31, 30, 
                  31, 30, 
                  31, 31, 
                  30, 31, 
                  30, 31]

monthDays y m = monthLengths y !! ((m-1+1200) `mod` 12)

dayNr (Date year month day) = 
  (y*365 + y `div` 4 - y `div` 100 + y `div` 400 + 
   sum (take (month-1) (monthLengths year)) + day) `mod` 7
 where
   y = year - 1
