-- Lexer library in Haskell
--
-- This module was (mainly) part of Parselib3, by John Hughes
--
--  By John Hughes (1997)
--  Additions and modifications: Pablo E. Martinez Lopez (1999)

module Lexer 
 ( Token(..)   -- A type to classify tokens
 , Tokenized, TokenList -- Type synonyms
 , string      -- Given a list of tokens, succeeds iff that 
               --   list is the initial segment
 , word        -- Parses a sequence of letters and digits beginning 
               --   with a letter
 , number      -- Parses a sequence of digits
 , as          -- Invokes the argument parser, and changes the result 
               --   adding symbol information
 , strip       -- Filters all the tokens of a given kind
 , kind        -- Succeeds iff the first token is of a given kind
 , lit         -- Succeeds iff the first token is a given symbol of 
               --   kind Symbol (tokenized input)
 , anyLitOf    -- Succeeds iff the first token is one of the given symbols
 , key         -- Strips the Symbol off the lit
 , skipto      -- Parses until a given parser succeeds
 , lexer       -- A simple lexical analyser for programming languages
 )
 where

import Char
import Position
import BasicParser
import Error

infix  4 `as`

type Tokenized a = (a,Token)
type TokenList s = PosList (Tokenized s)

-- Combinators for lexical analysis
string :: (Eq t, Show t) => [t] -> Parser t [t]
string s = foldr (<:>) (return []) (map literal s)

word, number :: Parser Char String
word = satisfy isAlpha <:> many (satisfy isIdChar)
number = some (satisfy isDigit)

isIdChar c = isAlphaNum c || c `elem` "_'"

-- Combinators for a separate lexical analyser
as :: Parser t a -> sym -> Parser t (Position, (a,sym))
parser `as` t = binM (\p v -> (p, (v,t))) position parser
              
data Token = Symbol | Ident | Numeral | StrConst    | WhiteSpace | Other String
             deriving (Eq,Show)

strip :: Eq tok => tok -> [(pos, (a, tok))] -> [(pos, (a, tok))]
strip t' = filter (\(_,(_,t)) -> t/=t')

kind :: (Eq tok, Show tok) => tok -> Parser (a,tok) a
kind t = do (v,t') <- token `withError` (show t ++ " expected!")
            if t==t' 
             then return v 
             else fail (show t ++ " expected!")
             
lit :: (Eq t, Show t) => t -> Parser (t,Token) (t,Token)        
lit s = literal (s,Symbol)

anyLitOf ls = anyOf lit ls `orelse` fail ("Any of " ++ show ls ++ " expected!")

key :: (Eq t, Show t) => t -> Parser (t,Token) t        
key s = return fst $> lit s

skipto :: Parser t a -> Parser t a
skipto p = p `orelse` (return id <* token $> skipto p)

lexer :: [String]           -> [String]                 -> 
         Parser Char String -> (String -> Maybe String)
          -> String -> Error (TokenList String)
lexer symbols keywords comment procIdents = 
   fmap (strip WhiteSpace)                                      .  
   parse (many (lextoken symbols keywords comment procIdents)) . 
   addPositions

lextoken :: [String]           -> [String]                -> 
            Parser Char String -> (String -> Maybe String)
             -> Parser Char (Position,(String,Token))
lextoken symbols keywords comment procIdents = 
  do pos <- position
     (_,t) <- auxlextoken symbols keywords comment procIdents 
     return (pos,t)

auxlextoken symbols keywords comment procIdents = 
       number `as` Numeral
        `orelse`
       (some (satisfy isSpace)) `as` WhiteSpace
        `orelse`
       comment `as` WhiteSpace
        `orelse`
       (anyOf string symbols) `as` Symbol
        `orelse`
       (word >>= \w->
        if (w `elem` keywords) 
         then return w `as` Symbol
         else case procIdents w of
               Just s  -> return w `as` (Other s)
               Nothing -> return w `as` Ident
       )
        `orelse`
       (return id 
         <* literal '"' 
         $> many (satisfy (/='"')) <* literal '"') `as` StrConst


