-- Monad library in Haskell
--
-- This module was (mainly) part of Parselib3, by John Hughes
--
--  By John Hughes (1997)
--  Additions and modifications: Pablo E. Martinez Lopez (1999)

module MoreMonad 
 ( module Monad
 , using, (<@)                   -- the flipped version of fmap
 , binM                          -- lifting of a given binary operation
 , anyOf                         -- a multiple orelse
 , ($>)                          -- lifting of functional application to monad M
 , doM, (<*), (*>), pairM, (<*>) -- pairing and selection of one component
 , consM, (<:>)                  -- lifting of list operations to monad M
 , replicateM                    -- lifting of replicate
 , MonadOption(..)               -- option
 , many, some                    -- repetition
 , lassoc, strictlassoc
 , rassoc, noassoc               -- For binary operations
 , sepBy                         -- for lists
 )
 where

import Monad
import MyPrelude(strictfoldl)

-------------
-- Operations useful in any monad
-------------

infixl 8 $>, <*, `doM`, *>
infixr 6 <:>, `consM`, <*>, `pairM`

anyOf :: MonadPlus m => (a -> m b) -> [a] -> m b
anyOf f = foldr mplus mzero . map f

using, (<@) :: Monad m => m a -> (a -> b) -> m b
using m f = liftM f m
(<@) = using

binM :: Monad m => (a -> b -> c) -> m a -> m b -> m c
binM = liftM2

($>) :: Monad m => m (a->b) -> m a -> m b              
($>) = ap

doM, (<*) :: Monad m => m a -> m b -> m a
doM = binM const
(<*) = doM

(*>) :: Monad m => m a -> m b -> m b
(*>) = (>>)

pairM, (<*>) :: Monad m => m a -> m b -> m (a,b)
pairM = binM (\x y -> (x,y))
(<*>) = pairM

replicateM 0 m = return []
replicateM n m = do a  <- m
                    as <- replicateM (n-1) m
                    return (a:as)

consM, (<:>) :: Monad m => m a -> m [a] -> m [a]
consM = binM (:)
(<:>) = consM

-------------
-- A new kind of alternative, and some functions depending on it
-------------
class MonadPlus m => MonadOption m where
  option :: m a -> a -> m a 
  
  option m a = m `mplus` return a
  
many, some :: MonadOption m => m a -> m [a]  
many p = some p `option` []
some p = p <:> many p

-- This parser is not recursive due to left-recursion
lassoc, strictlassoc :: MonadOption m => m (a->a->a) -> m a -> m a 
lassoc op e = do v <- e
                 opws <- many (op <*> e)
                 return (foldl (\u (oper,w)->u `oper` w) v opws)
strictlassoc op e = do v <- e
                       opws <- many (op <*> e)
                       return (strictfoldl (\u (oper,w) -> u `oper` w) v opws)

                 
rassoc :: MonadOption m => m (a->a->a) -> m a -> m a 
rassoc op e = do v <- e
                 (op $> return v $> rassoc op e) `option` v
              
noassoc :: MonadOption m => m (a->a->a) -> m a -> m a 
noassoc op e = do v <- e
                  (op $> return v $> e) `option` v

sepBy :: MonadOption m => m a -> m b -> m [a]               
p `sepBy` s = (consM p (many (return id <* s $> p))) `option` []

  
