-- Positions library in Haskell
--
-- This module was (mainly) part of Parselib3, by John Hughes
--
--  By John Hughes (1997)
--  Additions and modifications: Pablo E. Martinez Lopez (1999)

-- I want to consider the way Niklas Rojemo adds positions (toghether with the lexer...)

module Position
  (Position(..)     -- Type to indicate the positions in a file
  , PosList         -- List of elements with positions
  , addPositions    -- Adds positions to a string
  , displayPos      -- Displays the positions in format suitable for reporting errors
  , column          -- Extract the column of a given position
  , pos             -- Extract the position of the next token in a PosList
  )
  where

-------------
-- Positions
-------------
data Position = NoPos                 -- there is no information
              | Pos Int Int String    -- indicates a position within a file
              | EndofFile             -- indicates the end of the file
              deriving (Eq,Ord,Show)

type PosList t = [(Position, t)]
        
-- Adds positional information to every character in a string
--   (behaves lazily, so the addition is made only if neccesary)
addPositions :: String -> PosList Char
addPositions xs = [(Pos i j l,c) 
                  | (l,i) <- zip (lines xs) [1..],
                    let l' = tabs 0 l,
                    (c,j) <- zip (l'++"\n") [1..length l'+1]
                  ]
                  
   -- take the opportunity to force l
tabs k [] = []
tabs k ('\t':cs) | k`mod`8==7 = ' ':tabs(k+1) cs
                 | otherwise  = ' ':tabs(k+1) ('\t':cs)
tabs k (c:cs) = c:tabs(k+1)cs
        
-- Displays a position in human readable form
displayPos :: Position -> String
displayPos EndofFile = "end of file\n"
displayPos (Pos i j l) =
    "line "++show i++":\n"++                                   -- indicates the line;
    l++"\n"++                                                  -- shows the line;
    [if c=='\t' then '\t' else ' ' | c<-take (j-1) l]++"^\n"   -- mark the column with ^.
displayPos NoPos = "unknown position\n"

-- Projects the column in a given position (NoPos has no column)
column :: Position -> Int
column (Pos _ j _) = j
column EndofFile = 0

-- Extracts the current position of a token-list
pos :: PosList t -> Position
pos [] = EndofFile
pos ((p,_):_) = p

