module PredsSolving(solve,solveRT) where

import List(union, sort)

import MoreMonad

import ResidualLang
import SpecMonad
import ResidualLangFVs
import ResidualReduction

import PredsSimplify
import SolvingFailure
import SolvingFunS
import SolvingMG
--------------------------
---- Begin Alejandro -----
--------------------------
import SolvingDData
--------------------------
----- End Alejandro -----
--------------------------

------------------------------------------------------------
-- Solves a context by instantiating variables that have
--  no further chance of being instantiated by unification.
------------------------------------------------------------
solve :: Bool -> [ResTypeVble] -> [ResSchemeVble]
          -> Preds -> SpecM (ContextTerm, Preds)
solve forcing ftvs fsvs delta = 
  do yes <- solveContexts
     if yes
      then solveRT forcing ftvs fsvs delta
      else theEnd forcing ftvs fsvs delta

theEnd _ _ _ delta = return (CHole, delta)

------
-- Executes the next level of the solving algorithm,
--  and calculates the compound resulting context
--
-- OBS: this context combination prevents tail
--      recursion.
--  It would be better shifted to an upper level...
--   (that is, accumulate all (hs,vs,hole) and then construct)
------
nextLevel loop forcing ftvs fsvs delta (hs,vs) hole =
   do (cxt', delta'')  <- loop forcing ftvs fsvs delta
      xv               <- freshresidualtermBasedOn "xv"
      let rescxt = CELet xv (CELam hs hole)
                            (noredfill cxt' (REApp (RVar xv) vs))
      return (rescxt, delta'')

------------------------------------------------------------
-- Picks a type variable, and solves it, if it can.
------------------------------------------------------------
solveRT forcing ftvs fsvs delta =
  do delta'       <- unfoldAll delta
     (gs, ws, os) <- clFreeRTV True delta'
     mt           <- pickRTVar ftvs ws delta'
     case mt of
      Just (t, d) -> 
       do traceSM ("Go: "    ++ show (sort gs)) ()
          traceSM ("Wait: "  ++ show (sort ws)) ()
          traceSM ("Other: " ++ show (sort os)) ()
          traceSM "RTVar to solve:" t
          traceSM ("Cxt(summary):" ++ show (takeEvidence delta')) ()
          traceSM "Context:" delta'
          (hs,vs,delta'') <-
            case d of
              IsFunS _ _ -> solveFunS forcing ftvs fsvs t delta'
              IsFixS _ _ -> solveFixS forcing ftvs fsvs t delta'
              ConstrsOf _ -> solveDData forcing ftvs fsvs t delta'
              GuardedDinData _ _ _ -> solveDData forcing ftvs fsvs t delta'
              HasC _ _ _ -> solveDData forcing ftvs fsvs t delta'
              _          -> noSolution forcing msgAmbiguous [(EVar 0,d)] delta'
          nextLevel solveRT forcing ftvs fsvs delta'' (hs,vs) CHole
      Nothing -> solveNRRS forcing ftvs fsvs delta'

------------------------------------------------------------
-- Picks a (nonrecursive) scheme variable, and solves it.
------------------------------------------------------------
solveNRRS forcing ftvs fsvs delta = 
  do delta'     <- return delta --unfoldAll delta
     (gs,ws,os) <- clFreeRSV True delta'
     ms         <- pickNRRSVar fsvs ws delta'
     case ms of
      Just s  -> do traceSM ("SV Go: "    ++ show (sort gs)) ()
                    traceSM ("SV Wait: "  ++ show (sort ws)) ()
                    traceSM ("SV Other: " ++ show (sort os)) ()
                    traceSM "RSVar to solve:" s
                    traceSM ("Cxt(summary):" ++ show (takeEvidence delta')) ()
                    traceSM "Context:" delta'
                    (hs, vs, delta'',hole) <- solveMG forcing ftvs fsvs s delta'
                    nextLevel solveRT forcing ftvs fsvs delta'' (hs,vs) hole
      Nothing -> solveRRS forcing ftvs fsvs delta'

------------------------------------------------------------
-- Picks a (recursive) scheme variable, and solves it.
------------------------------------------------------------
solveRRS forcing ftvs fsvs delta = 
  do (hs,vs,delta')  <- simplifySolving ftvs fsvs delta
     when forcing (checkFailure delta')
     nextLevel theEnd forcing ftvs fsvs delta' (hs,vs) CHole
{-
  do delta'     <- unfoldAll delta
     (gs,ws,os) <- clFreeRSV True delta'
     ms <- pickRRSVar fsvs ws delta
     case ms of
      Just s  -> 
       do traceSM ("SV Go: "    ++ show (sort gs)) ()
          traceSM ("SV Wait: "  ++ show (sort ws)) ()
          traceSM ("SV Other: " ++ show (sort os)) ()
          traceSM "RSVar to solve:" s
          traceSM "Context:" delta
-}
     
------
-- Picks a type variable that does not appear in the type or context
------
pickRTVar _ _ [] = return Nothing
pickRTVar fvs ncftvs ((_,d@(IsInt tau)):delta) = 
  do tau' <- indir tau
     case tau' of
      RTVar t -> if not (t `elem` fvs) && not (t `elem` ncftvs)
                  then return (Just (t, d))
                  else pickRTVar fvs ncftvs delta
      _       -> pickRTVar fvs ncftvs delta
pickRTVar fvs ncftvs ((_,d@(IsBool tau)):delta) = 
  do tau' <- indir tau
     case tau' of
      RTVar t -> if not (t `elem` fvs) && not (t `elem` ncftvs)
                  then return (Just (t, d))
                  else pickRTVar fvs ncftvs delta
      _       -> pickRTVar fvs ncftvs delta
pickRTVar fvs ncftvs ((_,d@(IsFunS ub tau)):delta) = 
  do tau' <- indir tau
     case tau' of
      RTVar t -> 
       do ub' <- unfoldAll ub
          case ub of
           (RTTypedS _ (RTStaticExp _) _) 
             -> if not (t `elem` fvs)  -- CHEAT!! This is because static
                                       --  functions are unlikely to be
                                       --  constrained meaningfully more 
                                       --  than once... (is that true?)
                 then return (Just (t, d))
                 else pickRTVar fvs ncftvs delta
           _ -> if not (t `elem` fvs) && not (t `elem` ncftvs)
                 then return (Just (t, d))
                 else pickRTVar fvs ncftvs delta
      _       -> pickRTVar fvs ncftvs delta
--pickRTVar fvs ncftvs ((_,d@(IsFunS tau _)):delta) = 
--  do tau' <- indir tau
--     case tau' of
--      RTVar t -> if not (t `elem` fvs) && not (t `elem` ncftvs)
--                  then return (Just (t, d))
--                  else pickRTVar fvs ncftvs delta
--      _       -> pickRTVar fvs ncftvs delta
pickRTVar fvs ncftvs ((_,d@(IsFixS tauU tau)):delta) = 
  do tau' <- indir tau
     case tau' of
      RTVar t -> if not (t `elem` fvs) && not (t `elem` ncftvs)
                  then do tauU' <- indir tauU
                          case tauU' of
                            RTVar _ -> pickRTVar fvs ncftvs delta
                            _       -> return (Just (t, d))
                  else pickRTVar fvs ncftvs delta
      _       -> pickRTVar fvs ncftvs delta

-------------------------------
--- Begin Alejandro -----------
-------------------------------
pickRTVar fvs ncftvs ((_,d@(HasC td _ tau)):delta) = 
  do tau' <- indir td 
     case tau' of
      RTVar t -> if not (t `elem` fvs) && not (t `elem` ncftvs)
                  then return (Just (t, d))
                  else pickRTVar fvs ncftvs delta
      _       -> pickRTVar fvs ncftvs delta


pickRTVar fvs ncftvs ((_,d@(ConstrsOf td)):delta) = 
  do tau' <- indir td 
     case tau' of
      RTVar t -> if not (t `elem` fvs) && not (t `elem` ncftvs)
                  then return (Just (t, d))
                  else pickRTVar fvs ncftvs delta
      _       -> pickRTVar fvs ncftvs delta


pickRTVar fvs ncftvs ((_,d@(GuardedDinData _ td _)):delta) = 
  do tau' <- indir td 
     case tau' of
      RTVar t -> if not (t `elem` fvs) && not (t `elem` ncftvs)
                  then return (Just (t, d))
                  else pickRTVar fvs ncftvs delta
      _       -> pickRTVar fvs ncftvs delta

-------------------------------
----- End Alejandro -----------
-------------------------------

pickRTVar fvs ncftvs (_:delta) = pickRTVar fvs ncftvs delta

------
-- Picks a scheme variable that does not appear in the type or context
--  and that cannot be further constrained.
------
pickNRRSVar _ _ [] = return Nothing
pickNRRSVar fvs nbfvs ((_,IsMG (RSForAll _ _ _) sigma):delta) = 
  do sigma' <- indir sigma
     case sigma' of
      RSVar s -> if not (s `elem` fvs) -- && not (s `elem` nbfvs)
                                       -- WARNING!! Causes non correct specs
                  then return (Just s)
                  else pickNRRSVar fvs nbfvs delta
      _      -> pickNRRSVar fvs nbfvs delta
pickNRRSVar fvs nbfvs ((_,IsMGmorph _ (RSForAll _ _ _) sigma):delta) = 
  do sigma' <- indir sigma
     case sigma' of
      RSVar s -> if not (s `elem` fvs) && not (s `elem` nbfvs)
                  then return (Just s)
                  else pickNRRSVar fvs nbfvs delta
      _      -> pickNRRSVar fvs nbfvs delta
pickNRRSVar fvs nbfvs (_:delta) = pickNRRSVar fvs nbfvs delta

