module ResidualGeneralizable
  (generalize, regularize, instantiate, addOffset 
  ,Generalizable(..))
 where

import List (union)
import MoreMonad(liftM, liftM2, liftM3, liftM4)
import MyPrelude (unite)

import ResidualLang
import SpecMonad

---------------------------------------------------------
-- Quantifies the variables provided in the lists
---------------------------------------------------------
generalize :: [ResTypeVble] -> [ResSchemeVble] 
               -> QualifiedType -> SpecM ResidualTypeScheme
generalize ts ss qt = 
   do qt' <- gener ts ss qt
      let lts = length ts
      let lss = length ss
      return (RSForAll lts lss qt')

---------------------------------------------------------
-- Instantiates quantified variables to fresh ones, returning those.
---------------------------------------------------------
instantiate :: ResidualTypeScheme -> SpecM (QualifiedType, [ResTypeVble]
                                                         , [ResSchemeVble])
instantiate (RSForAll 0 0 qt) = return (qt, [], [])
instantiate (RSForAll _ _ qt) = inst qt
instantiate sigma = fail "Cannot instantiate a variable or a tuple!"

---------------------------------------------------------
-- Renames quantified variables to avoid clashes.
---------------------------------------------------------
addOffset :: ResTypeVble -> ResSchemeVble 
             -> ResidualTypeScheme -> SpecM ResidualTypeScheme
addOffset tm sm (RSForAll mt ms qt) = do qt' <- offset tm sm qt
                                         return (RSForAll (mt+tm) (ms+sm) qt')

---------------------------------------------------------
-- Transforms the GenVbles of a term to get them
--  in a 'regular' order
---------------------------------------------------------
regularize :: ResidualTypeScheme -> SpecM ResidualTypeScheme
regularize (RSForAll _ _ qt) = do qt' <- regular qt
                                  tm' <- maxRTGen
                                  sm' <- maxRSGen
                                  return (RSForAll tm' sm' qt')
regularize (RSTupleS ts)     = liftM RSTupleS (mapM regular ts)
regularize sigma             = do sigma' <- indir sigma
                                  case sigma' of
                                    RSVar _ -> return sigma'
                                    _       -> regularize sigma'

-- ---------------------------------------------
type PlusVbles t = (t,[ResTypeVble],[ResSchemeVble])
--type PlusVbles t = (t,[ResGenTypeVble],[ResGenSchemeVble])

-- ---------------------------------------------
-- Generalization for different elements
-- ---------------------------------------------
gener    :: Generalizable t => [ResTypeVble] -> [ResSchemeVble] -> t -> SpecM t 
gener   ts ss t = do resetGeneric; gener' ts ss  t

gener'   :: Generalizable t => [ResTypeVble] -> [ResSchemeVble] -> t -> SpecM t 
gener' ts ss  t = do (t',_,_) <- gr (\tv -> if (tv `elem` ts)
                                             then do tg <- aRTGenFor tv
                                                     return (tg,[],[])
                                             else return (RTVar tv,[],[])) 
                                    (\ot -> return (ot,[],[]))
                                    (\sv -> if (sv `elem` ss) 
                                             then do sg <- aRSGenFor sv
                                                     return (sg,[],[])
                                             else return (RSVar sv,[],[])) 
                                    (\os -> return (os,[],[]))
                                    t
                     return t'

-- ---------------------------------------------
-- Regularization for different elements
-- ---------------------------------------------
regular  :: Generalizable t => t -> SpecM t
regular       t = do resetGeneric; regular'      t

regular' :: Generalizable t => t -> SpecM t
regular'      t = do (t',_,_) <- gr (\tv -> return (RTVar tv,[],[]))
                                    (\t@(RTGen l g) -> 
                                        do cl <- scopeLevel
                                           if cl/=l 
                                            then return (t,[],[])
                                            else do tg <- aRTGenFor g
                                                    return (tg,[],[])
                                    ) 
                                    (\sv -> return (RSVar sv,[],[]))
                                    (\s@(RSGen l g) ->  
                                        do cl <- scopeLevel
                                           if cl/=l 
                                            then return (s,[],[])
                                            else do sg <- aRSGenFor g
                                                    return (sg,[],[])
                                    )
                                    t
                     return t'

-- ------------------------------------
--  Eliminating one level of quantification
-- ------------------------------------
inst :: Generalizable t => t -> SpecM (PlusVbles t)
inst t = do resetGeneric; inst' t

inst' :: Generalizable t => t -> SpecM (PlusVbles t)
inst' t = gr (\tv -> return (RTVar tv, [], []))
             (\t@(RTGen l g) -> 
                 do cl <- scopeLevel
                    case compare l cl of
                      EQ -> do nv@(RTVar v) <- aRTVarFor g
                               return (nv,[v],[])
                      GT -> let l1 = l-1 in l1 `seq`
                            return (RTGen l1 g,[],[])
                      LT -> return (t,[],[])
             )
             (\sv -> return (RSVar sv, [], []))
             (\s@(RSGen l g) ->
                 do cl <- scopeLevel
                    case compare l cl of
                     EQ -> do nv@(RSVar v) <- aRSVarFor g
                              return (nv, [], [v])
                     GT -> let l1 = l-1 in l1 `seq` 
                           return (RSGen l1 g, [], [])
                     LT -> return (s, [], [])
             )
             t

-- ---------------------------------------------
-- Renaming of bound variables to avoid clashes
-- ---------------------------------------------
offset  :: Generalizable t => ResTypeVble  ->  ResSchemeVble  -> t -> SpecM t
offset  tm sm t = do resetGeneric; offset' tm sm t

offset' :: Generalizable t => ResTypeVble  ->  ResSchemeVble  -> t -> SpecM t
offset' tm sm t =
  do (t',_,_) <- gr (\tv -> return (RTVar tv, [], []))
                    (\t@(RTGen l g) -> 
                        do cl <- scopeLevel
                           if cl/=l then return (t,[],[])
                                    else let gtm = g+tm in gtm `seq` 
                                         return (RTGen l gtm,[],[])
                    )
                    (\sv -> return (RSVar sv, [], []))
                    (\s@(RSGen l g) ->
                         do cl <- scopeLevel
                            if cl/=l then return (s,[],[])
                                     else let gsm = g+sm in gsm `seq` 
                                          return (RSGen l gsm,[],[])
                    )
                    t
     return t'

-- ---------------------------------------------
-- Generalization of type and scheme variables
-- ---------------------------------------------
class Generalizable t where
  gr :: (ResTypeVble   -> SpecM (PlusVbles ResType))       -> -- RTVar
        (ResType       -> SpecM (PlusVbles ResType))       -> -- RTGen
        (ResSchemeVble -> SpecM (PlusVbles ResTypeScheme)) -> -- RSVar
        (ResTypeScheme -> SpecM (PlusVbles ResTypeScheme)) -> -- RSGen
        t -> SpecM (PlusVbles t)

  gr _ _ _ _    t = return (t,[],[])

instance Generalizable Int
instance Generalizable Char

-- -------------------------------------------------------------
-- Generalization for residual types
instance Generalizable ResType where
  gr _  tg _  _  t@(RTGen _ _)     = tg t
  gr tf tg sf sg t@(RTVar _)       = do t' <- indir t
                                        case t' of
                                          RTVar tv    -> tf tv
                                          _           -> gr tf tg sf sg t'
  gr tf tg sf sg (RTTuple  ts)     = do (ts',tv,sv) <- gr tf tg sf sg ts 
                                        return (RTTuple ts',tv,sv)
  gr tf tg sf sg (RTTupleS ts)     = do (ts',tv,sv) <- gr tf tg sf sg ts
                                        return (RTTupleS ts',tv,sv)
  gr tf tg sf sg (RTFun t1 t2)     = do (t1',tv1,sv1) <- gr tf tg sf sg t1 
                                        (t2',tv2,sv2) <- gr tf tg sf sg t2
                                        return (RTFun t1' t2' 
                                               ,tv1`union`tv2
                                               ,sv1`union`sv2)
--  gr tf tg sf sg (RTStaticExp e)   = do (e',tv,sv) <- gr tf tg sf sg e
--                                        return (RTStaticExp e', tv, sv)
  gr tf tg sf sg (RTFixS ts t s)   = do (ts',tvts,svts) <- gr tf tg sf sg ts
                                        (t',tv1,sv1) <- gr tf tg sf sg t 
                                        (s',tv2,sv2) <- gr tf tg sf sg s
                                        return (RTFixS ts' t' s' 
                                               ,tv1`union`tv2`union`tvts
                                               ,sv1`union`sv2`union`svts)
  gr tf tg sf sg (RTTypedS ts t s) = do (ts',tvts,svts) <- gr tf tg sf sg ts
                                        (t',tv1,sv1) <- gr tf tg sf sg t 
                                        (s',tv2,sv2) <- gr tf tg sf sg s
                                        return (RTTypedS ts' t' s' 
                                               ,tv1`union`tv2`union`tvts
                                               ,sv1`union`sv2`union`svts)
  gr tf tg sf sg (RTPoly s)        = do (s',tv,sv) <- gr tf tg sf sg s
                                        return (RTPoly s', tv, sv)
  ----------------------------
  ------- Begin Alejandro ----
  ----------------------------
  gr tf tg sf sg (RTDData t ts)  = do (t',tv,sv) <- gr tf tg sf sg t
                                      (ts',tvs',svs') <- gr tf tg sf sg ts
                                      return (RTDData t' ts', 
                                              tv `union` tvs', 
                                              sv `union` svs')
  ----------------------------
  ------- End Alejandro ------
  ----------------------------
  gr tf tg sf sg (RTData d ts ss)  = do (ts',tv1,sv1) <- gr tf tg sf sg ts 
                                        (ss',tv2,sv2) <- gr tf tg sf sg ss 
                                        return (RTData d ts' ss'           
                                               ,tv1`union`tv2              
                                               ,sv1`union`sv2)             
  gr tf tg sf sg (RTConstruc c ts) = do (ts',tv,sv) <- gr tf tg sf sg ts   
                                        return (RTConstruc c ts',tv,sv)    
  gr _  _  _  _  t                 =    return (t,[],[])

-- -------------------------------------------------------------
-- Generalization for residual types schemes
instance Generalizable ResTypeScheme where
  gr _  _  _  sg s@(RSGen _ _)       = sg s
  gr tf tg sf sg s@(RSVar _)         = do s' <- indir s
                                          case s' of
                                           RSVar os -> sf os
                                           _        -> gr tf tg sf sg s'
  gr tf tg sf sg (RSForAll mt ms qt) = do enterQuantifierScope
                                          (qt',tv,sv) <- gr tf tg sf sg qt 
                                          leaveQuantifierScope
                                          return (RSForAll mt ms qt',tv,sv)
  gr tf tg sf sg (RSTupleS ss)       = do (ss',tv,sv) <- gr tf tg sf sg ss 
                                          return (RSTupleS ss', tv, sv)
  gr _  _  _  _  s                   = return (s,[],[])
 
-- -------------------------------------------------------------
-- Generalization for qualified types
instance Generalizable QualType where
  gr tf tg sf sg (ps :=> t) = do (t' ,tvt,svt) <- gr tf tg sf sg t
                                 (ps',tvp,svp) <- gr tf tg sf sg ps
                                 return (ps' :=> t'
                                        ,tvt`union`tvp
                                        ,svt`union`svp)

--------
instance Generalizable a => Generalizable [a] where
  gr _  _  _  _  []     = return ([],[],[])
  gr tf tg sf sg (x:xs) = do (x' ,tvx ,svx)  <- gr tf tg sf sg x
                             (xs',tvxs,svxs) <- gr tf tg sf sg xs
                             return (x':xs'
                                    ,tvx `union` tvxs
                                    ,svx `union` svxs)

--------
instance (Generalizable a, Generalizable b) => Generalizable (a,b) where
  gr tf tg sf sg (a,b) = do (a',tva,sva) <- gr tf tg sf sg a
                            (b',tvb,svb) <- gr tf tg sf sg b
                            return ((a',b')
                                   ,tva `union` tvb
                                   ,sva `union` svb)

-- -------------------------------------------------------------
-- Generalization for predicates
instance Generalizable Pred where
  gr tf tg sf sg (IsMG s1 s2)           = do (s1',tv1,sv1) <- gr tf tg sf sg s1 
                                             (s2',tv2,sv2) <- gr tf tg sf sg s2
                                             return (IsMG s1' s2'
                                                    ,tv1 `union` tv2
                                                    ,sv1 `union` sv2)
  gr tf tg sf sg (IsMGmorph t s1 s2)    = do (t' ,tvt,svt) <- gr tf tg sf sg t
                                             (s1',tv1,sv1) <- gr tf tg sf sg s1 
                                             (s2',tv2,sv2) <- gr tf tg sf sg s2
                                             return (IsMGmorph t' s1' s2'
                                                    ,tv1 `union` tv2 `union` tvt
                                                    ,sv1 `union` sv2 `union` svt)
  gr tf tg sf sg (IsInt t)              = do (t' ,tv ,sv)  <- gr tf tg sf sg t
                                             return (IsInt  t', tv, sv)
  gr tf tg sf sg (IsBool t)             = do (t' ,tv ,sv)  <- gr tf tg sf sg t
                                             return (IsBool t', tv, sv)
  gr tf tg sf sg (IsFunS t1 t2)         = do (t1',tv1,sv1) <- gr tf tg sf sg t1
                                             (t2',tv2,sv2) <- gr tf tg sf sg t2 
                                             return (IsFunS t1' t2'
                                                    ,tv1 `union` tv2
                                                    ,sv1 `union` sv2)
  gr tf tg sf sg (IsFixS t1 t2)         = do (t1',tv1,sv1) <- gr tf tg sf sg t1
                                             (t2',tv2,sv2) <- gr tf tg sf sg t2 
                                             return (IsFixS t1' t2'
                                                    ,tv1 `union` tv2
                                                    ,sv1 `union` sv2)
  gr tf tg sf sg (IsIf tr tb t1 t2)     = do (tr',tvr,svr) <- gr tf tg sf sg tr
                                             (tb',tvb,svb) <- gr tf tg sf sg tb 
                                             (t1',tv1,sv1) <- gr tf tg sf sg t1 
                                             (t2',tv2,sv2) <- gr tf tg sf sg t2
                                             return (IsIf tr' tb' t1' t2'
                                                    ,unite [tvr,tvb,tv1,tv2]
                                                    ,unite [svr,svb,sv1,sv2])
  gr tf tg sf sg (IsCase tr te cts)     = do (tr',tvr,svr) <- gr tf tg sf sg tr
                                             (te',tvb,svb) <- gr tf tg sf sg te 
                                             (cts',tv1,sv1)<- gr tf tg sf sg cts 
                                             return (IsCase tr' te' cts'
                                                    ,unite [tvr,tvb,tv1]
                                                    ,unite [svr,svb,sv1])
  gr tf tg sf sg (IsArithOp nop 
                            t t1 t2)    = do (t' ,tv ,sv)  <- gr tf tg sf sg t 
                                             (t1',tv1,sv1) <- gr tf tg sf sg t1 
                                             (t2',tv2,sv2) <- gr tf tg sf sg t2
                                             return (IsArithOp nop t' t1' t2'
                                                    ,unite [tv,tv1,tv2]
                                                    ,unite [sv,sv1,sv2])
  gr tf tg sf sg (IsBoolOp  bop 
                            t t1 t2)    = do (t' ,tv ,sv)  <- gr tf tg sf sg t 
                                             (t1',tv1,sv1) <- gr tf tg sf sg t1 
                                             (t2',tv2,sv2) <- gr tf tg sf sg t2
                                             return (IsBoolOp  bop t' t1' t2'
                                                    ,unite [tv,tv1,tv2]
                                                    ,unite [sv,sv1,sv2])
  gr tf tg sf sg (IsRelOp   rop 
                            t t1 t2)    = do (t' ,tv ,sv)  <- gr tf tg sf sg t 
                                             (t1',tv1,sv1) <- gr tf tg sf sg t1 
                                             (t2',tv2,sv2) <- gr tf tg sf sg t2
                                             return (IsRelOp   rop t' t1' t2'
                                                    ,unite [tv,tv1,tv2]
                                                    ,unite [sv,sv1,sv2])
  gr tf tg sf sg (IsConstrOf sdt t)     = do (t' ,tv ,sv)  <- gr tf tg sf sg t 
                                             return (IsConstrOf sdt t',tv,sv)
  gr tf tg sf sg (Guarded    t d)       = do (t' ,tv ,sv)  <- gr tf tg sf sg t 
                                             (d' ,tvd,svd) <- gr tf tg sf sg d
                                             return (Guarded t' d'
                                                    ,tv `union` tvd
                                                    ,sv `union` svd)
  gr tf tg sf sg (GuardedNeg t d)       = do (t' ,tv ,sv)  <- gr tf tg sf sg t 
                                             (d' ,tvd,svd) <- gr tf tg sf sg d
                                             return (GuardedNeg t' d'
                                                    ,tv `union` tvd
                                                    ,sv `union` svd)
  gr tf tg sf sg (GuardedConstr c t d)  = do (t' ,tv ,sv)  <- gr tf tg sf sg t 
                                             (d' ,tvd,svd) <- gr tf tg sf sg d
                                             return (GuardedConstr c t' d'
                                                    ,tv `union` tvd
                                                    ,sv `union` svd)
  gr tf tg sf sg (Unify t1 t2)          = do (t1',tv1,sv1) <- gr tf tg sf sg t1
                                             (t2',tv2,sv2) <- gr tf tg sf sg t2
                                             return (Unify t1' t2'
                                                    ,tv1 `union` tv2
                                                    ,sv1 `union` sv2)
  -------------------------
  --- Begin Alejandro -----
  -------------------------
  gr tf tg sf sg (GuardedDinData c t p) = 
            do (t', tv, sv) <- gr tf tg sf sg t
               (p' , ptv, psv) <- gr tf tg sf sg p
               return(GuardedDinData c t' p', 
                         tv `union` ptv, 
                         sv `union` psv)

  gr tf tg sf sg (HasC t1 c t2) = 
            do (t1', tv1, sv1) <- gr tf tg sf sg t1
               (t2' , tv2, sv2) <- gr tf tg sf sg t2
               return(HasC t1' c t2', 
                         tv1 `union` tv2, 
                         sv1 `union` sv2)

  gr tf tg sf sg (ConstrsOf t) = 
            do (t', tv, sv) <- gr tf tg sf sg t
               return(ConstrsOf t', tv, sv)

  gr tf tg sf sg (IndexOf t) = 
            do (t', tv, sv) <- gr tf tg sf sg t
               return(IndexOf t', tv, sv)

  gr tf tg sf sg (UnifyData t1 cname t2) = 
            do (t1', tv1, sv1) <- gr tf tg sf sg t1
               (t2' , tv2, sv2) <- gr tf tg sf sg t2
               return(UnifyData t1' cname t2', 
                         tv1 `union` tv2, 
                         sv1 `union` sv2)
  
  -------------------------
  ----- End Alejandro -----
  -------------------------
  gr _  _  _  _  pred                   = return (pred,[],[])

-- -------------------------------------------------------------
-- Generalization for terms
instance Generalizable ResTerm where
 gr tf tg sf sg (RApp   e1 e2)   = do (e1',tv1,sv1) <- gr tf tg sf sg e1
                                      (e2',tv2,sv2) <- gr tf tg sf sg e2
                                      return (RApp e1' e2'
                                             ,unite [tv1,tv2]
                                             ,unite [sv1,sv2])
 gr tf tg sf sg (RLam x e)       = do (e' ,tve,sve) <- gr tf tg sf sg e
                                      return (RLam x e', tve, sve)
 gr tf tg sf sg (RFix e)         = do (e' ,tve,sve) <- gr tf tg sf sg e
                                      return (RFix e', tve, sve)
 gr tf tg sf sg (RLet xes e)     = do (xs',tvs,svs) <- gr tf tg sf sg xes
                                      (e' ,tve,sve) <- gr tf tg sf sg e
                                      return (RLet xs' e'
                                             ,tvs `union` tve
                                             ,svs `union` sve)
 gr tf tg sf sg (RArithOp
                  nop e1 e2)     = do (e1',tv1,sv1) <- gr tf tg sf sg e1
                                      (e2',tv2,sv2) <- gr tf tg sf sg e2
                                      return (RArithOp nop e1' e2'
                                             ,tv1 `union` tv2
                                             ,sv1 `union` sv2)
 gr tf tg sf sg (RIf e e1 e2)    = do (e' ,tv ,sv ) <- gr tf tg sf sg e
                                      (e1',tv1,sv1) <- gr tf tg sf sg e1
                                      (e2',tv2,sv2) <- gr tf tg sf sg e2
                                      return (RIf e' e1' e2'
                                             ,unite [tv,tv1,tv2]
                                             ,unite [sv,sv1,sv2])
 gr tf tg sf sg (RBoolOp 
                  bop e1 e2)     = do (e1',tv1,sv1) <- gr tf tg sf sg e1
                                      (e2',tv2,sv2) <- gr tf tg sf sg e2
                                      return (RBoolOp  bop e1' e2'
                                             ,tv1 `union` tv2
                                             ,sv1 `union` sv2)
 gr tf tg sf sg (RRelOp 
                  rop   e1 e2)   = do (e1',tv1,sv1) <- gr tf tg sf sg e1
                                      (e2',tv2,sv2) <- gr tf tg sf sg e2
                                      return (RRelOp   rop e1' e2'
                                             ,unite [tv1,tv2]
                                             ,unite [sv1,sv2])
 gr tf tg sf sg (RTuple es)      = do (es',tv ,sv ) <- gr tf tg sf sg es 
                                      return (RTuple es', tv, sv)
 gr tf tg sf sg (RProj i n e)    = do (e' ,tve,sve) <- gr tf tg sf sg e
                                      return (RProj i n e',tve ,sve)
 gr tf tg sf sg (RTupleS es)     = do (es',tv ,sv ) <- gr tf tg sf sg es 
                                      return (RTupleS es', tv, sv)
 gr tf tg sf sg (RProjS i n e)   = do (e' ,tve,sve) <- gr tf tg sf sg e
                                      return (RProjS i n e' ,tve ,sve)
 gr tf tg sf sg (REFill   v e)   = do (v' ,tvv,svv) <- gr tf tg sf sg v
                                      (e' ,tve,sve) <- gr tf tg sf sg e
                                      return (REFill v' e'
                                             ,unite [tvv,tve]
                                             ,unite [svv,sve])
 gr tf tg sf sg (RELam h e)      = do (e' ,tv ,sv ) <- gr tf tg sf sg e
                                      return (RELam h e', tv, sv)
 gr tf tg sf sg (RELet x e1 e2)  = do (e1',tv1,sv1) <- gr tf tg sf sg e1
                                      (e2',tv2,sv2) <- gr tf tg sf sg e2
                                      return (RELet x e1' e2'
                                             ,tv1 `union` tv2
                                             ,sv1 `union` sv2)
 gr tf tg sf sg (REApp   e v)    = do (e' ,tve,sve) <- gr tf tg sf sg e
                                      (v' ,tvv,svv) <- gr tf tg sf sg v
                                      return (REApp e' v'
                                             ,unite [tve,tvv]
                                             ,unite [sve,svv])
 gr tf tg sf sg (REIf v e1 e2)   = do (v' ,tv ,sv ) <- gr tf tg sf sg v
                                      (e1',tv1,sv1) <- gr tf tg sf sg e1
                                      (e2',tv2,sv2) <- gr tf tg sf sg e2
                                      return (REIf v' e1' e2'
                                             ,unite [tv,tv1,tv2]
                                             ,unite [sv,sv1,sv2])
 gr tf tg sf sg (RECase v pes)   = do (v' ,tv ,sv ) <- gr tf tg sf sg v
                                      (pes',tv1,sv1)<- gr tf tg sf sg pes
                                      return (RECase v' pes'
                                             ,tv`union`tv1
                                             ,sv`union`sv1)
 gr tf tg sf sg (REAppS e ea)    = do (e' ,tv ,sv ) <- gr tf tg sf sg e
                                      (es',tvs,svs) <- gr tf tg sf sg ea
                                      return (REAppS e' es'
                                             ,tv `union` tvs
                                             ,sv `union` svs)
 gr tf tg sf sg (RData ds e)     = do (e' ,tv,sv) <- gr tf tg sf sg e    
                                      return (RData ds e',tv,sv)         
 gr tf tg sf sg (RConstruc c es) = do (es',tv,sv) <- gr tf tg sf sg es   
                                      return (RConstruc c es',tv,sv)     
 gr tf tg sf sg (RCase e   cs)   = do (e' ,tv, sv)  <- gr tf tg sf sg e  
                                      (cs',tvc,svc) <- gr tf tg sf sg cs 
                                      return (RCase e' cs'               
                                             ,tv`union`tvc               
                                             ,sv`union`svc)              
 gr _  _  _  _  e                =    return (e,[],[])

--------
instance Generalizable EvTerm where
 gr tf tg sf sg (ECxt  c)        = do (c' ,tv ,sv)  <- gr tf tg sf sg c
                                      return (ECxt c', tv, sv)
 gr tf tg sf sg (EComp v   e)    = do (v' ,tvv,svv) <- gr tf tg sf sg v
                                      (e' ,tve,sve) <- gr tf tg sf sg e
                                      return (EComp v' e'
                                             ,unite [tvv,tve]
                                             ,unite [svv,sve])
 gr tf tg sf sg (EStaticExp e)   = do (e' ,tv ,sv)  <- gr tf tg sf sg e
                                      return (EStaticExp e', tv, sv)
 gr tf tg sf sg (EAppS e ea)     = do (e' ,tv ,sv)  <- gr tf tg sf sg e
                                      (es',tvs,svs) <- gr tf tg sf sg ea
                                      return (EAppS e' es'
                                             ,tv `union` tvs
                                             ,sv `union` svs)
 gr tf tg sf sg (ETupleS vs)     = do (vs',tv ,sv)  <- gr tf tg sf sg vs
                                      return (ETupleS vs', tv, sv)
 gr tf tg sf sg (EProjS i n v)   = do (v' ,tvv,svv) <- gr tf tg sf sg v
                                      return (EProjS i n v',tvv ,svv)
 gr _  _  _  _  v                =    return (v,[],[])

--------
instance Generalizable CxtTerm where
 gr tf tg sf sg (CELet x c e)    = do (c' ,tvc,svc) <- gr tf tg sf sg c
                                      (e' ,tve,sve) <- gr tf tg sf sg e
                                      return (CELet x c' e'
                                             ,tvc `union` tve
                                             ,svc `union` sve)
 gr tf tg sf sg (CELam xs ct)    = do (ct',tv ,sv ) <- gr tf tg sf sg ct
                                      return (CELam xs ct', tv, sv)
 gr tf tg sf sg (CEApp   c v)    = do (c' ,tvc,svc) <- gr tf tg sf sg c
                                      (v' ,tvv,svv) <- gr tf tg sf sg v
                                      return (CEApp c' v'
                                             ,unite [tvc,tvv]
                                             ,unite [svc,svv])
 gr _ _ _ _ c                    =    return (c,[],[])

