module ResidualReduction where

import MyPrelude
import ResidualLang
import BasicData

-- ------------------------------------
-- Filling contexts
-- ------------------------------------
noredfill, redfill :: ContextTerm -> ResidualTerm -> ResidualTerm
noredfill c e = fill c e
redfill c e = ered (fill c e)

fill CHole           e = e
fill (CELam h c)     e = RELam h (fill c e)
fill (CEApp c v)     e = REApp (fill c e) v
fill (CELet x c e')  e = RELet x (fill c e) e'

-- ------------------------------------
--  Reduction of evidence
-- ------------------------------------
ered :: ResidualTerm -> ResidualTerm
ered (RApp e1 e2)              = RApp (ered e1) (ered e2)
ered (RLam x e)                = RLam x (ered e)
ered (RFix e)                  = RFix (ered e)
ered (RLet xes e)              = RLet (map (\(x,e) -> (x,ered e)) xes)
                                      (ered e)
ered (RArithOp nop e1 e2)      = RArithOp nop (ered e1) (ered e2)
ered (RIf e e1 e2)             = RIf (ered e) (ered e1) (ered e2)
ered (RBoolOp bop e1 e2)       = RBoolOp bop (ered e1) (ered e2)
ered (RRelOp rop e1 e2)        = RRelOp rop (ered e1) (ered e2)
ered (RTuple es)               = RTuple (map ered es)
ered (RProj i n e)             = RProj i n (ered e)
ered (RTupleS es)              = RTupleS (map ered es)
ered (RProjS i n e)            = let e' = ered e
                                  in case e' of
                                      RTupleS es -> 
                                        if i>length es
                                         then error (outboundIn "ered")
                                         else es!!(i-1)
                                      _ -> RProjS i n e'
ered (REFill v e)              = eredfill v (ered e)
ered (RELam [] e)              = ered e
ered (RELam h e)               = RELam h (ered e)
ered (RELet x e1 e2)           = betared e2 ressubst [x] [ered e1]
ered (REIf v e1 e2)            = case v of
                                  EBool b -> if b then ered e1 else ered e2
                                  _       -> REIf v (ered e1) (ered e2)
ered (RECase v pes)            = case v of
                                  EConstr c -> chooseBranch c pes
                                  _       -> 
                                    RECase v
                                      (map (\(p,e) -> (p,ered e)) pes)
ered (REApp e [])              = ered e
ered (REApp e v)               = let e' = ered e
                                  in case e' of
                                      RELam h e'' 
                                        -> if length h == length v
                                            then betared e'' evsubst h v 
                                            else error (canthpnSo ("Term: "
                                                        ++ show (REApp e v)))
                                      _ -> REApp e' v
ered (REAppS e ea)             = let e' = ered e
                                  in case e' of
                                      RLam x e'' 
                                        -> betared e'' ressubst [x] [ea]
                                      _ -> REAppS e' (ered ea)
ered (RData ds e)              = RData ds (ered e)                        
ered (RConstruc c es)          = RConstruc c (map ered es)                
ered (RCase e cs)              = RCase       (ered e)                     
                                             (map (\(p,e) -> (p, ered e)) 
                                                  cs)                     
--------------------------------
------- Begin Alejandro --------
--------------------------------
ered (RDConstruc c v cs)   = RDConstruc c (ered v) (map ered cs)
ered (RProtoCase e et cs)  = RProtoCase (ered e) (ered et)                                                              (map (\(p,e) -> (p, ered e)) 
                                                  cs)                     
--------------------------------
-------- End Alejandro ---------
--------------------------------
        
ered e                         = e 

--------------
betared :: ResTerm -> (ResTerm -> a -> b -> ResTerm) -> [a] -> [b] -> ResTerm
betared e subst xs es = ered (accuml (subst . ered) e xs es)

--------------
chooseBranch _ [] = error "Static case does not cover all cases!"
chooseBranch c ((c', ei):ces) = if c==c' then ei else chooseBranch c ces

--------------
eredfill (ECxt c)          e = redfill c e  
eredfill (EComp v    v')   e = eredfill v (eredfill v' e)
eredfill (EProjS i n v   ) e = RProjS i n (eredfill v e)
eredfill (ETupleS vs)      e = RTupleS (map (\v -> eredfill v e) vs)
eredfill v                 e = REFill v e

-- ------------------------------------
--  Substitution for residual terms
-- ------------------------------------
class ResSubst ty where
  ressubst :: ty -> ResTermVble -> ResidualTerm -> ty

  ressubst t _ _ = t

instance ResSubst Char
instance ResSubst ResType

instance ResSubst ResTerm where
 ressubst e@(RVar x)           x' e' = if x==x' then e' else e
 ressubst (RApp   e1 e2)       x' e' = RApp          (ressubst e1  x' e') 
                                                     (ressubst e2  x' e')
 ressubst e@(RLam x e'')       x' e' = if x==x' then e else 
                                       RLam x        (ressubst e'' x' e')
 ressubst (RFix e)             x' e' = RFix          (ressubst e   x' e')
 ressubst (RLet xes e)         x' e' = let xs = map fst xes in
                                       RLet          (ressubst xes x' e')
                                                     (if x' `elem` xs then e 
                                                                      else 
                                                      ressubst e   x' e'
                                                     )
 ressubst (RArithOp nop e1 e2) x' e' = RArithOp nop  (ressubst e1  x' e') 
                                                     (ressubst e2  x' e')
 ressubst (RIf e e1 e2)        x' e' = RIf           (ressubst e   x' e') 
                                                     (ressubst e1  x' e') 
                                                     (ressubst e2  x' e')
 ressubst (RBoolOp bop e1 e2)  x' e' = RBoolOp bop   (ressubst e1  x' e') 
                                                     (ressubst e2  x' e')
 ressubst (RRelOp rop   e1 e2) x' e' = RRelOp rop    (ressubst e1  x' e') 
                                                     (ressubst e2  x' e')
 ressubst (RTuple es)          x' e' = RTuple        (ressubst es  x' e')
 ressubst (RProj i n e  )      x' e' = RProj i n     (ressubst e   x' e')  
 ressubst (RTupleS es)         x' e' = RTupleS       (ressubst es  x' e')
 ressubst (RProjS i n e  )     x' e' = RProjS i n    (ressubst e   x' e')  
 ressubst (REFill   v e)       x' e' = REFill        (ressubst v   x' e')
                                                     (ressubst e   x' e')   
 ressubst (RELam h e)          x' e' = RELam h       (ressubst e   x' e')
 ressubst e@(RELet x e1 e2)    x' e' = if x==x' then e else 
                                       RELet x       (ressubst e1  x' e')
                                                     (ressubst e2  x' e')
 ressubst (REApp   e v)        x' e' = REApp         (ressubst e   x' e') 
                                                     (ressubst v   x' e')
 ressubst (REIf v e1 e2)       x' e' = REIf          (ressubst v   x' e')
                                                     (ressubst e1  x' e') 
                                                     (ressubst e2  x' e')
 ressubst (RECase v pes)       x' e' = RECase        (ressubst v   x' e')
                                                     (ressubst pes x' e') 
 ressubst (REAppS e ea)        x' e' = REAppS        (ressubst e   x' e') 
                                                     (ressubst ea  x' e')
 ressubst (RData ds e)         x' e' = RData ds      (ressubst e   x' e') 
 ressubst (RConstruc c es)     x' e' = RConstruc c   (ressubst es  x' e') 
 -------------------------------
 ---- Begin Alejandro ----------
 -------------------------------
 ressubst (RDConstruc c v cs) x' e' = RDConstruc c (ressubst v x' e') 
                                                   (ressubst cs x' e')
 ressubst (RProtoCase e et cs) x' e' = RProtoCase    (ressubst e   x' e') 
                                                     (ressubst et  x' e') 
                                                     (ressubst cs  x' e') 
 -------------------------------
 ------ End Alejandro ----------
 -------------------------------

 ressubst (RCase e   cs)       x' e' = RCase         (ressubst e   x' e') 
                                                     (ressubst cs  x' e') 
 ressubst e                    _  _  = e 
 
----------
instance ResSubst a => ResSubst [a] where
  ressubst as x e = map (\a -> ressubst a x e) as

----------
instance (ResSubst a,ResSubst b) => ResSubst (a,b) where
  ressubst (a,b) x e = (ressubst a x e, ressubst  b x e)

-- ------------------------------------
instance ResSubst EvTerm where
  ressubst (EComp v1   v2)   x' e' = EComp      (ressubst v1 x' e')   
                                                (ressubst v2 x' e')
  ressubst (ETupleS vs)      x' e' = ETupleS    (ressubst vs x' e')
  ressubst (EProjS i n v'  ) x' e' = EProjS i n (ressubst v' x' e')  
  ressubst (EAppS e ea)      x' e' = EAppS      (ressubst e  x' e')
                                                (ressubst ea x' e')
  ressubst (EStaticExp e)    x' e' = EStaticExp (ressubst e  x' e')
  ressubst v                 _  _  = v

-- ------------------------------------
--  Substitution for evidence
-- ------------------------------------
class EvSubst ty where
  evsubst :: ty -> EvVble -> EvTerm -> ty

  evsubst t _ _ = t

instance EvSubst Char
instance EvSubst ResType

-- ------------------------------------
instance EvSubst ResTerm where
 evsubst (RApp   e1 e2)        h v = RApp         (evsubst e1  h v)
                                                  (evsubst e2  h v)
 evsubst (RLam x e)            h v = RLam x       (evsubst e   h v)
 evsubst (RFix e)              h v = RFix         (evsubst e   h v)
 evsubst (RLet xes e)          h v = RLet         (evsubst xes h v)
                                                  (evsubst e   h v)
 evsubst (RArithOp nop e1 e2)  h v = RArithOp nop (evsubst e1  h v) 
                                                  (evsubst e2  h v)
 evsubst (RIf e e1 e2)         h v = RIf          (evsubst e   h v) 
                                                  (evsubst e1  h v) 
                                                  (evsubst e2  h v)
 evsubst (RBoolOp bop e1 e2)   h v = RBoolOp bop  (evsubst e1  h v) 
                                                  (evsubst e2  h v)
 evsubst (RRelOp nop   e1 e2)  h v = RRelOp nop   (evsubst e1  h v) 
                                                  (evsubst e2  h v)
 evsubst (RTuple es)           h v = RTuple       (evsubst es  h v)
 evsubst (RProj  i n e  )      h v = RProj  i n   (evsubst e   h v)  
 evsubst (RTupleS es)          h v = RTupleS      (evsubst es  h v)
 evsubst (RProjS i n e  )      h v = RProjS i n   (evsubst e   h v) 
 evsubst e@(REVar h')          h v = if h==h' then evidence2residual v else e
 evsubst (REApp   e v')        h v = REApp        (evsubst e   h v) 
                                                  (evsubst v'  h v)
 evsubst e'@(RELam h'  e)      h v = if h `elem` h' then e' else 
                                     RELam h'     (evsubst e   h v)
 evsubst (RELet x e1 e2)       h v = RELet x      (evsubst e1  h v)
                                                  (evsubst e2  h v)
 evsubst (REIf v' e1 e2)       h v = REIf         (evsubst v'  h v)
                                                  (evsubst e1  h v) 
                                                  (evsubst e2  h v)
 evsubst (RECase v' pes)       h v = RECase       (evsubst v'  h v)
                                                  (evsubst pes h v) 
 evsubst (REFill   v' e)       h v = REFill       (evsubst v'  h v) 
                                                  (evsubst e   h v)
 evsubst (REAppS e ea)         h v = REAppS       (evsubst e   h v) 
                                                  (evsubst ea  h v)
 evsubst (RData ds e)          h v = RData ds     (evsubst e   h v)   
 evsubst (RConstruc c es)      h v = RConstruc c  (evsubst es  h v)   
 -------------------------------
 ---- Begin Alejandro ----------
 -------------------------------
 evsubst (RDConstruc c v' es)      h v = 
    let v''  = evsubst v' h v  
        es'  = evsubst es h v 
    in RDConstruc c v'' es'    

 evsubst (RProtoCase e et cs)  h v = RProtoCase (evsubst e h v)   
                                                (evsubst et h v)
                                                (evsubst cs h v)

 -------------------------------
 ------ End Alejandro ----------
 -------------------------------
 evsubst (RCase e   cs)        h v = RCase        (evsubst e   h v)   
                                                  (evsubst cs  h v)   
 evsubst e                     _ _ = e  
 
-- ------------------------------------
instance EvSubst EvTerm where
  evsubst v'@(EVar h')      h v = if h==h' then v else v' 
  evsubst (ECxt c)          h v = ECxt       (evsubst c  h v)
  evsubst (EComp v1   v2)   h v = EComp      (evsubst v1 h v) 
                                             (evsubst v2 h v)
  evsubst (ETupleS vs)      h v = ETupleS    (evsubst vs h v)
  evsubst (EProjS i n v'  ) h v = EProjS i n (evsubst v' h v)  
  evsubst (EAppS e ea)      h v = EAppS      (evsubst e  h v)
                                             (evsubst ea h v)
  evsubst (EStaticExp e)    h v = EStaticExp (evsubst e  h v)
  evsubst v                 _ _ = v

-- ------------------------------------
instance EvSubst CxtTerm where
  evsubst v'@(CELam h' c)     h v = if h `elem` h' then v' else 
                                    CELam h'  (evsubst c  h v)
  evsubst (CEApp   c v')      h v = CEApp     (evsubst c  h v) 
                                              (evsubst v' h v)
  evsubst (CELet x c e)       h v = CELet x   (evsubst c  h v)
                                              (evsubst e  h v)
  evsubst c                   _ _ = c

----------
instance EvSubst a => EvSubst [a] where
  evsubst as h v = map (\a -> evsubst a h v) as

----------
instance (EvSubst a,EvSubst b) => EvSubst (a,b) where
  evsubst (a,b) h v = (evsubst a h v, evsubst  b h v)

{-
-- ------------------------------------
--  Substitution for types
-- ------------------------------------
class TySubst ty where
  tysubst  :: ty -> ResTypeVble   -> ResType       -> ty
  schsubst :: ty -> ResSchemeVble -> ResTypeScheme -> ty

  tysubst  t _ _ = t
  schsubst t _ _ = t

instance TySubst Char

-- ------------------------------------
instance TySubst ResTerm where
 -------
 tysubst (RApp t' e1 e2)       t tau = RApp         (tysubst t'  t tau) 
                                                    (tysubst e1  t tau) 
                                                    (tysubst e2  t tau)
 tysubst (RLam x e)            t tau = RLam x       (tysubst e   t tau)
 tysubst (RFix e)              t tau = RFix         (tysubst e   t tau)
 tysubst (RLet xes e)          t tau = RLet         (tysubst xes t tau)
                                                    (tysubst e   t tau)
 tysubst (RArithOp nop e1 e2)  t tau = RArithOp nop (tysubst e1  t tau) 
                                                    (tysubst e2  t tau)
 tysubst (RIf e e1 e2)         t tau = RIf          (tysubst e   t tau) 
                                                    (tysubst e1  t tau) 
                                                    (tysubst e2  t tau)
 tysubst (RBoolOp bop e1 e2)   t tau = RBoolOp  bop (tysubst e1  t tau) 
                                                    (tysubst e2  t tau)
 tysubst (RRelOp rop t' e1 e2) t tau = RRelOp   rop (tysubst t'  t tau)
                                                    (tysubst e1  t tau) 
                                                    (tysubst e2  t tau)
 tysubst (RTuple es)           t tau = RTuple       (tysubst es  t tau)
 tysubst (RProj  i n e t')     t tau = RProj  i n   (tysubst e   t tau) 
                                                    (tysubst t'  t tau)
 tysubst (RTupleS es)          t tau = RTupleS      (tysubst es  t tau)
 tysubst (RProjS i n e t')     t tau = RProjS i n   (tysubst e   t tau) 
                                                    (tysubst t'  t tau)
 tysubst (REApp t' e v)        t tau = REApp        (tysubst t'  t tau)
                                                    (tysubst e   t tau) 
                                                    (tysubst v   t tau)
 tysubst (RELam h (e,t'))      t tau = RELam h      (tysubst e   t tau
                                                    ,tysubst t'  t tau)
 tysubst (RELet x e1 e2)       t tau = RELet x      (tysubst e1  t tau) 
                                                    (tysubst e2  t tau)
 tysubst (REIf v e1 e2)        t tau = REIf         (tysubst v   t tau) 
                                                    (tysubst e1  t tau) 
                                                    (tysubst e2  t tau)
 tysubst (RECase v pes)        t tau = RECase       (tysubst v   t tau) 
                                                    (tysubst pes t tau) 
 tysubst (REFill t' v e)       t tau = REFill       (tysubst t'  t tau) 
                                                    (tysubst v   t tau) 
                                                    (tysubst e   t tau)   
 tysubst (REAppS e ea)         t tau = REAppS       (tysubst e   t tau) 
                                                    (tysubst ea  t tau)
 tysubst (RTApp e ts ss)       t tau = RTApp        (tysubst e   t tau) 
                                                    (tysubst ts  t tau)
                                                    (tysubst ss  t tau)
 tysubst e'@(RTLam ts ss e)    t tau = if t `elem` ts then e' else
                                       RTLam ts ss  (tysubst e   t tau)
 tysubst (RData ds e)          t tau = RData ds     (tysubst e   t tau) 
 tysubst (RConstruc c es)      t tau = RConstruc c  (tysubst es  t tau) 
 tysubst (RCase e t' cs)       t tau = RCase        (tysubst e   t tau) 
                                                    (tysubst t'  t tau) 
                                                    (tysubst cs  t tau) 
 tysubst e                     _ _   = e 

 -------
 schsubst (RApp t e1 e2)       s sigma = RApp        (schsubst t   s sigma) 
                                                     (schsubst e1  s sigma) 
                                                     (schsubst e2  s sigma)
 schsubst (RLam x e)           s sigma = RLam x      (schsubst e   s sigma)
 schsubst (RFix e)             s sigma = RFix        (schsubst e   s sigma)
 schsubst (RLet xes e)         s sigma = RLet        (schsubst xes s sigma)
                                                     (schsubst e   s sigma)
 schsubst (RArithOp nop e1 e2) s sigma = RArithOp nop 
                                                     (schsubst e1  s sigma) 
                                                     (schsubst e2  s sigma)
 schsubst (RIf e e1 e2)        s sigma = RIf         (schsubst e   s sigma) 
                                                     (schsubst e1  s sigma) 
                                                     (schsubst e2  s sigma)
 schsubst (RBoolOp bop e1 e2)  s sigma = RBoolOp bop (schsubst e1  s sigma) 
                                                     (schsubst e2  s sigma)
 schsubst (RRelOp rop t e1 e2) s sigma = RRelOp rop  (schsubst t   s sigma)
                                                     (schsubst e1  s sigma) 
                                                     (schsubst e2  s sigma)
 schsubst (RTuple es)          s sigma = RTuple      (schsubst es  s sigma)
 schsubst (RProj  i n e t)     s sigma = RProj  i n  (schsubst e   s sigma) 
                                                     (schsubst t   s sigma)
 schsubst (RTupleS es)         s sigma = RTupleS     (schsubst es  s sigma)
 schsubst (RProjS i n e t)     s sigma = RProjS i n  (schsubst e   s sigma) 
                                                     (schsubst t   s sigma)
 schsubst (REApp t e v)        s sigma = REApp       (schsubst t   s sigma)
                                                     (schsubst e   s sigma) 
                                                     (schsubst v   s sigma)
 schsubst (RELam h (e,t))      s sigma = RELam h     (schsubst e   s sigma
                                                     ,schsubst t   s sigma)
 schsubst (RELet x e1 e2)      s sigma = RELet x     (schsubst e1  s sigma) 
                                                     (schsubst e2  s sigma)
 schsubst (REIf v e1 e2)       s sigma = REIf        (schsubst v   s sigma) 
                                                     (schsubst e1  s sigma) 
                                                     (schsubst e2  s sigma)
 schsubst (RECase v pes)       s sigma = RECase      (schsubst v   s sigma) 
                                                     (schsubst pes s sigma) 
 schsubst (REFill t v e)       s sigma = REFill      (schsubst t   s sigma) 
                                                     (schsubst v   s sigma) 
                                                     (schsubst e   s sigma)   
 schsubst (REAppS e ea)        s sigma = REAppS      (schsubst e   s sigma) 
                                                     (schsubst ea  s sigma)
 schsubst (RTApp e ts ss)      s sigma = RTApp       (schsubst e   s sigma) 
                                                     (schsubst ts  s sigma)
                                                     (schsubst ss  s sigma)
 schsubst e'@(RTLam ts ss e)   s sigma = if s `elem` ss then e' else
                                         RTLam ts ss (schsubst e  s sigma)
 schsubst (RData ds e)         s sigma = RData ds    (schsubst e  s sigma)
 schsubst (RConstruc c es)     s sigma = RConstruc c (schsubst es s sigma)
 schsubst (RCase e t' cs)      s sigma = RCase       (schsubst e  s sigma)
                                                     (schsubst t' s sigma)
                                                     (schsubst cs s sigma)
 schsubst e                    _ _     = e 

-- ------------------------------------
instance TySubst EvTerm where
  -------
  tysubst (ECxt c)          t tau = ECxt       (tysubst c  t tau)
  tysubst (EComp v1 t' v2)  t tau = EComp      (tysubst v1 t tau) 
                                               (tysubst t' t tau)
                                               (tysubst v2 t tau)
  tysubst (ETupleS vs)      t tau = ETupleS    (tysubst vs t tau)
  tysubst (EProjS i n v t') t tau = EProjS i n (tysubst v  t tau) 
                                               (tysubst t' t tau)
  tysubst (EAppS e ea)      t tau = EAppS      (tysubst e  t tau)
                                               (tysubst ea t tau)
  tysubst (EStaticExp e)    t tau = EStaticExp (tysubst e  t tau)
  tysubst v                 _  _    = v

  -------
  schsubst (ECxt c)          s sigma = ECxt       (schsubst c  s sigma)
  schsubst (EComp v1 t v2)   s sigma = EComp      (schsubst v1 s sigma) 
                                                  (schsubst t  s sigma)
                                                  (schsubst v2 s sigma)
  schsubst (ETupleS vs)      s sigma = ETupleS    (schsubst vs s sigma)
  schsubst (EProjS i n v t)  s sigma = EProjS i n (schsubst v  s sigma) 
                                                  (schsubst t  s sigma)
  schsubst (EAppS e ea)      s sigma = EAppS      (schsubst e  s sigma)
                                                  (schsubst ea s sigma)
  schsubst (EStaticExp e)    s sigma = EStaticExp (schsubst e  s sigma)
  schsubst v                 _ _     = v

-- ------------------------------------
instance TySubst CxtTerm where
  -------
  tysubst (CELam h (c,t')) t tau = CELam h     (tysubst c  t tau
                                               ,tysubst t' t tau)
  tysubst (CEApp t' c v)   t tau = CEApp       (tysubst t' t tau)
                                               (tysubst c  t tau) 
                                               (tysubst v  t tau)
  tysubst (CELet x c e)    t tau = CELet x     (tysubst c  t tau) 
                                               (tysubst e  t tau)
  tysubst (CTLam ts ss c)  t tau = CTLam ts ss (tysubst c  t tau)
  tysubst (CTApp c t' s)   t tau = CTApp       (tysubst c  t tau)
                                               (tysubst t' t tau) 
                                               (tysubst s  t tau)
  tysubst c                _ _   = c

  -------
  schsubst (CELam h (c,t))  s sigma = CELam h     (schsubst c  s sigma
                                                  ,schsubst t  s sigma)
  schsubst (CEApp t c v)    s sigma = CEApp       (schsubst t  s sigma)
                                                  (schsubst c  s sigma) 
                                                  (schsubst v  s sigma)
  schsubst (CELet x c e)    s sigma = CELet x     (schsubst c  s sigma) 
                                                  (schsubst e  s sigma)
  schsubst (CTLam ts ss c)  s sigma = CTLam ts ss (schsubst c  s sigma)
  schsubst (CTApp c t s')   s sigma = CTApp       (schsubst c  s sigma)
                                                  (schsubst t  s sigma) 
                                                  (schsubst s' s sigma)
  schsubst c                _ _     = c

-- ------------------------------------
instance TySubst ResType where
  -------
  tysubst tau'@(RTVar t')    t tau = if (t'==t) then tau else tau'
  tysubst (RTTuple  ts)      t tau = RTTuple      (tysubst ts t tau)
  tysubst (RTFun    t1 t2)   t tau = RTFun        (tysubst t1 t tau) 
                                                  (tysubst t2 t tau)
  tysubst (RTStaticExp e)    t tau = RTStaticExp  (tysubst e  t tau)
  tysubst (RTFixS ts t' s)   t tau = RTFixS       (tysubst ts t tau)
                                                  (tysubst t' t tau) 
                                                  (tysubst s  t tau)
  tysubst (RTTypedS ts t' s) t tau = RTTypedS     (tysubst ts t tau)
                                                  (tysubst t' t tau) 
                                                  (tysubst s  t tau)
  tysubst (RTPoly   s)       t tau = RTPoly       (tysubst s  t tau)
  tysubst (RTTupleS ts)      t tau = RTTupleS     (tysubst ts t tau)
  tysubst (RTData d ts ss)   t tau = RTData d     (tysubst ts t tau)  
                                                  (tysubst ss t tau)  
  tysubst (RTConstruc c ts)  t tau = RTConstruc c (tysubst ts t tau)  
  tysubst tau                _ _   = tau
  
  -------
  schsubst (RTTuple  ts)      s sigma = RTTuple      (schsubst ts s sigma)
  schsubst (RTFun    t1 t2)   s sigma = RTFun        (schsubst t1 s sigma) 
                                                     (schsubst t2 s sigma)
  schsubst (RTStaticExp e)    s sigma = RTStaticExp  (schsubst e  s sigma)
  schsubst (RTFixS ts t s')   s sigma = RTFixS       (schsubst ts s sigma)
                                                     (schsubst t  s sigma) 
                                                     (schsubst s' s sigma)
  schsubst (RTTypedS ts t s') s sigma = RTTypedS     (schsubst ts s sigma)
                                                     (schsubst t  s sigma) 
                                                     (schsubst s' s sigma)
  schsubst (RTPoly   s')      s sigma = RTPoly       (schsubst s' s sigma)
  schsubst (RTTupleS ts)      s sigma = RTTupleS     (schsubst ts s sigma)
  schsubst (RTData d ts ss)   s sigma = RTData d     (schsubst ts s sigma)
                                                     (schsubst ss s sigma)
  schsubst (RTConstruc c ts)  s sigma = RTConstruc c (schsubst ts s sigma)
  schsubst tau                _  _    = tau

-- ------------------------------------
instance TySubst ResTypeScheme where
  tysubst (RSForAll tm sm qt) t tau = RSForAll tm sm (tysubst qt t tau)
  tysubst (RSTupleS ts)       t tau = RSTupleS (tysubst ts t tau) 
  tysubst sigma               _  _  = sigma

  schsubst sigma'@(RSVar s')   s sigma = if s'==s then sigma else sigma'
  schsubst (RSForAll tm sm qt) s sigma = RSForAll tm sm (schsubst qt s sigma)
  schsubst (RSTupleS ts)       s sigma = RSTupleS (schsubst ts s sigma) 
  schsubst sigma               _  _    = sigma

-- ------------------------------------
instance TySubst QualType where
  tysubst  (preds:=>t') t tau   = tysubst  preds t tau   :=> tysubst  t' t tau
  schsubst (preds:=>t)  s sigma = schsubst preds s sigma :=> schsubst t  s sigma

----------
instance TySubst a => TySubst [a] where
  tysubst  as t tau   = map (\a -> tysubst  a t tau)   as
  schsubst as s sigma = map (\a -> schsubst a s sigma) as

----------
instance (TySubst a,TySubst b) => TySubst (a,b) where
  tysubst  (a,b) t tau   = (tysubst  a t tau,   tysubst  b t tau)
  schsubst (a,b) s sigma = (schsubst a s sigma, schsubst b s sigma)

-- ------------------------------------
instance TySubst Pred where
  -------
  tysubst (IsMG       s1 s2)        t tau = IsMG    (tysubst s1  t tau)
                                                    (tysubst s2  t tau)
  tysubst (IsMGmorph t' s1 s2)      t tau = IsMGmorph 
                                                    (tysubst t'  t tau)
                                                    (tysubst s1  t tau)
                                                    (tysubst s2  t tau)
  tysubst (IsInt      t')           t tau = IsInt   (tysubst t'  t tau)
  tysubst (IsBool     t')           t tau = IsBool  (tysubst t'  t tau)
  tysubst (IsFunS     t' t'')       t tau = IsFunS  (tysubst t'  t tau)
                                                    (tysubst t'' t tau)
  tysubst (IsFixS     t' t'')       t tau = IsFixS  (tysubst t'  t tau)
                                                    (tysubst t'' t tau)
  tysubst (IsIf       tr tb t1 t2)  t tau = IsIf    (tysubst tr  t tau)
                                                    (tysubst tb  t tau)
                                                    (tysubst t1  t tau)
                                                    (tysubst t2  t tau)
  tysubst (IsCase     tr te cts)    t tau = IsCase  (tysubst tr  t tau)
                                                    (tysubst te  t tau)
                                                    (tysubst cts t tau)
  tysubst (IsArithOp  nop t' t1 t2) t tau = IsArithOp nop
                                                    (tysubst t'  t tau)
                                                    (tysubst t1  t tau)
                                                    (tysubst t2  t tau)
  tysubst (IsBoolOp   bop t' t1 t2) t tau = IsBoolOp  bop
                                                    (tysubst t'  t tau)
                                                    (tysubst t1  t tau)
                                                    (tysubst t2  t tau)
  tysubst (IsRelOp    rop t' t1 t2) t tau = IsRelOp   rop
                                                    (tysubst t'  t tau)
                                                    (tysubst t1  t tau)
                                                    (tysubst t2  t tau)
  tysubst (IsConstrOf sdt t1)       t tau = IsConstrOf sdt 
                                                    (tysubst t1  t tau)
  tysubst (Guarded    t' d)         t tau = Guarded (tysubst t'  t tau)
                                                    (tysubst d   t tau)
  tysubst (GuardedNeg t' d)         t tau = GuardedNeg
                                                    (tysubst t'  t tau)
                                                    (tysubst d   t tau)
  tysubst (GuardedConstr c t' d)    t tau = GuardedConstr c
                                                    (tysubst t'  t tau)
                                                    (tysubst d   t tau)

  tysubst (Unify      t1 t2)        t tau = Unify   (tysubst t1  t tau)
                                                    (tysubst t2  t tau)
  tysubst d                         _ _   = d   

  -------
  schsubst (IsMG       s1 s2)       s sigma = IsMG    (schsubst s1  s sigma)
                                                      (schsubst s2  s sigma)
  schsubst (IsMGmorph t s1 s2)      s sigma = IsMGmorph
                                                      (schsubst t   s sigma)
                                                      (schsubst s1  s sigma)
                                                      (schsubst s2  s sigma)
  schsubst (IsInt      t)           s sigma = IsInt   (schsubst t   s sigma)
  schsubst (IsBool     t)           s sigma = IsBool  (schsubst t   s sigma)
  schsubst (IsFunS     t t')        s sigma = IsFunS  (schsubst t   s sigma)
                                                      (schsubst t'  s sigma)
  schsubst (IsFixS     t t')        s sigma = IsFixS  (schsubst t   s sigma)
                                                      (schsubst t'  s sigma)
  schsubst (IsIf       tr tb t1 t2) s sigma = IsIf    (schsubst tr  s sigma)
                                                      (schsubst tb  s sigma)
                                                      (schsubst t1  s sigma)
                                                      (schsubst t2  s sigma)
  schsubst (IsCase     tr te cts)   s sigma = IsCase  (schsubst tr  s sigma)
                                                      (schsubst te  s sigma)
                                                      (schsubst cts s sigma)
  schsubst (IsArithOp  nop t t1 t2) s sigma = IsArithOp nop
                                                      (schsubst t   s sigma)
                                                      (schsubst t1  s sigma)
                                                      (schsubst t2  s sigma)
  schsubst (IsBoolOp   bop t t1 t2) s sigma = IsBoolOp  bop
                                                      (schsubst t   s sigma)
                                                      (schsubst t1  s sigma)
                                                      (schsubst t2  s sigma)
  schsubst (IsRelOp    rop t t1 t2) s sigma = IsRelOp   rop
                                                      (schsubst t   s sigma)
                                                      (schsubst t1  s sigma)
                                                      (schsubst t2  s sigma)
  schsubst (IsConstrOf sdt t1)      s sigma = IsConstrOf sdt 
                                                      (schsubst t1  s sigma)
  schsubst (Guarded    t d)         s sigma = Guarded (schsubst t   s sigma)
                                                      (schsubst d   s sigma)
  schsubst (GuardedNeg t d)         s sigma = GuardedNeg
                                                      (schsubst t   s sigma)
                                                      (schsubst d   s sigma)
  schsubst (GuardedConstr c t d)    s sigma = GuardedConstr c
                                                      (schsubst t   s sigma)
                                                      (schsubst d   s sigma)

  schsubst (Unify      t1 t2)       s sigma = Unify   (schsubst t1  s sigma)
                                                      (schsubst t2  s sigma)
  schsubst d                        _ _     = d   
-}

-- Term equality
class EqResidual t where
  eqR :: t -> t -> Bool

instance EqResidual ResTerm where
  eqR (RApp e1  e2)
      (RApp e1' e2')            = eqR e1 e1' && eqR e2 e2'
  eqR (RLam x e)   
      (RLam x' e')              = x==x' && eqR e e'
  eqR (RFix e) 
      (RFix e')                 = eqR e e'
  eqR (RLet xes  e)  
      (RLet xes' e')            = eqR xes xes' && eqR e e'
  eqR (RArithOp nop  e1  e2)
      (RArithOp nop' e1' e2')   = nop==nop' && eqR e1 e1' && eqR e2 e2'
  eqR (RIf e  e1  e2)
      (RIf e' e1' e2')          = eqR e e' && eqR e1 e1' && eqR e2 e2'
  eqR (RBoolOp bop  e1  e2)
      (RBoolOp bop' e1' e2')    = bop==bop' && eqR e1 e1' && eqR e2 e2'
  eqR (RRelOp rop    e1  e2)
      (RRelOp rop'   e1' e2')   = rop==rop' && eqR e1 e1' && eqR e2 e2'
  eqR (RTuple es)
      (RTuple es')              = eqR es es'
  eqR (RProj i  n  e   )
      (RProj i' n' e'  )        = i==i' && n==n' && eqR e e'
  eqR (RTupleS es)
      (RTupleS es')             = eqR es es'
  eqR (RProjS i  n  e   ) 
      (RProjS i' n' e'  )       = i==i' && n==n' && eqR e e' 
  eqR (REFill   v  e )
      (REFill   v' e')          = eqR v v' && eqR e e'
  eqR (RELam h   e    ) 
      (RELam h'  e'   )         = h==h' && eqR e e'
  eqR (RELet x  e1  e2)
      (RELet x' e1' e2')        = x==x' && eqR e1 e1' && eqR e2 e2'
  eqR (REIf v  e1  e2)
      (REIf v' e1' e2')         = eqR v v' && eqR e1 e1' && eqR e2 e2'
  eqR (RECase v  pes)
      (RECase v' pes')          = eqR v v' && eqR pes pes'
  eqR (REApp   e  v)  
      (REApp   e' v')           = eqR e e' && eqR v v'
  eqR (REAppS e  ea)
      (REAppS e' ea')           = eqR e e' && eqR ea ea'
  eqR (RData _ e)                                        
      (RData _ e')              = eqR e e'               
  eqR (RConstruc c  e)                                   
      (RConstruc c' e')         = c==c' && eqR e e'      
  eqR (RCase e     cs)                                   
      (RCase e'    cs')         = eqR e e' && eqR cs cs' 
  eqR e e'                      = e==e'             

instance EqResidual EvTerm where
  eqR v v' = True

instance EqResidual a => EqResidual [a] where
  eqR [] [] = True
  eqR (x:xs) (x':xs') = eqR x x' && eqR xs xs'

instance (EqResidual a,EqResidual b) => EqResidual (a,b) where
  eqR (a,b) (a',b') = eqR a a' && eqR b b'

instance EqResidual ResType where
  eqR _ _ = True

instance EqResidual Char where
  eqR = (==)


-------------------------
---- Begin Alejandro ----
-------------------------
-- Function for transform a protocase in a case
proto2case :: ResidualTerm -> ResidualTerm

proto2case (RProtoCase e et cs) = 
   case et of 
        (REConstrs valids) -> let branchs = concat (map (f valids) cs)
                              in RCase e (map g branchs)  
        _  -> (RProtoCase e et cs)  
   where f constrs b = -- error (show (constructorOfBranch b))
					   if constructorOfBranch b `elem` constrs
                       then [b] else [] 
         g (p, e) = (p, proto2case e)

proto2case (RApp e1 e2)  = RApp (proto2case e1) (proto2case e2)
proto2case (RLam x e)    = RLam x (proto2case e)
proto2case (RFix e)      = RFix (proto2case e)
proto2case (RLet xes e)  = RLet (map (\(x,e) -> (x,proto2case e)) xes)
                                      (proto2case e)
proto2case (RArithOp nop e1 e2) = RArithOp nop (proto2case e1) 
                                               (proto2case e2)
proto2case (RIf e e1 e2)  = RIf (proto2case e) (proto2case e1) 
                                (proto2case e2)
proto2case (RBoolOp bop e1 e2) = RBoolOp bop (proto2case e1) 
                                             (proto2case e2)
proto2case (RRelOp rop e1 e2)   = RRelOp rop (proto2case e1) (proto2case e2)
proto2case (RTuple es)          = RTuple (map proto2case es)
proto2case (RProj i n e)        = RProj i n (proto2case e)
proto2case (RTupleS es)         = RTupleS (map proto2case es)
proto2case (RProjS i n e)       = RProjS i n (proto2case e)
proto2case (RELam h e)          = RELam h (proto2case e)
proto2case (REIf v e1 e2)       = REIf v (proto2case e1) (proto2case e2)
proto2case (RECase v pes)       = RECase v (map g pes)
                                  where g (c,rt) = (c, proto2case rt)
proto2case (REApp e v)          = REApp (proto2case e) v
proto2case (REAppS e ea)        = REAppS (proto2case e) (proto2case ea)
proto2case (RConstruc c es)     = RConstruc c (map proto2case es)  
proto2case (RCase e cs)         = RCase (proto2case e) 
                                        (map g cs)
                                  where g (rt1, rt2) = (proto2case rt1,
                                                        proto2case rt2)
proto2case (RDConstruc c v cs) = RDConstruc c v (map proto2case cs)
proto2case (RData ds e)        = RData ds (proto2case e)
proto2case e = e


constructorOfBranch :: (ResidualTerm, ResidualTerm) -> ConstructorName
constructorOfBranch (RDConstruc cname (RInt n) _, _) = cname++"_"++(show n)
constructorOfBranch _ = fail "must be a legal branch"


-- Function for remove RTDData from a solved datatype
cleanDatasTag ds = concat (map cleanDataTag ds)

cleanDataTag (RDataDef dname [] [] [] [] ) = []
cleanDataTag (RDataDef dname [] [] [] constrs@(x:xs)) = 
 [ RDataDef dname [] [] [] (map cleanDataTagConstr constrs) ]

cleanDataTagConstr (Constructor name [rt]) = 
   Constructor name [cleanDataTagRType rt] 

-- Elim from residual types
-- Another?
cleanDataTagRType (RTDData name ws) = name
cleanDataTagRType (RTTuple rs) = RTTuple (map cleanDataTagRType rs)
cleanDataTagRType t = t

-------------------------
----- End Alejandro -----
-------------------------

