module Scripts (runEnvironment,nullEnv)where

import Trace
import SourceLang
import SourceLangParse
import SourceLangPrint
import SourceTypeCheck
import Error
import List
import Spec
import ResidualLangPrint
import SourceTypeCheckMonad
import IOExtensions
import Char

type ModuleName = String

type Command = String

type LoadedModule = (ModuleName,FilePath,SourceTerm)

type LoadedModules = [LoadedModule]

type Defs = [(ModuleName,DataDef)]

data Environment = Env LoadedModules [FilePath] [Command] Defs

--pathSep = '\\'
-- Unix: pathSep = '\\'
pathSep = '/'

--nullEnv = Env [] ["E:\\Facultad\\Cuarto\\PFA\\PTS-10.4\\examples"] [] []
-- Unix: nullEnv = Env [] ["E:\\Facultad\\Cuarto\\PFA\\PTS-10.4\\examples"] [] []
nullEnv = Env [] ["/home/fidel/Works/TypeSpec/code/PTS-V1.5/NewExamples"] [] []


isBackSpace :: Char -> Bool
isBackSpace c = (ord c == 8) || (c=='\b') || (ord c == 127)

modules :: Environment -> LoadedModules
modules (Env m _ _ _) = m

paths :: Environment -> [FilePath]
paths (Env _ p _ _) = p

commands :: Environment -> [Command]
commands (Env _ _ xs _) = xs

defs :: Environment -> [DataDef]
defs (Env _ _ _ defs) = map snd defs

removeDefs :: Environment -> String -> Environment
removeDefs (Env ms ps cmds defs) name = Env ms ps cmds (filter (\(x,_)->x/=name) defs)

namedDefs :: Environment -> Defs
namedDefs (Env _ _ _ defs) = defs

addPath :: Environment -> FilePath -> Environment
addPath  env                 [] = env
addPath (Env ms ps cmds defs) p = Env ms (p:ps) cmds defs

removePath :: Environment -> FilePath -> Environment
removePath (Env ms ps cmds defs) p = Env ms (filter (/=p) ps) cmds defs

appendModule :: Environment -> LoadedModule -> Environment
appendModule (Env ms ps cmds defs) m = Env (m:ms) ps cmds defs

appendDefs :: Environment -> ModuleName -> [DataDef] -> Environment
appendDefs (Env ms ps cmds defs) name defs' = Env ms ps cmds ((map (\x -> (name,x)) defs') ++ defs)

addCommand :: Environment -> Command -> Environment
addCommand (Env ms ps cmds defs) c | trim c==[] = Env ms ps cmds defs
                                   | otherwise  =  Env ms ps (c:cmds) defs

isAlreadyLoaded :: Environment -> ModuleName -> Bool
isAlreadyLoaded (Env ms _ _ _) name = name `elem` (map fst' ms)

clearModules :: Environment -> Environment
clearModules (Env ms ps cmds _) = Env [] ps cmds []

runEnvironment :: Environment -> IO()
runEnvironment env = do showPrompt (modules env)
                        n <- myGetLine "" (commands env)
                        let c = reverse n in
                            do e <- executeCommand c (addCommand env c)
                               case e of
                                    Just env' -> runEnvironment env'
                                    Nothing   -> return ()

showPrompt :: LoadedModules -> IO()
showPrompt [] = putStr "\nNo module loaded> "
showPrompt ((n,_,_):_) = putStr ("\n" ++ n ++ "> ")

executeCommand :: String -> Environment -> IO(Maybe Environment)
executeCommand (':':xs) env = executeCommand' xs env
executeCommand []       env = return (Just env)
executeCommand  ys      env = do putStr ("\n\"" ++ ys ++ "\" is not a valid command.")
                                 return (Just env)

executeCommand' :: String -> Environment -> IO(Maybe Environment)
executeCommand' (x:x':xs) | isSpace x' || isControl x' = executeCharCommand (toLower x:x':xs)
                          | otherwise = executeWordCommand (toLower x: toLower x':xs)
executeCommand' (x:xs) = executeCharCommand (toLower x:xs)


executeCharCommand :: String -> Environment -> IO(Maybe Environment)
executeCharCommand ('h':xs) env = do putStr help
                                     return (Just env)
executeCharCommand ('q':xs) env = do putStr "Go Smalltalk!"
                                     return Nothing
executeCharCommand ('l':xs) env = do env' <- loadModule env (fillExtension (trim xs))
                                     return (Just env')
executeCharCommand ('r':xs) env = do env' <- reloadModule env (trim xs)
                                     return (Just env')
executeCharCommand ('d':xs) env = do env' <- deleteModule env (trim xs)
                                     return (Just env')
executeCharCommand ('c':xs) env = return (Just (clearModules env))

executeCharCommand ('a':xs) env = return (Just (addPath env (trim xs)))

executeCharCommand ('o':xs) env = return (Just (removePath env (trim xs)))

executeCharCommand ('m':xs) env = do showAllModules env
                                     return (Just env)
executeCharCommand ('u':xs) env = runScript env (trim xs)
                                  
executeCharCommand ('t':xs) env = fromExpression env (trim xs)
                                      (\e -> do {putStr e; return (Just env)})
                                      (\x -> do {doTypeInference x; return (Just env)})
executeCharCommand ('e':xs) env = fromExpression env (trim xs)
                                      (\e -> do {putStr e; return (Just env)})
                                      (\x -> do {showExpression x; return (Just env)})
executeCharCommand ('s':xs) env = do env' <- switchTo env (trim xs)
                                     return (Just env')
executeCharCommand ys       env = do putStr ('\"':(ys ++ "\" is not a valid command."))
                                     return (Just env)

executeWordCommand :: String -> Environment -> IO(Maybe Environment)
executeWordCommand ('t':'s':xs) env = commandShell env xs processPTS
executeWordCommand ('p':'s':xs) env = commandShell env xs processPS
executeWordCommand ('e':'e':xs) env = commandShell env xs processEE
executeWordCommand ys           env = do putStr ('\"':(ys ++ "\" is not a valid command."))
                                         return (Just env)

commandShell :: Environment -> String -> (SourceTerm -> IO()) -> IO(Maybe Environment)
commandShell env params f = fromExpression env (trim params)
                               (\e -> do {putStr e; putStr "\n"; return (Just env)})
                               (\x -> do {f x; return (Just env)})


myGetLine :: String -> [Command] -> IO String
myGetLine xs ys= do c <- getCh
                    if (c=='\t') 
                       then processTab xs ys
                       else
                       if ((xs==[] && (isBackSpace c)) || (ord c > 127))
                          then myGetLine xs ys
                          else
                            do putChar c
                               eraseIfNeeded c xs
                               if (c=='\n') then return xs
                                            else do cs <- myGetLine (processBackSpace (c:xs)) ys
                                                    return cs
                     where
                        eraseIfNeeded _ [] = return()
                        eraseIfNeeded c _ = if (isBackSpace c) then putStr [' ',c] else return()

processBackSpace :: String -> String
processBackSpace [x]       | isBackSpace x = []
processBackSpace (c:c':xs) | isBackSpace c = xs
processBackSpace xs                        = xs

processTab :: String -> [Command] -> IO String
processTab xs []     = myGetLine xs []
processTab xs (y:ys) = do putStr (map (const '\b') xs)
                          putStr y
                          clearLine xs y
                          myGetLine (reverse y) ys

clearLine :: String -> Command -> IO()
clearLine [] ys = return ()
clearLine xs [] = let (l,l') = unzip (map (const (' ','\b')) xs)
                      in do putStr l
                            putStr l'
clearLine (x:xs) (y:ys) = clearLine xs ys

loadModule :: Environment -> String -> IO(Environment)
loadModule env name | isAlreadyLoaded env (filename name) = do putStr ("Module \"" ++ name ++ "\" already loaded!")
                                                               return env    
                    | otherwise = do goAhead env name
                        where
                          goAhead env name | isAbsolute name = doLoadModule' env name
                                           | otherwise = tryAllPaths env (paths env) name

tryAllPaths :: Environment -> [FilePath] -> ModuleName -> IO(Environment)
tryAllPaths env []     name = doLoadModule' env name
tryAllPaths env (x:xs) name = do m <- doLoadModule env (processPath x name) (\x -> return())
                                 case m of
                                    Nothing   -> tryAllPaths env xs name
                                    Just env' -> return env'

doLoadModule :: Environment -> String -> (String -> IO()) -> IO(Maybe Environment)
doLoadModule env name f = do val <- safeReadFile name
                             case val of
                                Error   xs -> do f xs
                                                 return Nothing
                                Success xs -> do putStr ("Parsing " ++ filename name ++ "...\n")
                                                 case parsesourceexpr xs of
                                                       Success e ->  do {env'<-addModule env e name; return (Just env')}
                                                       error     ->  do {putStr (show error); return (Just env)}

addModule :: Environment -> SourceTerm -> String -> IO(Environment)
addModule env (Data ddefs st) name = let common = intersectBy cmp ddefs (defs env)
                                         in if (common == []) 
                                               then return (appendDefs (appendModule env (filename name,name,st)) (filename name) ddefs)
                                               else do putStr ("Type redefined : " ++ datadefName(head common) ++ "...\n")
                                                       return env 
                                         where cmp x y =  datadefName x == datadefName y

doLoadModule' :: Environment -> String -> IO(Environment)
doLoadModule' env name = do x <- doLoadModule env name putStr
                            case x of
                                Nothing -> return env
                                Just env' -> return env'

reloadModule :: Environment -> String -> IO(Environment)
reloadModule env@(Env [] _ _ _) _  = do {putStr"No module available"; return env}
reloadModule (Env (x:xs) p cmds defs) [] = reloadModule' (Env xs p cmds defs) x
reloadModule env str = case getModule' env str of
                         Error msg -> do {putStr msg; return env}
                         Success (x,env') -> reloadModule' env' x

reloadModule' :: Environment -> LoadedModule -> IO(Environment)
reloadModule' env old@(name,path,st) = do x <- doLoadModule (removeDefs env name) path putStr
                                          case x of
                                               Nothing -> return (appendModule env old)
                                               Just env' -> if (modules env' == modules env) 
                                                                then return (appendModule env old)
                                                                else return env'

deleteModule :: Environment -> String -> IO(Environment)
deleteModule env@(Env [] _ _ _)       _   = do {putStr"No module available"; return env}
deleteModule (Env ((n,_,_):xs) p cmds defs) []  = return (Env xs p cmds (filter (\(x,_)->x/=n) defs))
deleteModule env@(Env xs p cmds defs) str = case getModule' env str of
                                                Error msg -> do {putStr msg; return env}
                                                Success (x,env') -> let (name,_,_) = x in
                                                                        return (Env (filter (/=x) xs) p cmds (filter (\(y,_)->name/=y) defs))

filename :: String -> String
filename xs = reverse (takeWhile (/=pathSep) (tail (dropWhile (/='.') (reverse xs))))

fillExtension :: String -> String
fillExtension xs = reverse (appendExtIfNeeded (reverse xs))
                where
                   appendExtIfNeeded name@(x:y:z:'.':xs) = name
                   appendExtIfNeeded name@(x:y:'.':xs)   = name
                   appendExtIfNeeded name@(x:'.':xs)     = name
                   appendExtIfNeeded name@('.':xs)       = "stp"  ++ name
                   appendExtIfNeeded name                = "stp." ++ name

processPath :: FilePath -> ModuleName -> FilePath
processPath path name = processPath' (reverse (splitPath path)) (splitPath name)

processPath' :: [String] -> [String] -> FilePath
processPath'  xs     []       = tail (foldr (\y ys -> ys ++ (pathSep:y)) [] xs)
processPath' (x:xs) ("..":ys) = processPath' xs     ys
processPath'  xs    (y:ys)    = processPath' (y:xs) ys


splitPath :: String -> [String]
splitPath ys@(x:xs) | x == pathSep = splitPath' xs
                    | otherwise    = splitPath' ys

splitPath'     :: String -> [String]
splitPath' ""   = []
splitPath' s    = let (l,s') = break (pathSep ==) s
                     in l : case s' of
                                 []      -> []
                                 (_:s'') -> splitPath' s''

safeReadFile :: String -> IO(Error String)
safeReadFile name = do val <- catch (readFile name) (\e -> return ('!':(show e)))
                       return(safeGetContents val)
                       where
                          safeGetContents ('!':xs) = Error xs
                          safeGetContents ys       = Success ys


safeReadScript :: [FilePath] -> String -> IO(Error String)
safeReadScript []     name = safeReadFile name
safeReadScript (x:xs) name = do m <- safeReadFile (processPath x name)
                                case m of
                                      Error   r  -> safeReadScript xs name
                                      ok         -> return ok

runScript :: Environment -> FilePath -> IO(Maybe Environment)
runScript env path = do val <- safeReadScript (paths env) path
                        case val of
                             Error   xs -> do putStr xs
                                              return (Just env)
                             Success xs -> do putStr ("Running script " ++ filename path ++ "...\n")
                                              runCommands (Just env) (lines xs)  

runCommands :: (Maybe Environment) -> [String] -> IO(Maybe Environment)
runCommands Nothing    _      = return Nothing
runCommands res        []     = return res
runCommands (Just env) (x:xs) = do e <- executeCommand x env
                                   runCommands e xs

isAbsolute :: String -> Bool
isAbsolute (_:':':xs) = True
isAbsolute  _         = False

showAllModules :: Environment -> IO()
showAllModules env = do putStr "Loaded modules:"
                        putStr (foldr (\x xs -> ('\n':'\t':x) ++ xs) [] (map fst' (modules env)))
                        return ()

isOneWord :: String -> Bool
isOneWord s = any isAlpha s && all (not . isSpace) s

fromExpression :: Environment -> String -> (String -> IO(Maybe Environment)) -> (SourceTerm -> IO(Maybe Environment)) -> IO(Maybe Environment)
fromExpression env param f g = case getModule env (trim param) of
                                     Success x -> g (Data (defs env) x)
                                     Error e -> if (isOneWord param) 
                                                   then f e
                                                   else parseExpression env param f g

parseExpression :: Environment -> String -> (String -> IO(Maybe Environment)) -> (SourceTerm -> IO(Maybe Environment)) -> IO(Maybe Environment)
parseExpression env param f g = case (parsesourceexpr param) of 
                                         Error   error -> f error
                                         Success exp   -> do putStr(ppsourceexpr (bindFreeVariables env exp (freeSV exp))); putStr "\n"
                                                             g (Data (defs env) (bindFreeVariables env exp (freeSV exp)))

bindFreeVariables :: Environment -> SourceTerm -> [String] -> SourceTerm
bindFreeVariables env (Data def st) fv = let sts  = map (\(x,y,z) -> (x,z)) (modules env)
                                             sts' = bindedModules env fv 
                                             in combineTerms st sts'

bindedModules :: Environment -> [String] -> [(String,SourceTerm)]
bindedModules env names = foldr available [] names
                          where available y xs= case getModule env y  of
                                                        Success exp -> (y,exp):xs
                                                        otherwise   -> xs

combineTerms :: SourceTerm -> [(String,SourceTerm)] -> SourceTerm
combineTerms exp sts = foldr buildLet exp sts

buildLet :: (String,SourceTerm) -> SourceTerm -> SourceTerm
buildLet (var,exp) st = Let Static [((var,STUnknown),exp)] st

combineDefs :: [DataDef] -> [(String,SourceTerm)] -> [DataDef]
combineDefs xs ys = foldr buildDef xs ys

buildDef :: (String,SourceTerm) -> [DataDef] -> [DataDef]
buildDef (_,(Data defs st)) xs = (xs `union` defs) \\ (xs `intersect` defs)
buildDef (_,_)              xs = xs


processPTS :: SourceTerm -> IO()
processPTS st = processSourceTerm st 
                                  False
                                  False
                                  False
                                  "PTS...\n"
                                  "Expression does not specialise.\n" 
                                  "Its principal type specialisation.\n----------------------------------\n"

processPS :: SourceTerm -> IO()
processPS st = processSourceTerm st 
                                 True
                                 False
                                 False
                                 "Predicate Solving...\n"
                                 "Expression can't be solved.\n" 
                                 "Solution(Predicate Solving).\n----------------------------\n"

processEE :: SourceTerm -> IO()
processEE st = processSourceTerm st 
                                 True
                                 True
                                 False
                                 "Evidence Elimination...\n"
                                 "The evidence can't be eliminated.\n" 
                                 "Solution(Evidence Elimination).\n-------------------------------\n"

processAR :: SourceTerm -> IO()
processAR st = processSourceTerm st 
                                 True
                                 True
                                 True
                                 "Arity Raising...\n"
                                 "Can't do Arity Raising.\n" 
                                 "Solution(Arity Raising).\n------------------------\n"

processSourceTerm :: SourceTerm -> Bool -> Bool -> Bool -> String -> String -> String -> IO()
processSourceTerm st doTS doEE doAR title errorMsg okMsg = 
 do x <- doTypeInference st
    case x of
       Nothing -> return ()
       Just ((te,tau),source_state) ->
        do  
          putStr ('\n':title)
          case specialisesource source_state doTS doEE doAR (te,tau) of
               Error msg -> do putStr ('\n':errorMsg)
                               putStr msg
                               return ()
               Success (e', sigma) -> do putStr ('\n':okMsg)
                                         putStr (ppresidualterm e')
                                         putStr ":: \n"
                                         putStr (ppresidualscheme sigma)
                                         return ()


doTypeInference :: SourceTerm -> IO(Maybe (TypedSourceTerm, SourceState))
doTypeInference st = case typechecksource st of
                        Error msg -> do putStr "Can't infer a type for the source program!\n"
                                        putStr msg
                                        return (Nothing )
                        Success (st'@((te, tau), source_state)) -> 
                                     do putStr "Source program.\n---------------\n"
                                        putStr (ppsourceexpr te)
                                        putStr "::\n"
                                        putStr (ppsourcetype tau)
                                        return (Just st')

getModule :: Environment -> String -> (Error SourceTerm)
getModule (Env  []    _ _ _) _  = Error "No module available"
getModule (Env (x:xs) _ _ _) [] = Success (trd' x)
getModule (Env  xs    _ _ _) n  = case filter ((== n) . fst') xs of
                                        []  -> Error ("There is no module named \"" ++ n ++ "\"")
                                        [x] -> Success (trd' x)


trim :: String -> String
trim xs = foldr backSpaces [] (dropWhile isSpace xs)
            where
               backSpaces ' ' [] = []
               backSpaces  x  xs = x:xs

showExpression :: SourceTerm -> IO()
showExpression t = putStr (ppsourceexpr t)

switchTo :: Environment -> ModuleName -> IO(Environment)
switchTo env name = case getModule' env name of
                        Error msg -> do {putStr msg; return env}
                        Success (x,Env xs p cmds defs) -> return (Env (x:xs) p cmds defs)

superFilter :: (a -> Bool) -> [a] -> ([a],[a])
superFilter _ []     = ([],[])
superFilter p (x:xs) = let (a,b) = superFilter p xs in
                            if (p x) then (x:a,b)
                                     else (a,x:b)   

getModule' :: Environment -> ModuleName -> Error (LoadedModule,Environment)
getModule' env@(Env m p cmds defs) name = case superFilter ((== name) . fst') m of
                                             ([],xs)  -> Error("There is no module named \"" ++ name ++ "\"")
                                             ([x],xs) -> Success (x,Env xs p cmds defs)

fst' :: (a,b,c) -> a
fst' (x,_,_) = x

snd' :: (a,b,c) -> b
snd' (_,y,_) = y

trd' :: (a,b,c) -> c
trd' (_,_,z) = z

help = "Valid commands: \n\t" ++
       ":h               - This Help \n\t" ++
       ":q               - Quit \n\t" ++
       ":a <path>        - Add a new default path \n\t" ++
       ":o <path>        - RemOve this path from the default path \n\t" ++
       ":u <script name> - RUn a script \n\t" ++
       ":l <module name> - Load a module \n\t" ++
       ":r [module name] - Reload the module \n\t" ++
       ":d [module name] - Delete the module \n\t" ++
       ":c               - Clear all modules \n\t" ++
       ":m               - Show all loaded Modules \n\t" ++
       ":s <module name> - Switch to another module \n\t" ++
       ":e [module name] - Show the module's Expression \n\t" ++
       ":t  [expression] - Infer the Type of the module's expression \n\t" ++
       ":ts [expression] - Type Specialisation! \n\t" ++
       ":ps [expression] - Predicate Solving \n\t" ++
       ":ee [expression] - Evidence Elimination \n\t" ++
       ":ar [expression] - Arity Raising \n"

