module SourceLang 
   (
     SourceTermVble, SourceTerm(..), TypedSourceTerm, isVar
   , Annot(..), showannot, tokenStatic, tokenDynamic
   , ArithOp(..), BoolOp(..), RelOp(..)
   , shownop, showbop, showrop, arithop, boolop, relop
   , SourceTypeVble, SourceType(..), STBase(..), isSTVar
   , sTypeInt, sTypeBool, FSTV(..), FSV(..)
   , stripEtas                          
   , DataDef(..), dataOf 
   , constrsOf -- Add by Alejandro                   
   , datadefName, datadefParams         
   , patternMainConstr, splitConstrType 
   )
 where

import List(union, (\\))
import BasicData

-- ------------------------------------
-- Annotations
-- ------------------------------------
data Annot = AVar Int | Static | Dynamic 
             deriving (Eq, Read, Show)

showannot :: Annot -> String
showannot (AVar i) = "" -- "^%"++show i 
showannot Dynamic  = "" -- "^D" -- = "_"
showannot Static   = "^S"           -- = "@"

tokenStatic  = "^S"
tokenDynamic = "^D"

-- ------------------------------------
-- Terms
-- ------------------------------------
type SourceTermVble  = String
type TypedSourceTerm = (SourceTerm, SourceType)
data SourceTerm      = 
    Var SourceTermVble
  -- Functions
  | App Annot SourceType SourceTerm SourceTerm   -- e1 @ (e2::t)
  | Lam Annot SourceTermVble SourceType 
              SourceTerm                         -- \x::t -> e
  | Fix Annot SourceTerm                         -- fix e
  | Let Annot 
        [ ((SourceTermVble, SourceType), SourceTerm) ]
        SourceTerm                               -- let^ann xi::ti = ei in e
  -- Basic types: Int
  | Int Annot Int                                -- n^ann
  | Lift SourceTerm                              -- lift e
  | ArithOp Annot ArithOp SourceTerm SourceTerm  -- e1 nop^ann e2
  -- Basic types: Bool
  | Bool Annot Bool                              -- b^ann
  | If Annot SourceTerm SourceTerm SourceTerm    -- if^ann e then e1 else e2
  | BoolOp Annot BoolOp SourceTerm SourceTerm    -- e1 bop^ann e2
  | RelOp Annot RelOp SourceType 
          SourceTerm SourceTerm                  -- (e1::t) rop^ann (e2::t)
  -- Tuples
  | Tuple Annot [ SourceTerm ]                   -- (e1,..,en)^ann
  | Proj Annot Int Int SourceType SourceTerm     -- Proj_i_n^ann (e::t)
  -- Polyvariance
  | Poly SourceTerm                              -- poly e
  | Spec SourceTerm                              -- spec e
  | Polymorph SourceTerm                         -- polym e
  -- Explicit typing
  | Typed SourceTerm SourceType                  -- e::t
   -- Datatypes                                          
  | Data [DataDef] SourceTerm                            
    -- data ^annot D T_1 T_2 = C1 | C2 T_1 | C3 T_1 T_2  
  | Construc Annot ConstructorName                       
             [(SourceTerm,SourceType)]                   
  | Case Annot SourceTerm SourceType                     
         [((SourceTerm, SourceType),SourceTerm)]         
    -- case x of ...                                     
  deriving (Eq,Show)

isVar :: SourceTerm -> Bool
isVar (Var _) = True
isVar _       = False

data ArithOp = Add | Sub | Mul | Exp
             deriving (Eq, Read, Show)
data BoolOp = Or | And 
            deriving (Eq, Read, Show)
data RelOp = Eq | Neq | Gt | Geq | Lt | Leq
           deriving (Eq, Read, Show)

shownop Add = "+"
shownop Sub = "-"
shownop Mul = "*"
shownop Exp = "^"

showbop Or  = "||"
showbop And = "&&"

showrop Eq  = "=="
showrop Neq = "/="
showrop Gt  = ">"
showrop Geq = ">="
showrop Lt  = "<"
showrop Leq = "<="

arithop Add = (+)
arithop Sub = (-)
arithop Mul = (*)
arithop Exp = (^)

boolop  Or  = (||)
boolop  And = (&&)

relop   Eq  = (==)
relop   Neq = (/=)
relop   Gt  = (>)
relop   Geq = (>=)
relop   Lt  = (<)
relop   Leq = (<=)

-- ------------------------------------
-- Types
-- ------------------------------------
type SourceTypeVble = Int

data STBase = SBVar Int | SBInt | SBBool
        deriving (Read, Show, Eq)

data SourceType = 
    STUnknown
  | STVar       SourceTypeVble    
  | STBase      Annot STBase
  | STTuple     Annot [ SourceType ]
  | STFun       Annot SourceType SourceType
  | STPoly      SourceType 
  | STData      Annot DataName [SourceType]
  deriving (Read, Show,Eq)

sTypeBool ann = STBase ann SBBool
sTypeInt  ann = STBase ann SBInt

isSTVar :: SourceType -> Bool
isSTVar (STVar _) = True
isSTVar _         = False

-----------------------------------
-- Free type variables 
-----------------------------------
class FSTV t where
  freeSTV  :: t -> [SourceTypeVble]
  freeSTV _ = []

instance FSTV SourceType where
  freeSTV (STVar i)         = [i]
  freeSTV (STFun _ a b)     = freeSTV a ++ freeSTV b
  freeSTV (STPoly a)        = freeSTV a
  freeSTV (STTuple _ ts)    = concat (map freeSTV ts)
  freeSTV (STData _ _ ts)   = concat (map freeSTV ts)    
  freeSTV _                 = []

instance FSTV SourceTerm where
  freeSTV (App _ ty e1 e2)     = freeSTV ty ++ freeSTV e1 ++ freeSTV e2
  freeSTV (Lam _ _ mt e)       = freeSTV mt ++ freeSTV e
  freeSTV (Fix _ e)            = freeSTV e
  freeSTV (Let _ ds e)         = freeSTV ds ++ freeSTV e
  freeSTV (Lift e)             = freeSTV e
  freeSTV (ArithOp _ _ e1 e2)  = freeSTV e1 ++ freeSTV e2
  freeSTV (If _ e e1 e2)       = freeSTV e  ++ freeSTV e1 ++ freeSTV e2
  freeSTV (BoolOp _ _ e1 e2)   = freeSTV e1 ++ freeSTV e2
  freeSTV (RelOp _ _ ty e1 e2) = freeSTV ty ++ freeSTV e1 ++ freeSTV e2
  freeSTV (Tuple _ es)         = freeSTV es
  freeSTV (Proj _ _ _ ty e)    = freeSTV ty ++ freeSTV e
  freeSTV (Poly e)             = freeSTV e
  freeSTV (Spec e)             = freeSTV e
  freeSTV (Polymorph e)        = freeSTV e
  freeSTV (Typed e t)          = freeSTV t  ++ freeSTV e
  freeSTV (Data _ e)           = freeSTV e                                
  freeSTV (Construc _ c args)  = freeSTV args                             
  freeSTV (Case _ e t es)      = freeSTV e  ++ freeSTV t ++ freeSTV es    
  freeSTV _                    = []

-----------------------------------
-- Auxiliaries
instance FSTV t => FSTV [t] where
  freeSTV []                = []
  freeSTV (t:ts)            = freeSTV t ++ freeSTV ts

instance (FSTV a, FSTV b) => FSTV (a,b) where
  freeSTV (a,b)             = freeSTV a ++ freeSTV b

instance FSTV Char

-----------------------------------
-- Free term variables 
-----------------------------------
class FSV e where
  freeSV  :: e -> [SourceTermVble]
  freeSV _ = []

instance FSV SourceTerm where
  freeSV (Var x)              = [x]
  freeSV (App _ _ e1 e2)      = freeSV e1 `union` freeSV e2
  freeSV (Lam _ x _ e)        = freeSV e \\ [x]
  freeSV (Fix _ e)            = freeSV e
-- Modified by @  
-- @@@@@@@@@@@@@@@@@@@@@@@@@@ 
  freeSV (Let _ ds e)         = (freeSV (map snd ds) `union` freeSV e) \\ (map (fst . fst) ds)
-- @@@@@@@@@@@@@@@@@@@@@@@@@
  freeSV (Lift e)             = freeSV e
  freeSV (ArithOp _ _ e1 e2)  = freeSV e1 `union` freeSV e2
  freeSV (If _ e e1 e2)       = freeSV e  `union` freeSV e1 `union` freeSV e2
  freeSV (BoolOp _ _ e1 e2)   = freeSV e1 `union` freeSV e2
  freeSV (RelOp _ _ _ e1 e2)  = freeSV e1 `union` freeSV e2
  freeSV (Tuple _ es)         = freeSV es
  freeSV (Proj _ _ _ _ e)     = freeSV e
  freeSV (Poly e)             = freeSV e
  freeSV (Spec e)             = freeSV e
  freeSV (Polymorph e)        = freeSV e
  freeSV (Typed e _)          = freeSV e
  freeSV (Data _ e)           = freeSV e                 
  freeSV (Construc _ c args)  = freeSV args              
  freeSV (Case _ e _ es)      = freeSV e ++ freeSV es    
  freeSV _                    = []

-----------------------------------
-- Auxiliaries
instance FSV e => FSV [e] where
  freeSV []                = []
  freeSV (e:es)            = freeSV e `union` freeSV es

instance (FSV a, FSV b) => FSV (a,b) where
  freeSV (a,b)             = freeSV a `union` freeSV b

instance FSV Char
instance FSV SourceType

-----------------------------------
-- Eta contraction for constructors
-----------------------------------
class EtaStrippable t where
  stripEtas :: t -> t
  stripEtas t = t

instance EtaStrippable SourceTerm where
  stripEtas (App ann t e1 e2)      = App ann t (stripEtas e1) (stripEtas e2)
  stripEtas e@(Lam _ _ _ _)        = let (xts, e') = stripEta [] e
                                      in foldr (\(a,x,t) e'' -> Lam a x t e'')
                                               (stripEtas e') xts
  stripEtas (Fix ann e)            = Fix ann (stripEtas e)
  stripEtas (Let ann ds e)         = Let ann (stripEtas ds) (stripEtas e)
  stripEtas (Lift e)               = Lift (stripEtas e)
  stripEtas (ArithOp ann o e1 e2)  = ArithOp ann o (stripEtas e1) (stripEtas e2)
  stripEtas (If ann e e1 e2)       = If ann (stripEtas e) (stripEtas e1) 
                                                          (stripEtas e2)
  stripEtas (BoolOp ann o e1 e2)   = BoolOp ann o (stripEtas e1) (stripEtas e2)
  stripEtas (RelOp ann o t e1 e2)  = RelOp ann o t (stripEtas e1) (stripEtas e2)
  stripEtas (Tuple ann es)         = Tuple ann (map stripEtas es)
  stripEtas (Proj ann i n t e)     = Proj ann i n t (stripEtas e)
  stripEtas (Poly e)               = Poly (stripEtas e)
  stripEtas (Spec e)               = Spec (stripEtas e)
  stripEtas (Polymorph e)          = Polymorph (stripEtas e)
  stripEtas (Typed e t)            = Typed (stripEtas e) t
  stripEtas (Data ddefs e)         = Data ddefs (stripEtas e)
  stripEtas (Construc ann c es)    = Construc ann c (stripEtas es)
  stripEtas (Case ann e t es)      = Case ann (stripEtas e) t (stripEtas es)
  stripEtas e                      = e
  
stripEta xts (Lam ann x t e) = stripEta ((ann,x,t):xts) e
stripEta xts e@(Construc ann c args) =
  let n = length args
      m = length xts
      (xts',etas) = splitAt (m-n) xts
   in if map (\(a,x,t)->Var x) (reverse etas) == map fst args
       then (reverse xts', Construc ann c [])
       else (reverse xts, e)
stripEta xts e = (reverse xts, e)

-----------------------------------
-- Auxiliaries
instance EtaStrippable e => EtaStrippable [e] where
  stripEtas []     = []
  stripEtas (e:es) = stripEtas e : stripEtas es

instance (EtaStrippable a, EtaStrippable b) => EtaStrippable (a,b) where
  stripEtas (a,b)  = (stripEtas a, stripEtas b)

instance EtaStrippable Char
instance EtaStrippable SourceType

-----------------------------------
-- Datatypes
-----------------------------------
data DataDef = DataDef Annot DataName [SourceType] [Constructor SourceType]  
               deriving (Eq, Read, Show)                  
                                               
datadefName (DataDef _ dname _ _) = dname
datadefParams (DataDef _ _ params _) = params

dataOf _ [] = error "Cannot happen! Have you messed with datatypes?"
dataOf cname (ddef@(DataDef _ _ _ cs):ddefs) =
  case (constrArgsOf cname cs) of
    Nothing   -> dataOf cname ddefs
    Just args -> (ddef, args)

------------------------
-- Begin Alejandro -----
------------------------
constrsOf _ [] = error "Cannot happen! Have you messed with datatypes?"
constrsOf dname (ddef@(DataDef _ d _ cs):ddefs) 
          | d == dname = map constrName cs 
          | otherwise  = constrsOf dname ddefs
            where constrName (Constructor name _) = name
------------------------
---- End Alejandro -----
------------------------

patternMainConstr (Construc _ c []) = c
patternMainConstr (App _ _ e1 e2) = 
  if isVar e2
   then patternMainConstr e1
   else error "Cannot happen! Does source typecheck failed?"
patternMainConstr _ = error "Cannot happen! Does source typecheck failed?"

splitConstrType d@(STData _ _ _) = ([],d)
splitConstrType (STFun _ t1 t2) = let (args,d) = splitConstrType t2
                                   in (t1:args,d)
splitConstrType _ = error ("Cannot happen! Constructor types are formed by"
                           ++ " funtions into a data.")

