module SourceLangPrint(ppsourceexpr, ppsourcetype, pptypefact) where

import Char
import Maybe
import MyPrelude
import PrettyPrinter
import SourceLang 
import BasicData

infixl 9 `xannot`

----------------------------
-- Printing expressions
----------------------------
ppsourceexpr :: SourceTerm -> String
ppsourceexpr e = pretty 80 60 (ppexp (stripEtas e)) 

ppsourcetype :: SourceType -> String
ppsourcetype t = pretty 80 60 (pptypeexp t)

----------------------------
-- Pretty printing expressions
----------------------------
pptexp (Typed e t) = paren (ppexp e <> pptypesig t)
pptexp e           = ppexp e

ppexp :: SourceTerm -> Doc
ppexp e@(Lam _ _ _ _) = pplam [] e
ppexp (Let ann decls e@(Let _ _ _ )) = 
  (("let" `xannot` ann) <> ppdecls decls)
  $$
  (nest 1 (text"in"))
  $$
  (ppexp e)
ppexp (Let ann decls e) = 
  sep [ ("let" `xannot` ann) <> ppdecls decls
      , nest 1 (text "in ") <> ppexp e
      ]
ppexp (If ann e e1 e2) = 
  (("if" `xannot` ann) <+> ppexp e)
  $$
  (nest 1 (text "then ") <> ppexp e1)
  $$ 
  (nest 1 (text "else ") <> ppexp e2)

ppexp (Data [] sr) = ppexp sr                       
ppexp (Data xs sr) = (ppdataexps xs) $$ (ppexp sr) 

ppexp e = ppbexp e

-------------------------------
pplam xts (Lam ann x t e) = pplam ((ann,x,t):xts) e
pplam xts e           = sep [ ppbounds (reverse xts) 
                            , nest 1 (ppexp e)
                            ]

ppbounds []          = text ""
ppbounds [(ann,x,t)]     =
  ("\\" `xannot` ann) <> ppsourcetermvble x t <+> text "->"
ppbounds ((ann,x,t):xts) = 
  ("\\" `xannot` ann) <> ppsourcetermvble x t <+> text "->" <+> ppbounds xts

ppsourcetermvble x t            = text x <> pptypesig t
 
-------------------------------
ppbexp (BoolOp ann Or e1 e2)    = ppbterm e1 <+> ppboolop ann Or 
                                             <+> ppbexp e2
ppbexp e                        = ppbterm e

-------------------------------
ppbterm (BoolOp ann And e1 e2)  = ppbfact e1 <+> ppboolop ann And 
                                             <+> ppbterm e2
ppbterm e                       = ppbfact e

-------------------------------
ppbfact (RelOp ann rop _ e1 e2) = ppnexp e1  <+> pprelop ann rop 
                                             <+> ppnexp e2
ppbfact e                       = ppnexp e

-------------------------------
ppnexp (ArithOp ann nop e1 e2)  | nop==Add || nop==Sub
                                = ppnexp e1  <+> pparithop ann nop 
                                             <+> ppnterm e2
ppnexp e                        = ppnterm e

-------------------------------
ppnterm (ArithOp ann Mul e1 e2) = ppnterm e1 <+> pparithop ann Mul 
                                             <+> ppnfact e2
ppnterm e                       = ppnfact e

-------------------------------
ppnfact (ArithOp ann Exp e1 e2) = ppnfact e1 <+> pparithop ann Exp 
                                             <+> ppfexp e2
ppnfact e                       = ppfexp e

-------------------------------
ppfexp (App ann _ e1 e2) = 
                       sep [ ppfexp e1
                           , nest 2 (("@" `xannot` ann) <+> ppprim e2)
                           ]
ppfexp e             = ppprim e

-------------------------------
ppprim (Poly e)           = pppr "poly" e
ppprim (Spec e)           = pppr "spec" e
ppprim (Polymorph e)      = pppr "polym" e
ppprim (Lift e)           = pppr "lift" e
ppprim (Fix ann e)        = ppprann ann "fix" e
ppprim (Proj ann i n _ e) = ppproj ann i n e
ppprim e                  = ppaexp e

ppaexp (Var s)             = ppsourcetermvble s STUnknown
ppaexp (Tuple Static [])   = text "*"
ppaexp (Tuple _ [])        = text "()"
ppaexp (Tuple ann (e1:es)) = sep (   [ text "(" <+> pptexp e1 ] 
                                  ++ map (\e -> text "," <+> pptexp e) es
                                  ++ [ ")" `xannot` ann ]
                                 )
ppaexp (Int ann n)         = show n `xannot` ann
ppaexp (Bool ann b)        = show b `xannot` ann
ppaexp et@(Typed _ _)      = paren (pptexp et)
ppaexp (Construc ann c xs) = ppargs (c `xannot` ann) 
                                    (\(e,_) -> ppaexp e) xs    
ppaexp e@(Case _ _ _ _)    = ppcaseStatement e                
ppaexp e                   = paren (ppexp e)

-----------
--begin @--
-----------
ppcaseStatement (Case ann st _ xs) =
   (("case" `xannot`  ann) <+> ppaexp st <+> text "of" )
   $$
   (nest 1 (ppcasestats xs))
   $$ text "esac"
---------
--end @--
---------

pppr        s e = text s <+> ppaexp e
ppprann ann s e = (s `xannot` ann) <+> ppaexp e

-----------
--begin @--
-----------
ppargs d pp [] = d
ppargs d pp xs = d <+> foldr1 (<+>) (map pp xs)

ppmany pp [] = text ""
ppmany pp [x] = pp x
ppmany pp (x:xs) =  (pp x) <+> (ppmany pp xs)

ppdataexps es = foldr ($$) (text "") (map ppdataexp es)

ppdataexp (DataDef ann name sts cs) = text "data" 
                                      <+> 
                                      (ppargs (name `xannot` ann) 
                                              pptypefact sts)
                                      <+>
                                      (text "=")
                                      <+> 
                                      (ppdataconstructors cs)
                                      <> text ";"

ppdataconstructors [] = text ""
ppdataconstructors [c] = ppdataconstructor c
ppdataconstructors (c:cs) 
   = sep (ppdataconstructor c :
          map (\c -> nest (-2) (text "|" <+> ppdataconstructor c)) cs)

ppdataconstructor (Constructor name sts)   
   = ppargs (text name) pptypefact sts

ppcasestats []  = text ""
ppcasestats [c] = ppcasestat c <> text ";"
ppcasestats (c:cs) = (ppcasestat c <> text ";")
                     $$ ppcasestats cs

ppcasestat ((st1,_),st2) = sep [ ppfexp st1 
                               , nest 2 ((text "->") <+> (ppexp st2))
                               ]
---------
--end @--
---------
----------------------------
-- Printing source types
----------------------------
pptypesig STUnknown = text ""
pptypesig t         = text "" -- text "::" <> pptypeexp t

pptypeexp (STFun ann t1 t2) = 
  sep [ pptypeprefact t1
      , nest 1 ("->" `xannot` ann)
      , nest 1 (pptypeexp t2)
      ]
pptypeexp t = pptypeprefact t

pptypeprefact (STData ann n xs@(_:_)) =                                 
     ppargs (n `xannot` ann) pptypefact xs                             
pptypeprefact (STPoly t)             = text "poly" <+> pptypefact t
pptypeprefact t                 = pptypefact t

pptypefact (STVar t)              = text "T_" <> text (show t)        
pptypefact (STBase ann b)         = showbase b `xannot` ann
--pptypefact (STPoly t)             = text "poly" <+> pptypefact t
pptypefact (STTuple _ [])         = text "()"
pptypefact (STTuple ann (t1:ts))  = sep (   [ text "(" <+> pptypeexp t1]
                                         ++ map (\t -> text "," 
                                                       <+> pptypeexp t) ts
                                         ++ [ ")" `xannot` ann ]
                                        )
pptypefact (STData ann n [])      = n `xannot` ann                   
pptypefact t                      = paren (pptypeexp t)


----------------------------
-- Printing definitions
----------------------------

s `xannot` a = if (showannot a == "")                           
                  then text s                                  
                  else text s <> (text (showannot a))         

showbase (SBVar b) = "Base_"++show b
showbase SBInt     = "Int"
showbase SBBool    = "Bool"

pparithop ann nop = (shownop nop) `xannot` ann
ppboolop  ann bop = (showbop bop) `xannot` ann
pprelop   ann rop = (showrop rop) `xannot` ann

ppproj ann 1 2 e = ppprann ann "fst" e
ppproj ann 2 2 e = ppprann ann "snd" e
ppproj ann i n e = ppprann ann ("Proj_"++show i++"_"++show n) e

ppdecls []     = text "" 
ppdecls [d]    = text " " <> ppdecl d
ppdecls (d:ds) = (text " " <> ppdecl d)
                 $$
                 foldr1 ($$) (map (\d -> nest (-1) (text ";" <+> ppdecl d)) ds)
                     
ppdecl ((x,t), e) = sep [ ppsourcetermvble x t <+> text "="
                        , nest 2 (ppexp e)
                        ]
        
paren d = sep [ text "(" <+> d, text ")" ]

