module SourceTypeCheck(typechecksource) where
import MyPrelude
import Error
import FMap
import Table
import AVL
import MoreMonad
import List (intersect,(\\))

import BasicData
import SourceLang
import SourceLangPrint
import SourceTypeCheckMonad
import SourceUnification
import SourceDataManagement

type Environment = [(SourceTermVble,SourceType)]

typechecksource :: SourceTerm -> Error (TypedSourceTerm, SourceState)
typechecksource e =
  let ftv = freeSTV e
      f = do setexcepts stype ftv
             i_tt_st <- tc [] e
             tt_st <- unfoldVbles True i_tt_st 
             case tt_st of
               (Data ddefs e, t) -> do ddefs' <- expandAnnotsFor ddefs
                                       return (Data ddefs' e,t)
               _ -> return tt_st
   in case (unSTCM f initialSourceState) of
        Error e                       -> Error e
        Success (tt_st, source_state) -> Success (tt_st, source_state)

tc :: Environment -> SourceTerm -> STCM TypedSourceTerm
tc g (Var x) = 
   case (assocM g x) of
     Success t -> return (Var x, t)
     Error e   -> fail e

tc g (App ann _ e1 e2) = 
   do (te1, t1) <- tc g e1
      (te2, t2) <- tc g e2
      t <- freshSType
--      error ("t1 = " ++ show t1 ++ "; t2 = " ++ show t2)
      ann' <- freshAnnotBasedOn ann
      unify t1 (STFun ann' t2 t)
      return (App ann' t2 te1 te2, t)

tc g (Lam ann x t' e) = 
   do t <- (case t' of
             STUnknown -> freshSType
             _         -> return t')
      (te, t1) <- tc ((x,t):g) e 
      ann' <- freshAnnotBasedOn ann
      return (Lam ann' x t te, STFun ann' t t1)

tc g (Fix ann e) = 
   do (te, t) <- tc g e
      t1      <- freshSType
      ann'    <- freshAnnotBasedOn ann
      unify t (STFun ann' t1 t1)
      case ann of
        Static -> do t2      <- freshSType
                     t3      <- freshSType
                     unify t1 (STFun ann' t2 t3)
        _      -> return ()
      return (Fix ann' te, t1)

tc g (Let ann decls e) = 
   do (tdecls, txs) <- tcdecls g decls 
      (te, t)       <- tc (txs++g) e 
      ann'          <- freshAnnotBasedOn ann
      return (Let ann' tdecls te, t)

tc _ (Int ann n)   = do ann' <- freshAnnotBasedOn ann
                        return (Int ann' n, sTypeInt ann')

tc g (Lift e) = 
   do (te, t) <- tc g e
      bt      <- freshSTBase
      unify t (STBase Static bt)
      return (Lift te, STBase Dynamic bt) 

tc g (ArithOp ann nop e1 e2) = 
   do (te1, t1) <- tc g e1
      (te2, t2) <- tc g e2
      ann'      <- freshAnnotBasedOn ann
      unify t1 (sTypeInt ann')
      unify t2 (sTypeInt ann')
      return (ArithOp ann' nop te1 te2, sTypeInt ann')

tc _ (Bool ann b)  = do ann' <- freshAnnotBasedOn ann
                        return (Bool ann' b, sTypeBool ann')

tc g (If ann e e1 e2) = 
   do (te, t)   <- tc g e
      ann'      <- freshAnnotBasedOn ann
      unify t (sTypeBool ann')
      (te1, t1) <- tc g e1
      (te2, t2) <- tc g e2
      unify t1 t2
      return (If ann' te te1 te2, t1)

tc g (BoolOp ann bop e1 e2) = 
   do (te1, t1) <- tc g e1
      (te2, t2) <- tc g e2
      ann'      <- freshAnnotBasedOn ann
      unify t1 (sTypeBool ann')
      unify t2 (sTypeBool ann')
      return (BoolOp ann' bop te1 te2, sTypeBool ann')

tc g (RelOp ann bop t e1 e2) = 
   do (te1, t1) <- tc g e1
      (te2, t2) <- tc g e2
      bt        <- freshSTBase
      ann'      <- freshAnnotBasedOn ann
      unify t1 (STBase ann' bt)
      unify t2 (STBase ann' bt)
      (case t of 
        STUnknown -> return ()
        _         -> unify t (STBase ann' bt))
      return (RelOp ann' bop t1 te1 te2, sTypeBool ann')

tc g (Tuple ann es) = 
   do (tes,ts) <- tcs g es 
      ann'     <- freshAnnotBasedOn ann
      return (Tuple ann' tes, STTuple ann' ts)

tc g (Proj ann i n _ e) = 
   do (te, t) <- tc g e
      ts      <- replicateM n freshSType
      ann'    <- freshAnnotBasedOn ann
      unify t (STTuple ann' ts)
      if i>length ts then error (outboundIn "Source type check (tc)") else do
      return (Proj ann' i n t te, ts!!(i-1))

tc g (Poly e) = 
   do (te, t) <- tc g e
      return (Poly te, STPoly t)

tc g (Spec e) = 
   do (te, t) <- tc g e
      t1      <- freshSType
      unify t (STPoly t1)
      return (Spec te, t1)

tc g (Polymorph e) = 
   do (te, t) <- tc g e
      return (Polymorph te, STPoly t)

tc g (Typed e t) =
   do (te, et) <- tc g e
      unify et t
      return (te, et)

-----------
--begin @--
-----------
tc g (Data dataDefs e) = 
   do dataDefs' <- tcDatas dataDefs
      (te, t) <- tc g e       
      return (Data dataDefs' te, t)

tc g (Construc ann name xs) = 
   do t     <- constructorType name
      ann'' <- freshAnnotBasedOn ann
      ann'  <- topAnnotation t
      unifyAnnot ann' ann''
      constr <- etaExpand ann' name xs t
      return (constr, t)

tc g (Case ann e _ cases) = 
   do (te, t) <- tc g e
      (cnames, tpatterns, texps, patterntypes, exptypes) <- procCases g cases
      let dname = head cnames
      DataDef _ _ _ constrs <- constructorDataType dname
      if not (null (cnames \\ map constrName constrs)) 
      -----------------------
      --- Begin Alejandro ---
      -----------------------
      {- Here, I expand the error message
         The previos code was:
       then fail "Non-disjoint patterns!" 
      -}
       then fail "Non-disjoint patterns or all constructors don't belong to the same datatype!" 
       else do unifyAll (t : patterntypes)
               unifyAll exptypes
               ann'  <- freshAnnotBasedOn ann
               ann'' <- topAnnotation t
               unifyAnnot ann' ann''
               let cases' = zip (zip tpatterns patterntypes) texps
               return (Case ann' te t cases',head exptypes)

tc _ e = error ("Case not covered: " ++ ppsourceexpr e)
---------
--end @--
---------

tcs g []     = return ([], [])
tcs g (e:es) = do (te,t)   <- tc g e
                  (tes,ts) <- tcs g es
                  return (te:tes, t:ts)

tcdecls g []  = return ([],[])
tcdecls g (d:ds) = do (td, tx)   <- tcdecl g d
                      (tds, txs) <- tcdecls g ds
                      return (td:tds, tx:txs)

tcdecl g ((x,t'),e) =
   do (te,t) <- tc g e
      (case t' of
        STUnknown -> return ()
        _         -> unify t t')
      return (((x,t),te), (x,t))

-----------
--begin @--
-----------
tcDatas :: [DataDef] -> STCM [DataDef]
tcDatas xs = 
  if (dataUsedIn xs `subseteq` dataNamesOf xs)
   then mapM tcData xs
   else fail ("Nonexistent constructor referenced in datatype definition")
  where
    dataNamesOf = map (\(DataDef _ name _ _ ) -> name) 
    
    dataUsedIn ds = concat [ dataUsedIn' t | DataDef _ _ _ cs <- ds
                                           , Constructor _ ts <- cs
                                           , t <- ts ] 
    
    dataUsedIn' (STTuple _ ts)     = concat (map dataUsedIn' ts)
    dataUsedIn' (STFun _ t1 t2)    = dataUsedIn' t1 ++ dataUsedIn' t2
    dataUsedIn' (STPoly t)         = dataUsedIn' t
    dataUsedIn' (STData _ name ts) = name : concat (map dataUsedIn' ts)
    dataUsedIn' _                  = []

tcData :: DataDef -> STCM DataDef
tcData (DataDef ann name sts cs) = 
  if (repeated sts)
   then fail ("Repeated type variables in datatype " ++ name)
   else case (foldr (checkConstructor sts) (Success()) cs) of
         Success _ -> do cs' <- freshAnnotsFor cs
                         let d = DataDef ann name sts cs'
                         addDataDef d
                         return d
         Error e   -> fail e
  where
    repeated []     = False
    repeated (x:xs) = (x `elem` xs) || repeated xs
    
    checkConstructor :: [SourceType] -> (Constructor SourceType) 
                        -> Error () -> Error ()
    checkConstructor _    _                    (Error e)  = Error e
    ----------------------
    -- Begin Alejandro ---
    ----------------------
    checkConstructor sts (Constructor name xs) (Success _)= 
      if ( freeVarsInConstr xs `subseteq` sts) 
       then Success ()
       else Error ("Free type variables in constructor " ++ name)
      where freeVarsInConstr = (map STVar).concat.(map freeSTV)  
   {- 
    -----------------------------------------------------------------
    The above code was modified by A. Russo in order to 
    fix an error. The original was
      if (filter isSTVar xs `subseteq` sts) 
       then Success ()
       else Error ("Free type variables in constructor " ++ name)

	See example data D T_1 = L (T_1, T_2)
     ----------------------------------------------------------------
    -}
    ----------------------
    -- End Alejandro -----
    ----------------------
    
----------------------------
-- Case alternatives
procCases g [] = fail "There should be some pattern!"
procCases g [c] =
   do (cname,  tp,  te,  pt,  et)  <- procCase g c
      return ([cname],[tp],[te],[pt],[et])
procCases g (c:cs) =
   do (cname,  tp,  te,  pt,  et)  <- procCase g c
      (cnames, tps, tes, pts, ets) <- procCases g cs
      return (cname:cnames,tp:tps,te:tes,pt:pts,et:ets)

procCase g ((pattern,_),exp) =
    do cname <- checkPattern pattern
       g' <- addCaseVariable g pattern
       (tpattern, patterntype) <- tc g' pattern
       (texp,     exptype)     <- tc g' exp
       return (cname, tpattern, texp, patterntype, exptype)

checkPattern (Construc _ c []) = return c
checkPattern (App _ _ e1 e2) = 
  if isVar e2
   then checkPattern e1
   else fail "Invalid pattern!"
checkPattern _ = fail "Invalid pattern!"


----------------------------
-- Patterns and environments
addCaseVariables :: Environment -> [SourceTerm] -> STCM Environment
addCaseVariables g []     = return g
addCaseVariables g (x:xs) = do g'  <- addCaseVariables g xs
                               g'' <- addCaseVariable g' x
                               return g''

addCaseVariable :: Environment -> SourceTerm -> STCM Environment
addCaseVariable g (Var i)         = do t <- freshSType
                                       return ((i,t) : g)
addCaseVariable g (App _ _ s1 s2) = do g1 <- addCaseVariable g s1
                                       addCaseVariable g1 s2
addCaseVariable g _               = return g

-------------------------------------
-- Eta expansion for constructors
etaExpand ann name [] datatype = etaExpand' 0 [] ann name datatype
-----------------------
--- Begin Alejandro ---
-----------------------
{- 
   We don't need this case because a constructor only appears
   in isolation.
   etaExpand ann name xs _        = return (Construc ann name xs)
-}
---------------------
--- End Alejandro ---
---------------------

etaExpand' i ts ann n (STData _ _ _) = 
   let xs' = foldr (\(y,t) ys -> (Var (etaName y),t):ys)
                   [] (zip [0..(i-1)] (reverse ts))
    in return (Construc ann n xs')
etaExpand' i ts ann n (STFun _ st st') = 
   let varName = etaName i
       var = Var varName           
    in do c <- etaExpand' (i+1) (st:ts) ann n st'
          return (Lam ann varName st c)

etaName i = "x_" ++ show i

-------------------------------------
-- Top annotation of a type
--  (no track of freshAnnot when t is a STVar...)
--   (it only appears in a statement "case e of x -> e' esac")
topAnnotation t = do t' <- unfoldVbles False t
                     case t' of
                      STBase ann _    -> return ann
                      STTuple ann _   -> return ann
                      STFun ann _ _   -> return ann
                      STPoly t''      -> topAnnotation t''
                      STData ann _ _  -> return ann 
                      _               -> freshAnnot
---------
--end @--
---------

