module SourceTypeCheckMonad where

import List (nub, intersect)
import MyPrelude
import Error
import FMap
import Table
import AVL
import MoreMonad

import MetaVbles
import SourceLang
import SourceLangPrint
import SourceDataManagement   
import BasicData

-- -------------------------------------------------------------
-- Monad to manage substitutions: a substitution is implicitly
--  build when you use the monad!
-- -------------------------------------------------------------
data STCM a = STCM (SourceState->Error(a,SourceState))
type SourceState = ( MetaVble SourceType      -- type assignments
                   , MetaVble Annot           -- annot assignments
                   , MetaVble STBase          -- base type assignments
                   , DataManager              -- datas            
                   )

initialSourceState :: SourceState
initialSourceState = ( initialmetavbles
                     , initialmetavbles
                     , initialmetavbles
                     , initialManager                             
                     )

stype    :: ST (MetaVble SourceType) a -> ST SourceState a
stype    f (st,ann,bt,datas) = let (i,st')  = f st                   
                                in (i, (st',ann ,bt ,datas))         

annot    :: ST (MetaVble Annot) a -> ST SourceState a
annot    f (st,ann,bt,datas) = let (i,ann') = f ann                  
                                in (i, (st ,ann',bt ,datas))         

basetype :: ST (MetaVble STBase) a -> ST SourceState a
basetype f (st,ann,bt,datas) = let (i,bt')  = f bt                   
                                in (i, (st ,ann ,bt' ,datas))        

unSTCM (STCM f) = f

instance Monad STCM where
  return x     = STCM (\s -> Success(x,s))
  STCM x >>= f = STCM (\s -> case x s of 
                            Error e         -> Error e
                            Success (v, s') -> unSTCM (f v) s'
                    )
  fail e       = STCM (\_ -> Error e)

setexcepts part e   = STCM (\s -> Success (part (setexceptsmetavble e) s))
fresh      part     = STCM (\s -> Success (part freshmetavble          s))
get        part x   = STCM (\s -> Success (part (getmetavble x)        s))
set        part x v = STCM (\s -> Success (part (setmetavble x v)      s))
getTable   part     = STCM (\s -> Success (part gettable               s))

remember   part v   = STCM (\s -> Success (part (remembermetavble v)   s))
forget     part v   = STCM (\s -> Success (part (forgetmetavble v)     s))
seen       part v   = STCM (\s -> Success (part (seenmetavble v)       s))
forgetAll  part     = STCM (\s -> Success (part forgetallmetavbles     s))
pushMemory part     = STCM (\s -> Success (part pushmemmetavble        s))
popMemory  part     = STCM (\s -> Success (part popmemmetavble         s))

freshSType  = liftM STVar (fresh stype)
freshAnnot  = liftM AVar  (fresh annot)
freshSTBase = liftM SBVar (fresh basetype)

freshAnnotBasedOn (AVar 0)   = freshAnnot
freshAnnotBasedOn ann        = return ann

-----------
--begin @--
-----------
getManager :: STCM DataManager
getManager = STCM (\s@(_,_,_,mngr)  -> Success (mngr,s))

addDataDef :: DataDef -> STCM ()
addDataDef ddef  = 
  STCM (\(st,an,bt,mngr) -> 
         let res = addData mngr ddef in
         case res of
           Error e -> Error e
           Success new_mngr 
                   -> Success ((),(st,an,bt,new_mngr))
       )

constructorDataType :: ConstructorName -> STCM DataDef
constructorDataType constr = 
   do mngr <- getManager
      case (datatypeOf mngr constr) of 
        Error msg -> fail msg
        Success (d, _) -> return d

constructorType :: ConstructorName -> STCM SourceType
constructorType constr = 
   do mngr <- getManager
      case (datatypeOf mngr constr) of 
        Error msg -> fail msg
        Success (DataDef ann name sts _, args) ->
         do ann'     <- freshAnnotBasedOn ann
             -- This works because ann is (AVar 0) or ground...
            replaces <- newTypesFor sts
            let cType = buildConstructorType ann' replaces name args
            return cType

newTypesFor [] = return []
newTypesFor (st:sts) = do nst  <- freshSType
                          nsts <- newTypesFor sts
                          return ((st,nst):nsts)
---------
--end @--
---------

---------------------
-- Begin Alejandro --
---------------------
constrsOfData :: DataName -> STCM [(ConstructorName, SourceType)]
constrsOfData dname = do mngr <- getManager
                         case (dataconstrs mngr dname) of 
                              Error msg  -> fail msg
                              Success cs -> constrsST cs

constrST :: ConstructorName -> STCM (ConstructorName, SourceType)
constrST cname = do t <- constructorType cname
                    return (cname, t)

constrsST :: [ConstructorName] -> STCM [(ConstructorName, SourceType)]
constrsST [] = return []
constrsST (cname:cs) = do (_ , st)  <- constrST cname 
                          rs <- constrsST cs 
                          return ((cname, st):rs)
---------------------
---- End Alejandro --
---------------------
-- %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
-- Unfolding variables for different kinds of objects
-- %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
class WithVbles t where
  isvble      :: t -> Bool
  indir       :: t -> STCM t
  unfoldVbles :: Bool -> t -> STCM t  -- May initialize auxiliary structures
  unfoldV     :: Bool -> t -> STCM t  -- Do NOT initialize auxiliary structures

  isvble  _       = False
  unfoldVbles _ t = return t
  indir         t = unfoldVbles False t
  unfoldV     b t = unfoldVbles b t

------------------------------
-- Unfolding for annotations
instance WithVbles Annot where
  isvble      (AVar _)    = True
  isvble      _           = False

  indir       vi@(AVar i) = do pa <- get annot i
                               case pa of
                                Yes a -> indir a
                                No _  -> return vi
  indir       t           = return t

  unfoldVbles final
              vi@(AVar i) = do pa <- get annot i
                               case pa of
                                Yes a -> unfoldVbles final a
                                No _  -> if final 
                                          then return Dynamic 
                                          -- No BTpolymorphism! So, any
                                          -- FAVar is considered to be Dynamic
                                          else return vi
  unfoldVbles _
              t           = return t

------------------------------
-- Unfolding for base types
instance WithVbles STBase where
  isvble      (SBVar _)    = True
  isvble      _            = False

  unfoldVbles final
              vi@(SBVar i) = do pa <- get basetype i
                                case pa of
                                 Yes a -> unfoldVbles final a
                                 No _  -> if final
                                           then return SBInt
                                           -- No overloading! So, any
                                           -- FSBVar is considered to be SBInt
                                           else return vi
  unfoldVbles _
              t            = return t

------------------------------
-- Unfolding for two-level types
instance WithVbles SourceType where
  isvble      (STVar _)         = True
  isvble      _                 = False
 
  indir       vi@(STVar i)      = do pa <- get stype i
                                     case pa of
                                      Yes a -> indir a
                                      No _  -> return vi
  indir       t                 = return t

  unfoldVbles final
              vi@(STVar i)      = do t <- get stype i
                                     case t of
                                      Yes t'-> unfoldVbles final t'
                                      No _  -> if final
                                                then return (sTypeInt Dynamic)
                                                -- No polymorphism! So, any
                                                -- FSTVar is considered to be Int
                                                else return vi
  unfoldVbles final
              (STBase ann b)    = liftM2 STBase  (unfoldVbles final ann)
                                                 (unfoldVbles final b) 
  unfoldVbles final
              (STTuple ann ts)  = liftM2 STTuple (unfoldVbles final ann)
                                                 (unfoldVbles final ts) 
  unfoldVbles final
              (STFun ann t1 t2) = liftM3 STFun   (unfoldVbles final ann)
                                                 (unfoldVbles final t1) 
                                                 (unfoldVbles final t2)
  unfoldVbles final
              (STPoly t)        = liftM STPoly   (unfoldVbles final t)

  unfoldVbles final                                                     
              (STData ann n xs) = liftM3 STData (unfoldVbles final ann) 
                                                (return n)              
                                                (unfoldVbles final xs)  
  unfoldVbles _                 
              t                 = return t
                     
-----------------------------------------
-- Unfolding for NoInformation, Maybe, and lists
instance WithVbles Char

instance WithVbles a => WithVbles [a] where
 unfoldVbles _     []     = return []
 unfoldVbles final (a:as) = liftM2 (:) (unfoldVbles final a) 
                                       (unfoldVbles final as)

instance (WithVbles a, WithVbles b) => WithVbles (a,b) where
 unfoldVbles final (a,b) = liftM2 (,) (unfoldVbles final a) 
                                      (unfoldVbles final b)

------------------------------
-- Unfolding for SourceTerms
instance WithVbles SourceTerm where
 isvble (Var _)      = True
 isvble _            = False

 unfoldVbles final e = 
  case e of
    (App iann it ie1 ie2)       -> do e1    <- unfoldVbles final ie1
                                      e2    <- unfoldVbles final ie2
                                      ann   <- unfoldVbles final iann
                                      t     <- unfoldVbles final it
                                      return (App ann t e1 e2)
    (Lam iann x i_t ie)         -> do t     <- unfoldVbles final i_t
                                      e     <- unfoldVbles final ie
                                      ann   <- unfoldVbles final iann
                                      return (Lam ann x t e)
    (Fix iann ie)               -> do e     <- unfoldVbles final ie
                                      ann   <- unfoldVbles final iann
                                      return (Fix ann e)
    (Let iann idecls ie)        -> do decls <- unfoldVbles final idecls
                                      e     <- unfoldVbles final ie
                                      ann   <- unfoldVbles final iann
                                      return (Let ann decls e)
    (Int iann n)                -> do ann   <- unfoldVbles final iann
                                      return (Int ann n)
    (Lift ie)                   -> do e     <- unfoldVbles final ie
                                      return (Lift e)
    (ArithOp iann nop ie1 ie2)  -> do e1    <- unfoldVbles final ie1
                                      e2    <- unfoldVbles final ie2
                                      ann   <- unfoldVbles final iann
                                      return (ArithOp ann nop e1 e2)
    (Bool iann b)               -> do ann   <- unfoldVbles final iann
                                      return (Bool ann b)
    (If iann ie ie1 ie2)        -> do e1    <- unfoldVbles final ie1
                                      e2    <- unfoldVbles final ie2
                                      e     <- unfoldVbles final ie
                                      ann   <- unfoldVbles final iann
                                      return (If ann e e1 e2)
    (BoolOp iann bop ie1 ie2)   -> do e1    <- unfoldVbles final ie1
                                      e2    <- unfoldVbles final ie2
                                      ann   <- unfoldVbles final iann
                                      return (BoolOp ann bop e1 e2)
    (RelOp iann rop it ie1 ie2) -> do e1    <- unfoldVbles final ie1
                                      e2    <- unfoldVbles final ie2
                                      t     <- unfoldVbles final it
                                      ann   <- unfoldVbles final iann
                                      return (RelOp ann rop t e1 e2)
    (Tuple iann ies)            -> do es    <- unfoldVbles final ies
                                      ann   <- unfoldVbles final iann
                                      return (Tuple ann es)
    (Proj iann i j it ie)       -> do e     <- unfoldVbles final ie
                                      t     <- unfoldVbles final it
                                      ann   <- unfoldVbles final iann
                                      return (Proj ann i j t e)
    (Poly ie)                   -> do e     <- unfoldVbles final ie
                                      return (Poly e)
    (Spec ie)                   -> do e     <- unfoldVbles final ie
                                      return (Spec e)
    (Polymorph ie)              -> do e     <- unfoldVbles final ie
                                      return (Polymorph e)
    (Typed ie it)               -> do e     <- unfoldVbles final ie
                                      t     <- unfoldVbles final it
                                      return (Typed e t)
    (Data xs st)                -> do st'   <- unfoldVbles final st     
                                      return (Data xs st')              
    (Construc iann name sts)    -> do sts'  <- unfoldVbles final sts    
                                      ann   <- unfoldVbles final iann   
                                      return (Construc ann name sts')   
    (Case iann st t xs)         -> do xs'   <- unfoldVbles final xs     
                                      st'   <- unfoldVbles final st     
                                      t'    <- unfoldVbles final t      
                                      ann   <- unfoldVbles final iann   
                                      return (Case ann st' t' xs')      
    _                           ->    return e
    
-----------
--begin @--
-----------
------------------------------                                                  
-- Unfolding Constructors
instance WithVbles a => WithVbles (Constructor a) where
 isvble _            = False

 unfoldVbles final (Constructor name xs) = 
    do xs' <- unfoldVbles final xs
       return (Constructor name xs')

------------------------------                                                  
-- Unfolding Data definitions
instance WithVbles DataDef where
 isvble _            = False

 unfoldVbles final (DataDef ann name sts cs) = 
    do sts' <- unfoldVbles final sts
       cs'  <- unfoldVbles final cs
       ann' <- unfoldVbles final ann
       return (DataDef ann' name sts' cs')

------------------------------
-- Traverse annots for
------------------------------
class TraverseAnnotsFor t where
  freshAnnotsFor    :: t -> STCM t 
  expandAnnotsFor   :: t -> STCM t 
  traverseAnnotsFor :: (Annot -> STCM Annot) -> t -> STCM t 

  freshAnnotsFor t  = traverseAnnotsFor freshAnnotBasedOn t
  expandAnnotsFor t = traverseAnnotsFor (unfoldVbles True) t
  traverseAnnotsFor _ t = return t

------------------------------
-- Traverse Annots for annotations
instance TraverseAnnotsFor Annot where
  traverseAnnotsFor f a = f a

------------------------------
-- Traverse Annots for base types
instance TraverseAnnotsFor STBase

------------------------------
-- Traverse Annots for two-level types
instance TraverseAnnotsFor SourceType where
  traverseAnnotsFor f (STBase ann b)    = liftM2 STBase      (traverseAnnotsFor f ann)
                                                             (traverseAnnotsFor f b) 
  traverseAnnotsFor f (STTuple ann ts)  = liftM2 STTuple     (traverseAnnotsFor f ann)
                                                             (traverseAnnotsFor f ts) 
  traverseAnnotsFor f (STFun ann t1 t2) = liftM3 STFun       (traverseAnnotsFor f ann)
                                                             (traverseAnnotsFor f t1) 
                                                             (traverseAnnotsFor f t2)
  traverseAnnotsFor f (STPoly t)        = liftM  STPoly      (traverseAnnotsFor f t)
  traverseAnnotsFor f (STData ann n xs) = liftM3 STData      (traverseAnnotsFor f ann)
                                                             (return n)               
                                                             (traverseAnnotsFor f xs) 
  traverseAnnotsFor f t                 = return t
                     
-----------------------------------------
-- Traverse Annots for NoInformation, Maybe, and lists
instance TraverseAnnotsFor Char

instance TraverseAnnotsFor a => TraverseAnnotsFor [a] where
 traverseAnnotsFor f []     = return []
 traverseAnnotsFor f (a:as) = liftM2 (:) (traverseAnnotsFor f a) 
                                         (traverseAnnotsFor f as)

instance (TraverseAnnotsFor a, TraverseAnnotsFor b) => TraverseAnnotsFor (a,b) where
 traverseAnnotsFor f (a,b) = liftM2 (,) (traverseAnnotsFor f a) 
                                        (traverseAnnotsFor f b)

------------------------------
-- Traverse Annots for SourceTerms
instance TraverseAnnotsFor SourceTerm where
 traverseAnnotsFor f e = 
  case e of
    (App iann it ie1 ie2)       -> do e1    <- traverseAnnotsFor f ie1
                                      e2    <- traverseAnnotsFor f ie2
                                      ann   <- traverseAnnotsFor f iann
                                      t     <- traverseAnnotsFor f it
                                      return (App ann t e1 e2)
    (Lam iann x i_t ie)         -> do t     <- traverseAnnotsFor f i_t
                                      e     <- traverseAnnotsFor f ie
                                      ann   <- traverseAnnotsFor f iann
                                      return (Lam ann x t e)
    (Fix iann ie)               -> do e     <- traverseAnnotsFor f ie
                                      ann   <- traverseAnnotsFor f iann
                                      return (Fix ann e)
    (Let iann idecls ie)        -> do decls <- traverseAnnotsFor f idecls
                                      e     <- traverseAnnotsFor f ie
                                      ann   <- traverseAnnotsFor f iann
                                      return (Let ann decls e)
    (Int iann n)                -> do ann   <- traverseAnnotsFor f iann
                                      return (Int ann n)
    (Lift ie)                   -> do e     <- traverseAnnotsFor f ie
                                      return (Lift e)
    (ArithOp iann nop ie1 ie2)  -> do e1    <- traverseAnnotsFor f ie1
                                      e2    <- traverseAnnotsFor f ie2
                                      ann   <- traverseAnnotsFor f iann
                                      return (ArithOp ann nop e1 e2)
    (Bool iann b)               -> do ann   <- traverseAnnotsFor f iann
                                      return (Bool ann b)
    (If iann ie ie1 ie2)        -> do e1    <- traverseAnnotsFor f ie1
                                      e2    <- traverseAnnotsFor f ie2
                                      e     <- traverseAnnotsFor f ie
                                      ann   <- traverseAnnotsFor f iann
                                      return (If ann e e1 e2)
    (BoolOp iann bop ie1 ie2)   -> do e1    <- traverseAnnotsFor f ie1
                                      e2    <- traverseAnnotsFor f ie2
                                      ann   <- traverseAnnotsFor f iann
                                      return (BoolOp ann bop e1 e2)
    (RelOp iann rop it ie1 ie2) -> do e1    <- traverseAnnotsFor f ie1
                                      e2    <- traverseAnnotsFor f ie2
                                      t     <- traverseAnnotsFor f it
                                      ann   <- traverseAnnotsFor f iann
                                      return (RelOp ann rop t e1 e2)
    (Tuple iann ies)            -> do es    <- traverseAnnotsFor f ies
                                      ann   <- traverseAnnotsFor f iann
                                      return (Tuple ann es)
    (Proj iann i j it ie)       -> do e     <- traverseAnnotsFor f ie
                                      t     <- traverseAnnotsFor f it
                                      ann   <- traverseAnnotsFor f iann
                                      return (Proj ann i j t e)
    (Poly ie)                   -> do e     <- traverseAnnotsFor f ie
                                      return (Poly e)
    (Spec ie)                   -> do e     <- traverseAnnotsFor f ie
                                      return (Spec e)
    (Polymorph ie)              -> do e     <- traverseAnnotsFor f ie
                                      return (Polymorph e)
    (Typed ie it)               -> do e     <- traverseAnnotsFor f ie
                                      t     <- traverseAnnotsFor f it
                                      return (Typed e t)
    (Data xs st)                -> do st'   <- traverseAnnotsFor f st     
                                      return (Data xs st')                
    (Construc iann name sts)    -> do sts'  <- traverseAnnotsFor f sts    
                                      ann   <- traverseAnnotsFor f iann   
                                      return (Construc ann name sts')     
    (Case iann st t xs)         -> do xs'   <- traverseAnnotsFor f xs     
                                      st'   <- traverseAnnotsFor f st     
                                      t'    <- traverseAnnotsFor f t      
                                      ann   <- traverseAnnotsFor f iann   
                                      return (Case ann st' t' xs')        
    _                           ->    return e
    
------------------------------                                                  
-- Traverse AnnotsConstructors
instance TraverseAnnotsFor a => TraverseAnnotsFor (Constructor a) where
 traverseAnnotsFor f (Constructor name xs) = 
    do xs' <- traverseAnnotsFor f xs
       return (Constructor name xs')

------------------------------                                                  
-- Traverse AnnotsData definitions
instance TraverseAnnotsFor DataDef where
 traverseAnnotsFor f (DataDef ann name sts cs) = 
    do sts' <- traverseAnnotsFor f sts
       cs'  <- traverseAnnotsFor f cs
       ann' <- traverseAnnotsFor f ann
       return (DataDef ann' name sts' cs')

---------
--end @--
---------

inspectfTC f t     = do t'  <- unfoldVbles False t
                        fail (f t')
inspectTC   t      = inspectfTC show t
           
debugTC       t = error (show t)

