module SpecMonad where

import IOExts 
import Maybe

import MyPrelude
import Error(Error(..))
import MoreMonad(liftM, liftM2, liftM3, liftM4)
import Monad(when)

import BasicData
import SourceLang
import qualified SourceLangPrint
import qualified SourceTypeCheckMonad
import ResidualLang
import ResidualLangPrint
import ResidualDataManagement
import MetaVbles
import SpecState
import SourceTypeCheckMonad(constrsOfData, unSTCM) -- Added by Ale

-- -------------------------------------------------------------
-- Monad to manage meta-variables 
-- -------------------------------------------------------------
data SpecM a = SM (State -> Error (a,State))
unSM (SM f) = f

runSM (SM f) = f initialState

-- -----------------------
-- The Monad!
instance Monad SpecM where
  return x   = SM (\s -> Success (x,s))
  SM x >>= f = SM (\s -> case x s of 
                          Error e         -> Error e
                          Success (v, s') -> unSM (f v) s'
                  )
  fail e = do str <- getTrace; fail' (e ++ "\n\n" ++ show str)

fail' e = SM (\_ -> Error e)

---------------------------------------------------------------
-- Toogling the flags
onSolveContextsFlag  = SM (\s -> Success ((), pokeSolveContextsFlag True s))
offSolveContextsFlag = SM (\s -> Success ((), pokeSolveContextsFlag False s))
solveContexts        = SM (\s -> let bf = peekSolveContextsFlag s 
                                  in Success (bf, s))
onEvidenceElimFlag   = SM (\s -> Success ((), pokeEvidenceElimFlag True s))
offEvidenceElimFlag  = SM (\s -> Success ((), pokeEvidenceElimFlag False s))
eliminateEvidence    = SM (\s -> let ee = peekEvidenceElimFlag s 
                                  in Success (ee, s))

---------------------------------------------------------------
-- Operations for metavariables of different kinds
setexcepts part e   = SM (\s -> Success (part (setexceptsmetavble e) s))
fresh      part     = SM (\s -> Success (part freshmetavble        s))
get        part x   = SM (\s -> Success (part (getmetavble x)      s))
set pname  part x v = do --traceSM ("Set "++pname++" to") (x,v)
                         set' part x v
set'       part x v = SM (\s -> Success (part (setmetavble x v)    s))
getTable   part     = SM (\s -> Success (part gettable             s))

remember   part v   = SM (\s -> Success (part (remembermetavble v) s))
forget     part v   = SM (\s -> Success (part (forgetmetavble v)   s))
seen       part v   = SM (\s -> Success (part (seenmetavble v)     s))
forgetAll  part     = SM (\s -> Success (part forgetallmetavbles   s))
pushMemory part     = SM (\s -> Success (part pushmemmetavble      s))
popMemory  part     = SM (\s -> Success (part popmemmetavble       s))

freshresidualtype   = liftM RTVar (fresh residualtype)
freshresidualscheme = liftM RSVar (fresh residualscheme)
freshtmptype        = liftM RTVar (fresh tmptype)
freshtmpscheme      = liftM RSVar (fresh tmpscheme)
freshevidence       = liftM EVar (fresh evidence)

newTypesFor _ [] = return []
newTypesFor fresh (st:sts) = 
   do nst  <- fresh
      nsts <- newTypesFor fresh sts
      return ((st,nst):nsts)

---------------------------------------------------------------
-- Operations for source typechecking
getSourceTCstate :: SpecM SourceTypeCheckMonad.SourceState
getSourceTCstate = SM (\s -> Success (peekSourceTCstate s, s))

setSourceTCstate :: SourceTypeCheckMonad.SourceState -> SpecM ()
setSourceTCstate stcs = SM (\s -> Success ((), pokeSourceTCstate stcs s))

sourceTypeCheck :: SourceTypeCheckMonad.STCM a -> SpecM a
sourceTypeCheck (SourceTypeCheckMonad.STCM f) = 
  do stcstate <- getSourceTCstate
     case f stcstate of
      Error msg            -> fail msg
      Success (a,new_stcs) -> do setSourceTCstate new_stcs
                                 return a

---------------------------------------------------------------
-- Operations for datatypes
addData :: RDataDef -> SpecM ()
forwardDataFor :: (DataName,ResType) -> SpecM ()

addData d = 
  SM (\s -> Success ((), updateData (\dd -> dd `mngrAddData` d) s))

forwardDataFor fd =
  SM (\s -> Success ((), updateData (\dd -> dd `mngrAddForwardData` fd) s))

getData :: SpecM RDataManager
getData = SM (\s -> Success (peekData s, s))

dataNamed, resolveDataNamed :: DataName -> SpecM (Preds, ResType)
dataNamed        d = dataNamed' False d
resolveDataNamed d = dataNamed' True d

dataNamed' :: Bool -> DataName -> SpecM (Preds, ResType)
dataNamed' b d = 
   do mngr <- getData
      case (mngrDataNamed mngr d) of 
        Error msg                                      -> fail msg
        Success (Right td)                             -> return ([],td)
        Success (Left (RDataDef name sts sss delta _)) ->
         do rep_ts <- if b then return (map (\t -> (t,t)) sts )
                           else newTypesFor freshresidualtype sts
            rep_ss <- newTypesFor freshresidualscheme sss
            let delta' = replaceIn (rep_ts,rep_ss) delta
            return (delta', RTData name (map snd rep_ts)
                                        (map snd rep_ss))
constructorRType c = 
   do mngr <- getData
      case (mngrDataOf mngr c) of 
        Error msg -> fail msg
        Success dmy@(RDataDef name sts sss delta cs, args) ->
         do rep_ts <- newTypesFor freshresidualtype sts
            rep_ss <- newTypesFor freshresidualscheme sss
            args'  <- unfoldAll args
            let cType = buildConstructorType rep_ts rep_ss name args'
            cType' <- unfoldAll cType
            delta' <- unfoldAll delta
            let delta'' = replaceIn (rep_ts,rep_ss) delta'
            return (delta'', cType')

-------------------------
---- Begin Alejandro ----
-------------------------

-- Get the residual type of a constructor
getconstructorRType c = 
   do mngr <- getData
      case (mngrDataOf mngr c) of 
        Error msg -> fail msg
        Success dmy@(RDataDef name sts sss delta cs, args) ->
		    let [(Constructor name [tv])] = filter f cs
            in do return tv
            where f (Constructor name [tv]) = name == c 


getconstructorRnames d = 
   do mngr <- getData
      case (mngrDataNamed mngr d) of 
        Error msg -> fail msg
        Success (Left dmy@(RDataDef name sts sss delta cs)) ->
		    let cnames = map f cs
                    in return cnames
        where f (Constructor name _) = name  


---------------------------------------------------------------
-- More Operations for datatypes (dynamic ones)

-- Retrieve each constructor and its source type.
constrsOfSourceData :: DataName -> SpecM [(ConstructorName, SourceType)]
constrsOfSourceData d = do ss <- getSourceTCstate
                           case (unSTCM (constrsOfData d) ss) of 
                                Error msg       -> fail msg 
                                Success (cs, _) -> return cs 
-------------------------
---- End Alejandro -----
-------------------------

---------------------------------------------------------------
-- Operations for temporary substition
gettmptype   l x   = SM (\s->let tts = peektmptype s 
                              in if l>=length tts
                                  then Error (outboundIn "gettmptype")
                                  else Success (fst (getmetavble x (tts!!l)), s))
gettmpscheme l x   = SM (\s->let tss = peektmpscheme s
                              in if l>=length tss
                                  then Error (outboundIn "gettmpscheme")
                                  else Success (fst (getmetavble x (tss!!l)), s))
gettmpLevels       = SM (\s->let tts = peektmptype s 
                              in Success (length tts, s))
newTemporarySubst  = do --traceSM "newTempSubst[" ()
                        newTemporarySubst'
newTemporarySubst' = SM (\s->Success ((), updatetmptype   (initialmetavbles:) $
                                          updatetmpscheme (initialmetavbles:) $
                                          s))
discardUnification = do --traceSM "Discarding unification]" ()
                        discardUnification'
discardUnification'= SM (\s->Success ((), updatetmptype tail   $
                                          updatetmpscheme tail $ s))
confirmUnification = do --traceSM "Confirming unification]" ()
                        confirmUnification'
confirmUnification'= SM (\s->Success ((), stateConfirmUnification s))

---------------------------------------------------------------
-- Operations for generic variables
resetGeneric         = SM (\s-> Success ((),pokeGeneric initialVbleGenerator s))
enterQuantifierScope = SM (\s-> Success ((),updateQuantifierScope (+1)       s))
leaveQuantifierScope = SM (\s-> Success ((),updateQuantifierScope (+(-1))    s))
      
scopeLevel = SM (\s -> Success (peekScopeLevel s, s))
maxRTGen   = SM (\s -> Success (peekMaxRTGen   s, s))
maxRSGen   = SM (\s -> Success (peekMaxRSGen   s, s))

aRTGenFor tv = 
  SM (\s ->
       let (gtv,s') = updateRTGen (\(ti,ta) -> lookUpdate tv ti ta) s
           l        = peekScopeLevel s
        in l `seq` gtv `seq` Success (RTGen l gtv, s')
     )
aRSGenFor sv = 
  SM (\s ->
       let (gsv,s') = updateRSGen (\(si,sa) -> lookUpdate sv si sa) s
           l        = peekScopeLevel s
        in l `seq` gsv `seq` Success (RSGen l gsv, s')
     )

isthereaRTVarFor gtv = SM (\s -> let mtv = projectRTGen (lookup gtv) s
                                  in Success (fmap RTVar mtv, s))
isthereaRSVarFor gsv = SM (\s -> let msv = projectRSGen (lookup gsv) s
                                  in Success (fmap RSVar msv, s))

aRTVarFor gtv = 
  do defined <- isthereaRTVarFor gtv
     case defined of
      Just tv -> return tv
      Nothing -> do tv <- freshresidualtype
                    tv `istheRTVarFor` gtv
                    return tv
aRSVarFor gsv = 
  do defined <- isthereaRSVarFor gsv
     case defined of
      Just sv -> return sv
      Nothing -> do sv <- freshresidualscheme
                    sv `istheRSVarFor` gsv
                    return sv
      
istheRTVarFor (RTVar tv) gtv = 
  SM (\s -> let (_, s') = updateRTGen (\(gti,ta) -> ((),gti,(gtv,tv):ta)) s
             in Success ((), s'))
istheRSVarFor (RSVar sv) gsv = 
  SM (\s -> let (_, s') = updateRSGen (\(gsi,sa) -> ((),gsi,(gsv,sv):sa)) s
             in Success ((), s'))

---------------------------------------------------------------
-- Operations for residual term variables
freshresidualtermBasedOn x = nextName x

nextName x = 
  SM (\s -> let (ix,s') = updateTermVbles (getNext x) s
             in Success (mkId x ix, s'))

-- %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
-- Unfolding variables for different kinds of objects
-- %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
class MetaVbles t where
  indir      :: t -> SpecM t          
  tmpindir   :: t -> SpecM t
  fullindir  :: t -> SpecM t
  tmpindir'  :: Int -> Int -> t -> SpecM t
  fullindir' :: Int -> t -> SpecM t
  indirOnly  :: [t] -> t -> SpecM t          

  indir          t = return t
  tmpindir'  _ _ t = return t
  fullindir' _   t = return t
  tmpindir       t = do ml <- gettmpLevels
                        tmpindir' ml ml t

  fullindir      t = do ml <- gettmpLevels
                        fullindir' ml t
  indirOnly  _   t = return t

-----
class Unfoldable t where
  unfoldG               ::    (ResType -> SpecM ResType) 
                           -> (ResTypeScheme -> SpecM ResTypeScheme) 
                           -> t -> SpecM t 
  unfoldVbles           :: t -> SpecM t 
  unfoldAll, unfoldAll' :: t -> SpecM t
  unfoldOnly            :: [ResType] -> t -> SpecM t

  unfoldG _ _ t = return t
  unfoldVbles t = unfoldG indir indir t
  unfoldAll'  t = unfoldG fullindir fullindir t
  unfoldAll   t = unfoldAll' t
  unfoldOnly ts t = unfoldG (indirOnly ts) return t
                   
instance MetaVbles Int
instance Unfoldable Int
instance MetaVbles Char
instance Unfoldable Char 
instance MetaVbles Bool
instance Unfoldable Bool 
instance MetaVbles ()
instance Unfoldable () 

---------------------------------
-- Unfolding for data definitions
---------------------------------
instance Unfoldable RDataDef where
  -----
  unfoldG f g (RDataDef dn ts ss delta cs) =
     liftM2 (RDataDef dn ts ss) (unfoldG f g delta) (unfoldG f g cs)

instance Unfoldable a => Unfoldable (Constructor a) where
  -----
  unfoldG f g (Constructor c args) = 
     liftM (Constructor c) (unfoldG f g args)

---------------------------------
-- Unfolding for residual types
---------------------------------
instance MetaVbles ResType where
  -----
  indir vi@(RTVar i) = do pa <- get residualtype i
                          case pa of
                            Yes a -> indir a
                            No _  -> return vi
  indir t            = return t
 
  -----
  tmpindir' 0  ml t            = return t                    
  tmpindir' cl ml vi@(RTVar i) = do pa <- gettmptype (cl-1) i
                                    case pa of
                                      Yes a -> tmpindir' ml ml a
                                      No _  -> tmpindir' (cl-1) ml vi
  tmpindir' _  _  t            = return t

  -----
  fullindir' ml vi@(RTVar _) = do nvi <- indir vi
                                  fvi <- tmpindir' ml ml nvi
                                  if fvi==nvi
                                   then return fvi
                                   else fullindir' ml fvi
  fullindir' _  t            = return t

  -----
  indirOnly ts vi@(RTVar i) = if vi `elem` ts
                               then do pa <- get residualtype i
                                       case pa of
                                         Yes a -> indir a
                                         No _  -> return vi
                               else return vi
  indirOnly _  t            = return t
---------------------------------
instance Unfoldable ResType where
  -----
  unfoldG f g vi@(RTVar _)         = do t <- f vi
                                        case t of
                                         RTVar _ -> return t
                                         _       -> unfoldG f g t
  unfoldG f g (RTTuple  taus)      = liftM  RTTuple     (unfoldG f g taus)
  unfoldG f g (RTTupleS taus)      = liftM  RTTupleS    (unfoldG f g taus) 
  unfoldG f g (RTFun    tau1 tau2) = liftM2 RTFun       (unfoldG f g tau1)
                                                        (unfoldG f g tau2)
  unfoldG f g (RTStaticExp e)      = liftM  RTStaticExp (unfoldG f g e)    
  unfoldG f g (RTFixS   taus 
                        tau sigma) = liftM3 RTFixS      (unfoldG f g taus)
                                                        (unfoldG f g tau)
                                                        (unfoldG f g sigma)
  unfoldG f g (RTTypedS taus 
                        tau sigma) = liftM3 RTTypedS    (unfoldG f g taus)
                                                        (unfoldG f g tau)
                                                        (unfoldG f g sigma)
  unfoldG f g (RTPoly   sigma)     = liftM  RTPoly      (unfoldG f g sigma)
  unfoldG f g (RTData d ts ss)     = liftM2 (RTData d)  (unfoldG f g ts)  
                                                        (unfoldG f g ss)  
  unfoldG f g (RTConstruc c taus)  = liftM  (RTConstruc c)                
                                                        (unfoldG f g taus)
  unfoldG _ _ t                    = return t

---------------------------------
-- Unfolding for residual types schemes
---------------------------------
instance MetaVbles ResTypeScheme where
  -----
  indir vi@(RSVar i) = do ps <- get residualscheme i
                          case ps of
                            Yes s -> indir s
                            No _  -> return vi
  indir t            = return t

  -----
  tmpindir' 0  _  s            = return s
  tmpindir' cl ml vi@(RSVar i) = do ps <- gettmpscheme (cl-1) i
                                    case ps of
                                      Yes s -> tmpindir' ml ml s
                                      No _  -> tmpindir' (cl-1) ml vi
  tmpindir' _  _  s            = return s

  -----
  fullindir' ml vi@(RSVar _) = do nvi <- indir vi
                                  fvi <- tmpindir' ml ml nvi
                                  if fvi==nvi
                                   then return fvi
                                   else fullindir' ml fvi
  fullindir' _  s = return s

---------------------------------
instance Unfoldable ResTypeScheme where
  -----
  unfoldG f g vi@(RSVar _)        = do s <- g vi
                                       case s of
                                        RSVar _ -> return s
                                        _       -> unfoldG f g s
  unfoldG f g (RSForAll tm sm qt) = liftM (RSForAll tm sm) (unfoldG f g qt)
  unfoldG f g (RSTupleS ss)       = liftM RSTupleS         (unfoldG f g ss) 
  unfoldG _ _ s                   = return s

---------------------------------
-- Unfolding for qualified types
---------------------------------
instance MetaVbles QualType
instance MetaVbles [a]
instance MetaVbles (a,b)

-----
instance Unfoldable QualType where
  unfoldG f g (ps :=> t) = liftM2 (:=>) (unfoldG f g ps)
                                        (unfoldG f g t)

-----
instance Unfoldable a => Unfoldable [a] where
  unfoldG f g []     = return []
  unfoldG f g (x:xs) = liftM2 (:) (unfoldG f g x) (unfoldG f g xs)

-----
instance (Unfoldable a, Unfoldable b) => Unfoldable (a,b) where
  unfoldG f g (a,b) = liftM2 (,) (unfoldG f g a) (unfoldG f g b)
                         
---------------------------------
-- Unfolding for predicates
---------------------------------
instance MetaVbles Pred

instance Unfoldable Pred where
  ----------
  unfoldG f g (IsMG s1 s2)            = liftM2 IsMG           (unfoldG f g s1)
                                                              (unfoldG f g s2)
  unfoldG f g (IsMGmorph t s1 s2)     = liftM3 IsMGmorph      (unfoldG f g t)
                                                              (unfoldG f g s1)
                                                              (unfoldG f g s2)
  unfoldG f g (IsInt  t)              = liftM  IsInt          (unfoldG f g t)
  unfoldG f g (IsBool t)              = liftM  IsBool         (unfoldG f g t)
  unfoldG f g (IsFunS t t')           = liftM2 IsFunS         (unfoldG f g t)
                                                              (unfoldG f g t')
  unfoldG f g (IsFixS t t')           = liftM2 IsFixS         (unfoldG f g t)
                                                              (unfoldG f g t')
  unfoldG f g (IsIf tr tb t1 t2)      = liftM4 IsIf           (unfoldG f g tr)
                                                              (unfoldG f g tb)
                                                              (unfoldG f g t1)
                                                              (unfoldG f g t2)
  unfoldG f g (IsCase tr te cts)      = liftM3 IsCase         (unfoldG f g tr)
                                                              (unfoldG f g te)
                                                              (unfoldG f g cts)
  unfoldG f g (IsArithOp nop t t1 t2) = liftM3 (IsArithOp nop)(unfoldG f g t) 
                                                              (unfoldG f g t1)
                                                              (unfoldG f g t2)
  unfoldG f g (IsBoolOp  bop t t1 t2) = liftM3 (IsBoolOp bop) (unfoldG f g t) 
                                                              (unfoldG f g t1)
                                                              (unfoldG f g t2)
  unfoldG f g (IsRelOp   rop t t1 t2) = liftM3 (IsRelOp rop)  (unfoldG f g t) 
                                                              (unfoldG f g t1)
                                                              (unfoldG f g t2)
  unfoldG f g (IsConstrOf sdt t)      = liftM  (IsConstrOf sdt)
                                                              (unfoldG f g t) 
  unfoldG f g (Guarded t d)           = liftM2 Guarded        (unfoldG f g t)
                                                              (unfoldG f g d)
  unfoldG f g (GuardedNeg t d)        = liftM2 GuardedNeg     (unfoldG f g t)
                                                              (unfoldG f g d)
  unfoldG f g (GuardedConstr c t d)   = liftM2 (GuardedConstr c)
                                                              (unfoldG f g t)
                                                              (unfoldG f g d)
  -------------------------
  ---- Begin Alejandro ----
  -------------------------
  unfoldG f g (GuardedDinData c t p) = liftM2 (GuardedDinData c)
                                                  (unfoldG f g t)
                                                  (unfoldG f g p)
  unfoldG f g (HasC t1 c t2) = liftM2 (\ut1-> \ut2 ->HasC ut1 c ut2) 
                                      (unfoldG f g t1)
                                      (unfoldG f g t2)
  unfoldG f g (ConstrsOf t) = liftM ConstrsOf (unfoldG f g t)


  unfoldG f g (IndexOf t) = liftM IndexOf (unfoldG f g t)


  unfoldG f g (UnifyData td c t) = liftM2 (\utd-> \ut -> 
                                           UnifyData utd c ut) 
                                          (unfoldG f g td)
                                          (unfoldG f g t)
  -------------------------
  ------ End Alejandro ----
  -------------------------
  unfoldG f g (Unify t1 t2)           = liftM2 Unify          (unfoldG f g t1)
                                                              (unfoldG f g t2)
  unfoldG _ _ pred                    = return pred

---------------------------------
instance MetaVbles ResTerm

instance Unfoldable ResTerm where
 unfoldG f g (RApp e1 e2)         = liftM2 RApp           (unfoldG f g e1)
                                                          (unfoldG f g e2)
 unfoldG f g (RLam x e)           = liftM  (RLam x)       (unfoldG f g e)
 unfoldG f g (RFix e)             = liftM  RFix           (unfoldG f g e)
 unfoldG f g (RLet xes e)         = liftM2 RLet           (unfoldG f g xes)
                                                          (unfoldG f g e)
 unfoldG f g (RArithOp nop e1 e2) = liftM2 (RArithOp nop) (unfoldG f g e1)
                                                          (unfoldG f g e2)
 unfoldG f g (RIf e e1 e2)        = liftM3 RIf            (unfoldG f g e)
                                                          (unfoldG f g e1)
                                                          (unfoldG f g e2)
 unfoldG f g (RBoolOp bop e1 e2)  = liftM2 (RBoolOp bop)  (unfoldG f g e1)
                                                          (unfoldG f g e2)
 unfoldG f g (RRelOp rop e1 e2)   = liftM2 (RRelOp rop)   (unfoldG f g e1)
                                                          (unfoldG f g e2)
 unfoldG f g (RTuple es)          = liftM  RTuple         (unfoldG f g es) 
 unfoldG f g (RProj i n e)        = liftM  (RProj i n)    (unfoldG f g e)
 unfoldG f g (RTupleS es)         = liftM  RTupleS        (unfoldG f g es) 
 unfoldG f g (RProjS i n e)       = liftM  (RProjS i n)   (unfoldG f g e)
 unfoldG f g (RData ds e)         = liftM  (RData ds)     (unfoldG f g e) 
 unfoldG f g (RConstruc c es)     = liftM  (RConstruc c)  (unfoldG f g es)
 unfoldG f g (RCase e cs)         = liftM2 RCase          (unfoldG f g e) 
                                                          (unfoldG f g cs)
 unfoldG f g (REFill v e)         = liftM2 REFill         (unfoldG f g v)
                                                          (unfoldG f g e)
 unfoldG f g (RELam h e)          = liftM  (RELam h)      (unfoldG f g e)
 unfoldG f g (RELet x e1 e2)      = liftM2 (RELet x)      (unfoldG f g e1)
                                                          (unfoldG f g e2)
 unfoldG f g (REApp e v)          = liftM2 REApp          (unfoldG f g e)
                                                          (unfoldG f g v)
 unfoldG f g (REIf v e1 e2)       = liftM3 REIf           (unfoldG f g v)
                                                          (unfoldG f g e1)
                                                          (unfoldG f g e2)
 unfoldG f g (RECase v pes)       = liftM2 RECase         (unfoldG f g v)
                                                          (unfoldG f g pes)
 unfoldG f g (REAppS e ea)        = liftM2 REAppS         (unfoldG f g e)
                                                          (unfoldG f g ea)
 unfoldG _ _ e                    = return e

---------------------------------
instance MetaVbles EvTerm

instance Unfoldable EvTerm where
 unfoldG f g (ECxt c)         = liftM  ECxt         (unfoldG f g c)
 unfoldG f g (EComp v e)      = liftM2 EComp        (unfoldG f g v)
                                                    (unfoldG f g e)
 unfoldG f g (EStaticExp e)   = liftM  EStaticExp   (unfoldG f g e)
 unfoldG f g (EAppS e ea)     = liftM2 EAppS        (unfoldG f g e)
                                                    (unfoldG f g ea)
 unfoldG f g (ETupleS vs)     = liftM  ETupleS      (unfoldG f g vs)
 unfoldG f g (EProjS i n v)   = liftM  (EProjS i n) (unfoldG f g v)
 unfoldG _ _ v                = return v

---------------------------------
instance MetaVbles CxtTerm

instance Unfoldable CxtTerm where
 unfoldG f g (CELet x c e)   = liftM2 (CELet x)  (unfoldG f g c)
                                                 (unfoldG f g e)
 unfoldG f g (CELam xs c)    = liftM  (CELam xs) (unfoldG f g c)
 unfoldG f g (CEApp c v)     = liftM2 CEApp      (unfoldG f g c)
                                                 (unfoldG f g v)
 unfoldG _ _ c               = return c

instance Unfoldable a => Unfoldable (Maybe a) where
  unfoldG f g Nothing  = return Nothing
  unfoldG f g (Just a) = do a' <- unfoldG f g a
                            return (Just a')

---------------------------------
---------------------------------
-- Debugging tools
---------------------------------
---------------------------------
-- Operations for traces
getTrace = SM (\s -> let str = peekTrace s
                      in Success (str, s)
              )
onTracing   = SM (\s -> Success ((), pokeTraceFlag True s))
offTracing  = SM (\s -> Success ((), pokeTraceFlag False s))
traceActive = SM (\s -> let trf = peekTraceFlag s 
                         in Success (trf, s))
stopTrace = do yes <- traceActive
               if yes
                then do str <- getTrace
                        return str -- fail str --return str
                else return ()
traceSM msg x = -- return ()
             do yes <- traceActive
                if yes 
                 then do x' <- unfoldAll x  
                         let msgf = "\n" ++ msg ++ inspect_function x'
                          in if (trace msgf msgf) =="" 
                              then return ()
                              else return ()
                 else return ()

traceSMcooked msg x = 
             do yes <- traceActive
                if yes 
                 then do x' <- unfoldAll x  
                         let msgf = "\n" ++ msg ++ inspect_function x'
                          in if (trace msgf msgf) =="" 
                              then return ()
                              else return ()
                 else return ()

traceSMraw msg x = 
                let msgf = "\nRAW " ++ msg ++ show x
                 in if (trace msgf msgf) =="" 
                     then return ()
                     else return ()

---------------------------------
inspectf f t     = do t' <- unfoldAll t
                      fail (f t')

class Show t => Inspectable t where
  inspect_function :: t -> String
  inspect_function = show

inspect   t      = inspectf inspect_function t

instance Inspectable ResTerm where
  inspect_function = ppresidualterm

instance Inspectable ResType where
  inspect_function = ppresidualtype

instance Inspectable ResTypeScheme where
  inspect_function = ppresidualscheme

instance Inspectable Pred where
  inspect_function = ppresidualpredicate

instance Inspectable a => Inspectable [a] where
  inspect_function = (\xs -> "[\n" ++ xs ++ "]\n") . concat
                     . map (\a -> inspect_function a ++ "\n")

instance Inspectable a => Inspectable (Maybe a) where
  inspect_function Nothing = "Nothing"
  inspect_function (Just a) = "Just " ++ inspect_function a

instance (Inspectable a, Inspectable b) => Inspectable (a,b) where
  inspect_function (a,b) = "(" ++ inspect_function a ++ 
                           "," ++ inspect_function b ++
                           ")"

instance Inspectable EvTerm
instance Inspectable CxtTerm
instance Inspectable Int
instance Inspectable Char
instance Inspectable Bool
instance Inspectable ()

instance Inspectable SourceTerm where
  inspect_function = SourceLangPrint.ppsourceexpr

instance Inspectable SourceType where
  inspect_function = SourceLangPrint.ppsourcetype

--inspect   t      = inspectf show t
inspecte  t      = inspectf ppresidualterm t
inspectt  t      = inspectf ppresidualtype t
inspects  t      = inspectf ppresidualscheme t
inspectd  delta  = inspectf ppresidualpredicate delta
inspectds deltas = inspectf ppresidualpredicates deltas
           
debug       t = error (show t)
getState    = SM (\s->Success (s,s))
debugState    = SM (\s->error (show s))
debugSubst    = SM (\s->error (show $ fst $ getflattable $ getSubst $ s))
debugTmpSubst = SM (\s->error (show $ map fst $ map getflattable 
                                    $ getTmpSubst $ s))

valueInSM f = case runSM f of Success (v,_) -> v; Error msg -> error msg

